//***************************************************************************
//* bind.h
//*
//* description    : Modelization of a RSIP bind
//* copyright      : (C) 2001 Cedric de Launois
//*                  Universite Catholique de Louvain, Belgium
//* begin          : Sept 24 2001
//* email          : cedric@delaunois.org
//***************************************************************************

/****************************************************************************
 * Copyright (c) 1998 Free Software Foundation, Inc.                        *
 *                                                                          *
 * Permission is hereby granted, free of charge, to any person obtaining a  *
 * copy of this software and associated documentation files (the            *
 * "Software"), to deal in the Software without restriction, including      *
 * without limitation the rights to use, copy, modify, merge, publish,      *
 * distribute, distribute with modifications, sublicense, and/or sell       *
 * copies of the Software, and to permit persons to whom the Software is    *
 * furnished to do so, subject to the following conditions:                 *
 *                                                                          *
 * The above copyright notice and this permission notice shall be included  *
 * in all copies or substantial portions of the Software.                   *
 *                                                                          *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS  *
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF               *
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.   *
 * IN NO EVENT SHALL THE ABOVE COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,   *
 * DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR    *
 * OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR    *
 * THE USE OR OTHER DEALINGS IN THE SOFTWARE.                               *
 *                                                                          *
 * Except as contained in this notice, the name(s) of the above copyright   *
 * holders shall not be used in advertising or otherwise to promote the     *
 * sale, use or other dealings in this Software without prior written       *
 * authorization.                                                           *
 ****************************************************************************/

class Bind;

#ifndef _BIND_H_
#define _BIND_H_

#include <deque>
#include "resources.h"
#include "common.h"
#include "tunnel.h"

using namespace std;

class Bind {
 public:
	
	~Bind();
	// Destructor
	// PRE : /
	// POST: release the allocated resources


	Bind();
	// Constructor
	// PRE : /
	// POST: initialize the bind
	

	int addResource(Resource & res, bool aDC, bool pDC, 
			unsigned char nPorts, bool listen);
	// PRE : - r: resource desired or required (depending of aDC or pDC)
	//       - aDC: true for "don't care" address
	//       - pDC: true for "don't care" port
	//       - nbPorts: - number of consecutive ports starting at r.port
	//                    for the address r.address (max 256 ports)
	//                  - 0 to request all the ports for this address
	//       - listen : true to request a listen port
	// POST: returns -1 if the resource is not available for RSIP
	//                0 if resource not available
	//                1 if the resource requested was succesfully assigned
	//                2 if another resource was successfully assigned
	//                 (only if aDC or pDC is set)


	int addResource(Address &a, deque<uint16_t> &ports, 
			bool aDC, bool listen);
	// Assign multiple ports of the same address
	// PRE : - ports is the list of ports requested
	//       - the address is a
	//       - aDC : true for a "Don't care" address
	//       - listen : true to request a listen port
	// POST: returns -1 if the resource is not available for RSIP
	//                0 if the resources were not all available
	//                1 if the resources with the address desired
	//                  were succesfully assigned
	//                2 if another address was chosen for the ports


	inline void setLeaseTime(int lt) { deadline = time(0)+lt; }
	// PRE : lt >= 0
	// POST: lease time for this bind set


	inline int getLeaseTime() { return deadline-time(0); }
	// PRE : /
	// POST: return the instantaneous lease time for this bind


	void clearResources();
	// PRE : /
	// POST: free the resources of the bind


	inline Resource getFirst() { return resList->front().r; }
	// Returns the first resource of the bind


	inline unsigned int nbPorts() { return nPorts; }
	// Returns the number of ports owned by this bind
	// Full IP addresses count for 0 ports


	inline unsigned int nbIPs() { return nIPs; }
	// Returns the number of full IP addresses owned by this bind


	static inline void setPool(Resources* p) { pool = p; }
	// PRE : p is initialized with resources
	// POST: p is the pool of resources to be leased for all
	//       future binds


	Tunnel* tunnel;         // The tunnel for this resource
	int bindID;             // ID of the bind
	int deadline;           // deadline for this bind
	                        // will expire

 private:

	typedef struct resourceElement {

		Resource r;
		// The couple <address, port>

		unsigned char range;
		// Number of consecutive ports (< 256) starting at r.port
		// If range is 0, then the resource is the full IP

	} RE;

	unsigned int nPorts;    // The number of ports of this bind
	unsigned int nIPs;      // The number of IPs of this bind
	deque<RE>* resList;     // List of resources owned by this bind

	static Resources* pool; // The pool of resources for all binds
};

#endif








