//***************************************************************************
//* iptunnel.cpp
//*
//* description    : basic creation of tunnels
//* copyright      : (C) 2001 Cedric de Launois
//*                  Universite Catholique de Louvain, Belgium
//* begin          : Sept 13 2001
//* email          : cedric@delaunois.org
//*
//* Inspired from iptunnel.c by Alexey Kuznetsov, <kuznet@ms2.inr.ac.ru>
//***************************************************************************

/****************************************************************************
 * Copyright (c) 1998 Free Software Foundation, Inc.                        *
 *                                                                          *
 * Permission is hereby granted, free of charge, to any person obtaining a  *
 * copy of this software and associated documentation files (the            *
 * "Software"), to deal in the Software without restriction, including      *
 * without limitation the rights to use, copy, modify, merge, publish,      *
 * distribute, distribute with modifications, sublicense, and/or sell       *
 * copies of the Software, and to permit persons to whom the Software is    *
 * furnished to do so, subject to the following conditions:                 *
 *                                                                          *
 * The above copyright notice and this permission notice shall be included  *
 * in all copies or substantial portions of the Software.                   *
 *                                                                          *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS  *
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF               *
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.   *
 * IN NO EVENT SHALL THE ABOVE COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,   *
 * DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR    *
 * OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR    *
 * THE USE OR OTHER DEALINGS IN THE SOFTWARE.                               *
 *                                                                          *
 * Except as contained in this notice, the name(s) of the above copyright   *
 * holders shall not be used in advertising or otherwise to promote the     *
 * sale, use or other dealings in this Software without prior written       *
 * authorization.                                                           *
 ****************************************************************************/

#include "iptunnel.h"

typedef struct
{
	__u8 family;
	__u8 bytelen;
	__s16 bitlen;
	__u32 data[4];
} inet_prefix;

int get_addr_1(inet_prefix *addr, char *name, int family)
{
	char *cp;
	unsigned char *ap = (unsigned char*)addr->data;
	int i;

	memset(addr, 0, sizeof(*addr));

	if (strcmp(name, "default") == 0 || strcmp(name, "any") == 0) {
		addr->family = family;
		addr->bytelen = (family == AF_INET6 ? 16 : 4);
		addr->bitlen = -1;
		return 0;
	}

	if (strchr(name, ':')) {
		addr->family = AF_INET6;
		if (family != AF_UNSPEC && family != AF_INET6)
			return -1;
		if (inet_pton(AF_INET6, name, addr->data) <= 0)
			return -1;
		addr->bytelen = 16;
		addr->bitlen = -1;
		return 0;
	}

	addr->family = AF_INET;
	if (family != AF_UNSPEC && family != AF_INET)
		return -1;
	addr->bytelen = 4;
	addr->bitlen = -1;
	for (cp=name, i=0; *cp; cp++) {
		if (*cp <= '9' && *cp >= '0') {
			ap[i] = 10*ap[i] + (*cp-'0');
			continue;
		}
		if (*cp == '.' && ++i <= 3)
			continue;
		return -1;
	}
	return 0;
}

__u32 get_addr32(char *name)
{
	inet_prefix addr;
	if (get_addr_1(&addr, name, AF_INET)) {
		fprintf(stderr, "ip: %s is invalid IPv4 address\n", name);
		exit(1);
	}
	return addr.data[0];
}

int do_ioctl_get_ifindex(char *dev)
{
	struct ifreq ifr;
	int fd;
	int err;

	strcpy(ifr.ifr_name, dev);
	fd = socket(AF_INET, SOCK_DGRAM, 0);
	err = ioctl(fd, SIOCGIFINDEX, &ifr);
	close(fd);
	if (err) {
	  perror("ioctl");
	  return -1;
	}
	return ifr.ifr_ifindex;
}

int do_ioctl(int cmd, char *name, struct ip_tunnel_parm *p)
{
	struct ifreq ifr;
	int fd;
	int err;

	strncpy(ifr.ifr_name, name, IFNAMSIZ);
	ifr.ifr_ifru.ifru_data = (char*)p;
	fd = socket(AF_INET, SOCK_DGRAM, 0);
	err = ioctl(fd, cmd, &ifr);
	close(fd);
	if (err) {
	  perror("ioctl");
	  return -1;
	} else err = 1;

	return err;
}

int do_iptunnel(const char* cmd, const char* name, __u8 proto, char* local, char* remote, char* medium)
{
        int iocmd;
	char* tunl = "tunl0";
	char* gre = "gre0";
	char* sit = "sit0";
  	struct ip_tunnel_parm p;

	memset(&p, 0, sizeof(p));
	p.iph.protocol = proto;
	p.iph.saddr = get_addr32(local);
	p.iph.daddr = get_addr32(remote);
	p.iph.version = 4;
	p.iph.ihl = 5;

	if (medium) p.link = do_ioctl_get_ifindex(medium);
	else p.link = NULL;

	if (p.link==-1) return -1;

	strncpy(p.name, name, IFNAMSIZ);
#ifndef IP_DF
#define IP_DF		0x4000		/* Flag: "Don't Fragment"	*/
#endif
	p.iph.frag_off = htons(IP_DF);

	if (strcmp(cmd, "add") == 0) iocmd = SIOCADDTUNNEL;
	else if (strcmp(cmd, "del") == 0) iocmd = SIOCDELTUNNEL;
	else {
	  fprintf(stderr, "Unknown command %s.\n", cmd);
	  return -1;
	}

	switch (p.iph.protocol) {
	case IPPROTO_IPIP:
	  return do_ioctl(iocmd, tunl, &p);
	case IPPROTO_GRE:
	  return do_ioctl(iocmd, gre, &p);
	case IPPROTO_IPV6:
	  return do_ioctl(iocmd, sit, &p);
	default:	
	  fprintf(stderr, "cannot determine tunnel mode (ipip, gre or sit)\n");
	  return -1;
	}

	return -1;
}



