//***************************************************************************
//* listener.cpp
//*
//* description    : class Listener
//*                  Listen for data and accept new connections
//* copyright      : (C) 2001 Cedric de Launois
//*                  Universite Catholique de Louvain, Belgium
//* begin          : Sept 13 2001
//* email          : cedric@delaunois.org
//***************************************************************************

/****************************************************************************
 * Copyright (c) 1998 Free Software Foundation, Inc.                        *
 *                                                                          *
 * Permission is hereby granted, free of charge, to any person obtaining a  *
 * copy of this software and associated documentation files (the            *
 * "Software"), to deal in the Software without restriction, including      *
 * without limitation the rights to use, copy, modify, merge, publish,      *
 * distribute, distribute with modifications, sublicense, and/or sell       *
 * copies of the Software, and to permit persons to whom the Software is    *
 * furnished to do so, subject to the following conditions:                 *
 *                                                                          *
 * The above copyright notice and this permission notice shall be included  *
 * in all copies or substantial portions of the Software.                   *
 *                                                                          *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS  *
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF               *
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.   *
 * IN NO EVENT SHALL THE ABOVE COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,   *
 * DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR    *
 * OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR    *
 * THE USE OR OTHER DEALINGS IN THE SOFTWARE.                               *
 *                                                                          *
 * Except as contained in this notice, the name(s) of the above copyright   *
 * holders shall not be used in advertising or otherwise to promote the     *
 * sale, use or other dealings in this Software without prior written       *
 * authorization.                                                           *
 ****************************************************************************/

#include "listener.h"

#define for_each(map,i) for (i = map.begin(); i != map.end(); ++i )


Listener::Listener(Options* o) 
{
	assert(o!=NULL);
	opts = o;
}



Listener::~Listener() 
{
	map<int, TCPEndPoint*>::const_iterator i;
	map<int, NetworkEndPoint*>::const_iterator j;

	stop();

	for_each(listeners, i) delete i->second;
	for_each(connections, j) delete j->second;
}



bool Listener::start() 
{
	if (!listeners.size()) return false;

	if (pthread_create(&listener, NULL, (void* (*) (void*)) runListener, (void*)this))
		return false;
	else return true;
}



bool Listener::stop() 
{
	if (pthread_cancel(listener)!=0) return false;

	pthread_join(listener, NULL);
	return true;
}



void Listener::doListen()
{
	fd_set readfds;
	int max=0, retval=0;
	ProtocolHandler ph(opts);

	pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, NULL);


	while (1) {

		buildSelectList(readfds, max);


		// This function may be canceled by a pthread_cancel
		pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL);
		retval = select(max+1, &readfds, NULL, NULL, NULL);
		pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, NULL);

		if (retval==0) continue;
		if (retval>=1) dispatch(readfds, ph);
		else break; // A pthread_cancel was received. Stopping.
	}
}



void Listener::dispatch(fd_set &readfds, ProtocolHandler &ph) 
{
	TCPEndPoint* newtep;
	map<int, TCPEndPoint*>::const_iterator i;
	map<int, NetworkEndPoint*>::iterator j;

	for_each(listeners, i) {
		if (!FD_ISSET(i->first, &readfds)) continue;
		if ((newtep = i->second->acceptConnection(0))) {
			if (connections.size()-1 >= opts->max_tcp_clients) {
				logmsg(LOG_INFO, "Refusing opening connection (maximum reached)\n");
				delete newtep;
			} else {
				connections.insert(make_pair(newtep->getfd(), newtep));
				logmsg(LOG_INFO, "Connection open\n");
			}
		}
	}

	for_each(connections, j) {
		if (!FD_ISSET(j->first, &readfds)) continue;
		if (!ph.process(j->second)) {
			logmsg(LOG_INFO, "Connection closed\n");
			delete j->second;
			connections.erase(j);
		}
	}
}



void Listener::buildSelectList(fd_set &fds, int &max)
{
	map<int, TCPEndPoint*>::const_iterator i;
	map<int, NetworkEndPoint*>::const_iterator j;

	max = 0;
	FD_ZERO(&fds);

	for_each(listeners, i) {
		if (i->first>=0) {
			if (max<i->first) max=i->first;
			FD_SET(i->first, &fds);
		}
	}

	for_each(connections, j) {
		if (j->first>=0) {
			if (max<j->first) max=j->first;
			FD_SET(j->first, &fds);
		}
	}
}



void runListener(Listener* lr) 
{
	lr->doListen();
}





