//***************************************************************************
//* listener.h
//*
//* description    : class Listener
//*                  Listen for data and accept new connections
//* copyright      : (C) 2001 Cedric de Launois
//*                  Universite Catholique de Louvain, Belgium
//* begin          : Sept 13 2001
//* email          : cedric@delaunois.org
//***************************************************************************

/****************************************************************************
 * Copyright (c) 1998 Free Software Foundation, Inc.                        *
 *                                                                          *
 * Permission is hereby granted, free of charge, to any person obtaining a  *
 * copy of this software and associated documentation files (the            *
 * "Software"), to deal in the Software without restriction, including      *
 * without limitation the rights to use, copy, modify, merge, publish,      *
 * distribute, distribute with modifications, sublicense, and/or sell       *
 * copies of the Software, and to permit persons to whom the Software is    *
 * furnished to do so, subject to the following conditions:                 *
 *                                                                          *
 * The above copyright notice and this permission notice shall be included  *
 * in all copies or substantial portions of the Software.                   *
 *                                                                          *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS  *
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF               *
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.   *
 * IN NO EVENT SHALL THE ABOVE COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,   *
 * DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR    *
 * OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR    *
 * THE USE OR OTHER DEALINGS IN THE SOFTWARE.                               *
 *                                                                          *
 * Except as contained in this notice, the name(s) of the above copyright   *
 * holders shall not be used in advertising or otherwise to promote the     *
 * sale, use or other dealings in this Software without prior written       *
 * authorization.                                                           *
 ****************************************************************************/

class Listener;

#ifndef _LISTENER_H_
#define _LISTENER_H_

#include <pthread.h>
#include <unistd.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <string.h>
#include <errno.h>
#include <map>
#include "exceptions.h"
#include "netEndPoint.h"
#include "tcpEndPoint.h"
#include "udpEndPoint.h"
#include "protocolHdlr.h"
#include "address.h"

class Listener {

 public:

	Listener(Options* o);
	// Constructor
	// PRE : o not NULL
	// POST: listener initialized


	~Listener();
	// Destructor
	// A delete is made on all the network end points


	inline void listen(NetworkEndPoint* n) { 
		connections.insert(make_pair(n->getfd(), n)); 
	}
	// Tell the listener to receive datas coming on this end point.
	// PRE : n not NULL
	// POST: the listener is accepting data coming on this end point
	// NOTE: never do a delete on n after this call


	inline void accept(TCPEndPoint* n) { 
		listeners.insert(make_pair(n->getfd(), n)); 
	}
	// Tell the listener to listen for new connections opened
	// on this end point.
	// A call to function listen() above will be made for each new
	// connection accepted.
	// PRE : n is not null
	// POST: the listener is accepting new connection on this end point n
	// NOTE: never do a delete on n after this call


	bool start();
	// Start listening on the network end points
	// PRE : there's an end point to listen to
	// POST: returns false if there is no end point to listen to
	//                     or if the thread could not be created
	//               true otherwise


	bool stop();
	// Stop listening on the network end points
	// PRE : /
	// POST: returns false if the listener is not running
	//               true otherwise


 private:

	pthread_t listener;       // Thread listener
	Options* opts;            // Options


	map<int, TCPEndPoint*> listeners; 
	// The list of end points the listener listen to
	// table <file descriptor, NetworkEndPoint>


	map<int, NetworkEndPoint*> connections; 
	// The list of connections for which the listener is waiting data
	// table <file descriptor, NetworkEndPoint>


	void doListen();
	// Do the effective listening on the end points
	// Pool the list of end points, waiting for an event


	void dispatch(fd_set &, ProtocolHandler &ph);
	// Dispatch the incoming data or new connection
	// PRE : - one or more sockets in fd_set are ready for reading
	//       - ph is able to process the data
	// POST: the data received is transmitted to the protocol
	//       handler for processing


	void buildSelectList(fd_set &fds, int &max);
	// Build the select list from the listeners and connections lists


	friend void runListener(Listener*);   // The EP manager thread

};
	
void runListener(Listener*);

#endif
