//***************************************************************************
//* msgProcessor.cpp
//*
//* description    : class MessageProcessor
//*                  Processing of a RSIP message :
//*                    - create the reply
//*                    - modify the client state
//*                    - update the client state in the database
//* copyright      : (C) 2001 Cedric de Launois
//*                  Universite Catholique de Louvain, Belgium
//* begin          : Sept 24 2001
//* email          : cedric@delaunois.org
//***************************************************************************

/****************************************************************************
 * Copyright (c) 1998 Free Software Foundation, Inc.                        *
 *                                                                          *
 * Permission is hereby granted, free of charge, to any person obtaining a  *
 * copy of this software and associated documentation files (the            *
 * "Software"), to deal in the Software without restriction, including      *
 * without limitation the rights to use, copy, modify, merge, publish,      *
 * distribute, distribute with modifications, sublicense, and/or sell       *
 * copies of the Software, and to permit persons to whom the Software is    *
 * furnished to do so, subject to the following conditions:                 *
 *                                                                          *
 * The above copyright notice and this permission notice shall be included  *
 * in all copies or substantial portions of the Software.                   *
 *                                                                          *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS  *
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF               *
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.   *
 * IN NO EVENT SHALL THE ABOVE COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,   *
 * DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR    *
 * OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR    *
 * THE USE OR OTHER DEALINGS IN THE SOFTWARE.                               *
 *                                                                          *
 * Except as contained in this notice, the name(s) of the above copyright   *
 * holders shall not be used in advertising or otherwise to promote the     *
 * sale, use or other dealings in this Software without prior written       *
 * authorization.                                                           *
 ****************************************************************************/

#include "msgProcessor.h"

MessageProcessor::MessageProcessor(Shell* s, Proxy* p, Options* o) 
{
	shell = s;
	proxy = p;
	opts = o;
	Tunnel::setup(shell, proxy, opts);
}


MessageProcessor::~MessageProcessor() {}


int MessageProcessor::process(RSIPMessage* &msg, 
			      RSIPMessage* &reply,
			      RSIPClient* &cl) 
{

	// Check the RSIP version
	if (msg->getVersion()!=RSIP_VERSION) {
		reply = new RSIPMessage(ERROR_RESPONSE);
		reply->forgeError(UNSUPPORTED_RSIP_VERSION, 0, 0, 0, 0);
		return 1;
	}

	int status = 0;

	logmsg(LOG_NOTICE, "%s from : %s (%s)\n", 
	       msg->name(), 
	       cl->getAddress().toChar(), 
	       (cl->getProto()==UDP) ? "UDP" : "TCP");

	switch (msg->getType()) {
	case REGISTER_REQUEST:
		status = treatRegisterRequest(msg, reply, cl);
		break;
	case DEREGISTER_REQUEST:
		status = treatDeRegisterRequest(msg, reply, cl);
		break;
	case ASSIGN_REQUEST_RSAIP:
		status = treatAssignRequestRSAIP(msg, reply, cl);
		break;
	case ASSIGN_REQUEST_RSAPIP:
		status = treatAssignRequestRSAPIP(msg, reply, cl);
		break;
	case EXTEND_REQUEST:
		status = treatExtendRequest(msg, reply, cl);
		break;
	case FREE_REQUEST:
		status = treatFreeRequest(msg, reply, cl);
		break;
	case LISTEN_REQUEST:
		status = treatListenRequest(msg, reply, cl);
		break;
	case QUERY_REQUEST:
		status = treatQueryRequest(msg, reply, cl);
		break;
	default:
		status = treatUnsupportedRequest(msg, reply, cl);
		break;
	}

	return status;
}



//***************************************************************************
// PRE : msg   : the message to process, of type REGISTER_REQUEST
//       clt   : the client who sent the message
// POST: reply : the relpy to send back to the client
//       - Returns 0 if an error occured which implies a deconnection of 
//                 the client
//       - Returns 1 otherwise
//***************************************************************************
int MessageProcessor::treatRegisterRequest(RSIPMessage* &msg, 
					   RSIPMessage* &reply,
					   RSIPClient* cl) 
{
	struct RSIPMsgParams msgp;

	if ((msgp.errorType=msg->parse(msgp))!=1) {
		reply = new RSIPMessage(ERROR_RESPONSE);
		reply->forge(msgp);
		logmsg(LOG_WARNING, "%s: parsing failed. Error %d\n", msg->name(), msgp.errorType);
		return 0;
	}

	msgp.errorType = check(msgp, cl, REGISTER_REQUEST);
	if (msgp.errorType!=1) {
		reply = new RSIPMessage(ERROR_RESPONSE);
		reply->forge(msgp);
		if (msgp.errorType!=ALREADY_REGISTERED) return 0;
		else return 1;
	}

	if (RSIPClient::size() > opts->max_clients) {
		reply = new RSIPMessage(ERROR_RESPONSE);
		reply->forgeError(REGISTRATION_DENIED, 0, 0, 0, 0);
		logmsg(LOG_WARNING, "Maximum (%i) number of clients reached !\n", opts->max_clients);
		return 0;
	}

	// INV : the message is valid and the client client is not registered
	msgp.clientID = cl->doRegister();
	msgp.leaseTime = opts->initial_register_lease_time - opts->net_delay;
	msgp.tunnelType = 0;

	reply = new RSIPMessage(REGISTER_RESPONSE);
	reply->forge(msgp);

	return 1;
}



//***************************************************************************
// PRE : msg   : the message to process, of type DEREGISTER_REQUEST
//       clt   : the client who sent the message
// POST: reply : the relpy to send back to the client
//       - Returns 0 if the connection has to be closed (i.e. for UDP protocol) 
//                 or if an error occured which implies a deconnection of 
//                 the client
//       - Returns 1 otherwise
//***************************************************************************
int MessageProcessor::treatDeRegisterRequest(RSIPMessage* &msg, 
					     RSIPMessage* &reply, 
					     RSIPClient* cl) 
{
	struct RSIPMsgParams msgp;

	if ((msgp.errorType=msg->parse(msgp))!=1) {
		reply = new RSIPMessage(ERROR_RESPONSE);
		reply->forge(msgp);
		logmsg(LOG_WARNING, "%s: parsing failed. Error %d\n", msg->name(), msgp.errorType);
		return 1;
	}

	msgp.errorType = check(msgp, cl, DEREGISTER_REQUEST);
	if (msgp.errorType!=1) {
		reply = new RSIPMessage(ERROR_RESPONSE);
		reply->forge(msgp);
		return 1;
	}

	reply = new RSIPMessage(DEREGISTER_RESPONSE);
	reply->forge(msgp);
	cl->deregister(false);

	return 0;

}



//***************************************************************************
// PRE : msg   : the message to process, of type ASSIGN_REQUEST_RSAIP
//       clt   : the client who sent the message
// POST: reply : the relpy to send back to the client
//       - Returns 0 if an error occured which implies a deconnection of 
//                 the client
//       - Returns 1 otherwise
//***************************************************************************
int MessageProcessor::treatAssignRequestRSAIP(RSIPMessage* &msg,
					      RSIPMessage* &reply, 
					      RSIPClient* cl) 
{
	struct RSIPMsgParams msgp;
	msgp.leaseTime = opts->initial_lease_time;

	if ((msgp.errorType=msg->parse(msgp))!=1) {
		reply = new RSIPMessage(ERROR_RESPONSE);
		msgp.clientID = cl->getClientID();
		reply->forge(msgp);
		logmsg(LOG_WARNING, "%s: parsing failed. Error %d\n", msg->name(), msgp.errorType);
		return 1;
	}

	msgp.errorType = check(msgp, cl, ASSIGN_REQUEST_RSAIP);
	if (msgp.errorType!=1) {
		reply = new RSIPMessage(ERROR_RESPONSE);
		reply->forge(msgp);
		return 1;
	}

	Bind* b= new Bind();

	msgp.errorType = assign(cl, msgp, ASSIGN_REQUEST_RSAIP, b);
	if (msgp.errorType!=1) {
		reply = new RSIPMessage(ERROR_RESPONSE);
		reply->forge(msgp);
		delete b;
		return 1;
	}

	msgp.bindID    = b->bindID;
	msgp.leaseTime = b->getLeaseTime() - opts->net_delay;
	msgp.flags    |= FLAG_TUNNEL_EP;
	msgp.tunnelEP.addrType = ADDR_IPV4;
	msgp.tunnelEP.address = opts->int_ip;

	reply = new RSIPMessage(ASSIGN_RESPONSE_RSAIP);
	reply->forge(msgp);
	
	return 1;
}



//***************************************************************************
// Same specifications as above... but for ASSIGN_REQUEST_RSAPIP
// Not implemented, return an UNSUPPORTED_MESSAGE Error
//***************************************************************************
int MessageProcessor::treatAssignRequestRSAPIP(RSIPMessage* &msg,
					       RSIPMessage* &reply, 
					       RSIPClient* cl) 
{
	struct RSIPMsgParams msgp;
	msgp.leaseTime = opts->initial_lease_time;

	if ((msgp.errorType=msg->parse(msgp))!=1) {
		reply = new RSIPMessage(ERROR_RESPONSE);
		msgp.clientID = cl->getClientID();
		reply->forge(msgp);
		logmsg(LOG_WARNING, "%s: parsing failed. Error %d\n", msg->name(), msgp.errorType);
		return 1;
	}

	msgp.errorType = check(msgp, cl, ASSIGN_REQUEST_RSAPIP);
	if (msgp.errorType!=1) {
		reply = new RSIPMessage(ERROR_RESPONSE);
		reply->forge(msgp);
		return 1;
	}

	Bind* b= new Bind();

	msgp.errorType = assign(cl, msgp, ASSIGN_REQUEST_RSAPIP, b);

	if (msgp.errorType!=1) {
		reply = new RSIPMessage(ERROR_RESPONSE);
		reply->forge(msgp);
		delete b;
		return 1;
	}

	msgp.bindID    = b->bindID;
	msgp.leaseTime = b->getLeaseTime() - opts->net_delay;
	msgp.flags    |= FLAG_TUNNEL_EP;
	msgp.tunnelEP.addrType = ADDR_IPV4;
	msgp.tunnelEP.address = opts->int_ip;

	reply = new RSIPMessage(ASSIGN_RESPONSE_RSAPIP);
	reply->forge(msgp);
	
	return 1;
}



//***************************************************************************
// Same specifications... but for EXTEND_REQUEST
//***************************************************************************
int MessageProcessor::treatExtendRequest(RSIPMessage* &msg, 
					 RSIPMessage* &reply, 
					 RSIPClient* cl) 
{
	struct RSIPMsgParams msgp;

	if ((msgp.errorType=msg->parse(msgp))!=1) { // An error occured when parsing
		reply = new RSIPMessage(ERROR_RESPONSE);
		reply->forge(msgp);
		logmsg(LOG_WARNING, "%s: parsing failed. Error %d\n", msg->name(), msgp.errorType);
		return 1;
	}

	msgp.errorType = check(msgp, cl, EXTEND_REQUEST);
	if (msgp.errorType!=1) {
		reply = new RSIPMessage(ERROR_RESPONSE);
		reply->forge(msgp);
		return 1;
	}

	logmsg(LOG_INFO, "Host %s: lease extended for %dsec\n", cl->getAddress().toChar(), msgp.leaseTime);

	msgp.leaseTime = cl->extend(msgp.bindID, msgp.leaseTime) - opts->net_delay;

	reply = new RSIPMessage(EXTEND_RESPONSE);
	reply->forge(msgp);

	return 1;
}



//***************************************************************************
// Same specifications... but for FREE_REQUEST
//***************************************************************************
int MessageProcessor::treatFreeRequest(RSIPMessage* &msg, 
				       RSIPMessage* &reply, 
				       RSIPClient* cl) 
{
	struct RSIPMsgParams msgp;

	if ((msgp.errorType=msg->parse(msgp))!=1) { // An error occured when parsing
		reply = new RSIPMessage(ERROR_RESPONSE);
		reply->forge(msgp);
		logmsg(LOG_WARNING, "%s: parsing failed. Error %d\n", msg->name(), msgp.errorType);
		return 1;
	}

	msgp.errorType = check(msgp, cl, FREE_REQUEST);
	if (msgp.errorType!=1) {
		reply = new RSIPMessage(ERROR_RESPONSE);
		reply->forge(msgp);
		return 1;
	}

	cl->free(msgp.bindID, false);

	reply = new RSIPMessage(FREE_RESPONSE);
	reply->forge(msgp);

	return 1;
}



//***************************************************************************
// Same specifications... but for LISTEN_REQUEST
//***************************************************************************
int MessageProcessor::treatListenRequest(RSIPMessage* &msg,
					 RSIPMessage* &reply, 
					 RSIPClient* cl) 
{
	struct RSIPMsgParams msgp;
	msgp.leaseTime = opts->initial_lease_time;

	if ((msgp.errorType=msg->parse(msgp))!=1) { // An error occured when parsing
		reply = new RSIPMessage(ERROR_RESPONSE);
		msgp.clientID = cl->getClientID();
		reply->forge(msgp);
		logmsg(LOG_WARNING, "%s: parsing failed. Error %d\n", msg->name(), msgp.errorType);
		return 1;
	}

	msgp.errorType = check(msgp, cl, LISTEN_REQUEST);
	if (msgp.errorType!=1) {
		reply = new RSIPMessage(ERROR_RESPONSE);
		reply->forge(msgp);
		return 1;
	}

	Bind* b= new Bind();

	msgp.errorType = assign(cl, msgp, LISTEN_REQUEST, b);

	if (msgp.errorType!=1) {
		reply = new RSIPMessage(ERROR_RESPONSE);
		reply->forge(msgp);
		delete b;
		return 1;
	}

	msgp.bindID    = b->bindID;
	msgp.leaseTime = b->getLeaseTime() - opts->net_delay;
	msgp.flags    |= FLAG_TUNNEL_EP;
	msgp.tunnelEP.addrType = ADDR_IPV4;
	msgp.tunnelEP.address  = opts->int_ip;

	reply = new RSIPMessage(LISTEN_RESPONSE);
	reply->forge(msgp);

	return 1;
}



//***************************************************************************
// Same specifications... but for QUERY_REQUEST
//***************************************************************************
int MessageProcessor::treatQueryRequest(RSIPMessage* &msg,
					RSIPMessage* &reply, 
					RSIPClient* cl) 
{
	struct RSIPMsgParams msgp;

	if ((msgp.errorType=msg->parse(msgp))!=1) { // An error occured when parsing
		reply = new RSIPMessage(ERROR_RESPONSE);
		reply->forge(msgp);
		logmsg(LOG_WARNING, "%s: parsing failed. Error %d\n", msg->name(), msgp.errorType);
		return 1;
	}

	if (!(msgp.flags & FLAG_COUNTER) && cl->getProto()==UDP) {
		reply = new RSIPMessage(ERROR_RESPONSE);
		msgp.errorType = MESSAGE_COUNTER_REQUIRED;
		reply->forge(msgp);
		logmsg(LOG_WARNING, "%s: counter required\n", msg->name());
		return 1;
	}

	msgp.errorType = check(msgp, cl, QUERY_REQUEST);
	if (msgp.errorType!=1) {
		reply = new RSIPMessage(ERROR_RESPONSE);
		reply->forge(msgp);
		return 1;
	}

	// We reply with a "don't know"
	reply = new RSIPMessage(QUERY_RESPONSE);
	reply->forge(msgp);

	return 1;
}



//***************************************************************************
// Same specifications... but for an UNSUPPORTED REQUEST
// Return an UNSUPPORTED_MESSAGE Error
//***************************************************************************
int MessageProcessor::treatUnsupportedRequest(RSIPMessage* &msg, RSIPMessage* &reply, RSIPClient* cl) 
{
	struct RSIPMsgParams msgp;

	if ((msgp.errorType=msg->parse(msgp))!=1) { // An error occured when parsing
		reply = new RSIPMessage(ERROR_RESPONSE);
		reply->forge(msgp);
		logmsg(LOG_WARNING, "%s: parsing failed. Error %d\n", msg->name(), msgp.errorType);
		return 1;
	}

	if (!(msgp.flags & FLAG_COUNTER) && cl->getProto()==UDP) {
		reply = new RSIPMessage(ERROR_RESPONSE);
		msgp.errorType = MESSAGE_COUNTER_REQUIRED;
		reply->forge(msgp);
		logmsg(LOG_WARNING, "%s: counter required\n", msg->name());
		return 1;
	}

	reply = new RSIPMessage(ERROR_RESPONSE);
	msgp.errorType = UNSUPPORTED_MESSAGE;
	reply->forge(msgp);

	return 1;
}



//***************************************************************************
// Check the state, the client ID and the bind ID of the client
// PRE : /
// POST: Returns an error number > 0 if one of these checked parameters is not valid
//       else return 0.
// NOTE: the error is the RFC3103 error number
//***************************************************************************
uint16_t MessageProcessor::check(struct RSIPMsgParams& msgp, RSIPClient* cl, uint8_t msgType)
{
	unsigned int flags = 0;

	if ((cl->getProto()==UDP) && !(msgp.flags & FLAG_COUNTER)) {
		logmsg(LOG_WARNING, "%s: Counter required\n", msgNames[msgType]);
		msgp.flags = flags;
		return MESSAGE_COUNTER_REQUIRED;
	}

	flags |= (msgp.flags & FLAG_COUNTER);

	// Checking State
	switch(msgType) {

	case REGISTER_REQUEST:
		if (cl->getState()!=STATE_UNREGISTERED) {
			msgp.clientID = cl->getClientID();
			flags |= FLAG_CLIENT_ID;
			logmsg(LOG_WARNING, "%s: Already registered\n", msgNames[msgType]);
			msgp.flags = flags;
			return ALREADY_REGISTERED;
		}
		return 1;

	case DEREGISTER_REQUEST:
		if (cl->getState() == STATE_UNREGISTERED) {
			logmsg(LOG_WARNING, "%s: Already unregistered\n", msgNames[msgType]);
			msgp.flags = flags;
			return ALREADY_UNREGISTERED;
		}
		break;

	default:
		if (cl->getState() == STATE_UNREGISTERED) {
			logmsg(LOG_WARNING, "%s: Register first\n", msgNames[msgType]);
			msgp.flags = flags;
			return REGISTER_FIRST;
		}
	}

	// Checking client ID
	flags |= FLAG_CLIENT_ID;
	if (msgp.clientID!=cl->getClientID()) {
		msgp.clientID = cl->getClientID();
		logmsg(LOG_WARNING, "%s: Bad client ID\n", msgNames[msgType]);
		msgp.flags = flags;
		return BAD_CLIENT_ID;
	} 
	msgp.clientID = cl->getClientID();

	// Checking bind ID
	switch(msgType) {
	case EXTEND_REQUEST:
	case FREE_REQUEST:
		if ((cl->getState() == STATE_REGISTERED) || !cl->checkBind(msgp.bindID)) {
			logmsg(LOG_WARNING, "%s: Bad bind ID\n", msgNames[msgType]);
			msgp.flags = flags;
			return BAD_BIND_ID;
		}
		flags |= FLAG_BIND_ID;
		break;
	default: 
		break;
	}

	// Checking tunnel type
	switch(msgType) {
	case ASSIGN_REQUEST_RSAIP:
	case ASSIGN_REQUEST_RSAPIP:
	case LISTEN_REQUEST:
		if (msgp.flags & FLAG_TUNNEL_TYPE) {

			switch(msgp.tunnelType) {
			case TFLAG_IP_IP:
				msgp.tunnelType = TUNNEL_IP_IP;
				break;
			case TFLAG_GRE:
				msgp.tunnelType = TUNNEL_GRE;
				break;
			default:
				logmsg(LOG_WARNING, "%s: Unsupported tunnel type\n", msgNames[msgType]);
				msgp.flags = flags;
				return BAD_TUNNEL_TYPE;
			}

		} else msgp.tunnelType = TUNNEL_IP_IP;
		break;

	default: 
		break;
	}

	// Checking lease time && availability of resources
	switch(msgType) {
	case ASSIGN_REQUEST_RSAIP:
	case ASSIGN_REQUEST_RSAPIP:
	case EXTEND_REQUEST:
	case LISTEN_REQUEST:
		if (msgp.leaseTime == 0) msgp.leaseTime = opts->initial_lease_time;
		else if (msgp.leaseTime > opts->max_lease_time) msgp.leaseTime = opts->max_lease_time;
		else if (msgp.leaseTime < opts->min_lease_time) msgp.leaseTime = opts->min_lease_time;
		break;
	default: 
		break;
	}

	// Checking limit
	switch(msgType) {
	case ASSIGN_REQUEST_RSAIP:
		// Checking the max number allowed of addresses per client
		if ((cl->nbIPs() >= 1) || (opts->rsa_method==0)) {
			logmsg(LOG_WARNING, "%s: Unallowed request\n", msgNames[msgType]);
			msgp.flags = flags;
			return LOCAL_ADDR_UNALLOWED;
		}
		break;
	case ASSIGN_REQUEST_RSAPIP:
	case LISTEN_REQUEST:
		// Checking the max number allowed of ports per client
		if (cl->nbPorts() + msgp.ports1.number > opts->max_host_ports) {
			logmsg(LOG_WARNING, "%s: Unallowed request\n", msgNames[msgType]);
			msgp.flags = flags;
			return LOCAL_ADDR_UNALLOWED;
		}
	default: 
		break;
	}

	return 1;
}


//***************************************************************************
// Assign a bind to a client and create the tunnel if needed.
// PRE : - b is not NULL
//       - msgType is ASSIGN_REQUEST_RSAIP, ASSIGN_REQUEST_RSAPIP 
//         or LISTEN_REQUEST
// POST: - if success, then the resources specified have been allocated to 
//         the client, and RETURNS 1 and
//         localPorts and addrLocal have been modified to reflect the
//         real assignement
//       - otherwise, nothing is done and RETURNS RFC3103 ERROR_CODE
//***************************************************************************
uint16_t MessageProcessor::assign(RSIPClient* cl,
				  struct RSIPMsgParams& msgp,
				  uint16_t msgType,
				  Bind* &b)
{
	bool aDC = false;
	bool pDC = true;
	bool listen = (msgType==LISTEN_REQUEST);
	int err, availability = 0;
	char idtnl[20];
	deque<uint16_t>* localPorts = &msgp.ports1.ports;
	unsigned char nbLocalPorts = msgp.ports1.number;
	Address from, realFrom, to, realTo;
	Resource r;

	if (!b) return INTERNAL_SERVER_ERROR;
	b->setLeaseTime(msgp.leaseTime);

	if (!msgp.address1.address.isDefined()) aDC = true; 


	// God ! all this stuff is ugly... Need to be reworked.

	// 1: getting resources from the pool
	if (msgType==ASSIGN_REQUEST_RSAIP) {
		// RSA-IP
		r = Resource(msgp.address1.address, -1);
		availability = b->addResource(r, true, true, 0, false);

	} else {
		// RSAP-IP or LISTEN
		if (nbLocalPorts==0) return BAD_PARAM;
		r = Resource(msgp.address1.address, 0);

		if (nbLocalPorts == localPorts->size()) {
			// Request for non-contiguous ports, ex: 1720,5000,5001,5002
			availability = b->addResource(r.address, *localPorts, aDC, listen);
		
		} else {
			// Request for contiguous ports, ex: 8446x255 
			// r.port is the first port of the list of contiguous ports
			switch (localPorts->size()) {
			case 1: 
				// Request for a list of requested contiguous ports
				pDC=false;
				r.port=localPorts->front();
				availability = b->addResource(r, aDC, pDC, nbLocalPorts, listen);
				break;

			case 0:
				// Request for a list of undefined contiguous ports
				availability = 0;

				if (listen) {
					pDC=true;
					availability = b->addResource(r, aDC, pDC, nbLocalPorts, listen);
					localPorts->push_back(r.port);
					break;
				}

				if (cl->getNextReservedPort()==0) {
					// New client
					// We search a free reserved range for the client
					pDC=false;
					r.port=opts->rsap_low_port;
					while (r.port+nbLocalPorts <= opts->rsap_high_port) {
						availability = b->addResource(r, aDC, pDC, nbLocalPorts, listen);
						if (availability>0) break;
						r.port += opts->max_host_ports;
					}

				} else {
					// Old client, we try its next reserved ports
					pDC=false;
					r.port=cl->getNextReservedPort();
					availability = b->addResource(r, aDC, pDC, nbLocalPorts, listen);
				}

				// Trying the hard way if failure
				if (availability<=0) {
					pDC=true;
					availability = b->addResource(r, aDC, pDC, nbLocalPorts, listen);
				}

				// Updating the list of assigned ports
				localPorts->push_back(r.port);
				if (availability>0) cl->setNextReservedPort(localPorts->front()+nbLocalPorts);

				break;

			default:
				// Bad param !
				logmsg(LOG_WARNING, "%s: Bad port parameter\n", msgNames[msgType]);
				return BAD_PARAM;
			}
		}

	}


	// 2: Setting the bind for the client
	switch (availability) {
	case 0:
		logmsg(LOG_WARNING, "%s: Resource in use\n", msgNames[msgType]);
		if (msgType==ASSIGN_REQUEST_RSAIP) return LOCAL_ADDR_INUSE;
		else return LOCAL_ADDRPORT_INUSE;

	case -1:
	case -2:
		logmsg(LOG_WARNING, "%s: Resource not allowed\n", msgNames[msgType]);
		if (msgType==ASSIGN_REQUEST_RSAIP) return LOCAL_ADDR_UNALLOWED;
		else return LOCAL_ADDRPORT_UNALLOWED;

	default:
		// 3: Setting up the tunnel
		getIP(opts->int_if, realFrom);
		getIP(opts->ext_if, from);
		realTo = cl->getAddress();  // Real destination address
		to = r.address;             // Tunnel end point address

		b->tunnel = new Tunnel(realFrom, realTo, from, to, msgp.tunnelType);
		if (msgType!=ASSIGN_REQUEST_RSAIP)
			b->tunnel->setPorts(nbLocalPorts, *localPorts);

		cl->assign(b);
		snprintf(idtnl, 20, "%ix%i", cl->getClientID(), b->bindID);
		string name = "rsip" + string(idtnl);     // The tunnel name

		if ((err=b->tunnel->create(name))!=1) {
			cl->free(b->bindID, false); // this also delete b
			b=new Bind(); // We return an empty bind
			logmsg(LOG_ERR, "%s: Can't establish tunnel. Error %d\n",msgNames[msgType] , err);
			return INTERNAL_SERVER_ERROR;
		}

		msgp.address1.address = r.address;
	}

	return 1;
}
