//***************************************************************************
//* msgProcessor.h
//*
//* description    : class MessageProcessor
//*                  Processing of a RSIP message :
//*                    - create the reply
//*                    - modify the client state
//*                    - update the client state in the database
//* copyright      : (C) 2001 Cedric de Launois
//*                  Universite Catholique de Louvain, Belgium
//* begin          : Sept 24 2001
//* email          : cedric@delaunois.org
//***************************************************************************

/****************************************************************************
 * Copyright (c) 1998 Free Software Foundation, Inc.                        *
 *                                                                          *
 * Permission is hereby granted, free of charge, to any person obtaining a  *
 * copy of this software and associated documentation files (the            *
 * "Software"), to deal in the Software without restriction, including      *
 * without limitation the rights to use, copy, modify, merge, publish,      *
 * distribute, distribute with modifications, sublicense, and/or sell       *
 * copies of the Software, and to permit persons to whom the Software is    *
 * furnished to do so, subject to the following conditions:                 *
 *                                                                          *
 * The above copyright notice and this permission notice shall be included  *
 * in all copies or substantial portions of the Software.                   *
 *                                                                          *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS  *
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF               *
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.   *
 * IN NO EVENT SHALL THE ABOVE COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,   *
 * DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR    *
 * OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR    *
 * THE USE OR OTHER DEALINGS IN THE SOFTWARE.                               *
 *                                                                          *
 * Except as contained in this notice, the name(s) of the above copyright   *
 * holders shall not be used in advertising or otherwise to promote the     *
 * sale, use or other dealings in this Software without prior written       *
 * authorization.                                                           *
 ****************************************************************************/

#ifndef _MESSAGE_PROCESSOR_H_
#define _MESSAGE_PROCESSOR_H_

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <pthread.h>
#include <time.h>
#include <stdio.h>
#include <string>
#include <deque>
#include "common.h"
#include "rsipMessage.h"
#include "rsipClient.h"
#include "resources.h"
#include "tunnel.h"
#include "netutils.h"
#include "address.h"
#include "shell.h"
#include "proxy.h"

class MessageProcessor {

 public:

	MessageProcessor(Shell* sh, Proxy* p, Options* o);
	// Constructor
	// Initialize a processing module


	~MessageProcessor();
	// Destructor


	int process(RSIPMessage* &msg, RSIPMessage* &reply, RSIPClient* &cl);
	// Processing of a RSIP message
	// PRE : msg is the message to process
	//       nep is the network end point from which the message is coming from
	// POST: reply is the reply to send back to the client
	//       The client state has been updated
	//       - Returns 0 if the client made a DEREGISTER_REQUEST
	//                 or if an error occured which implies a deconnection of 
	//                 the client
	//       - Returns 1 otherwise


 private:

	Shell* shell;  // To build tunnels etc.
	Proxy* proxy;  // To route RSIP traffic
	Options* opts; // Options

	int treatRegisterRequest(RSIPMessage* &msg, RSIPMessage* &reply, RSIPClient* clt);
	// PRE : msg   : the message to process, of type REGISTER_REQUEST
	//       clt   : the client who sent the message
	// POST: reply : the relpy to send back to the client
	//       - Returns 0 if an error occured which implies a deconnection of 
	//                 the client
	//       - Returns 1 otherwise


	int treatDeRegisterRequest(RSIPMessage* &msg, RSIPMessage* &reply, RSIPClient* clt);
	// PRE : msg   : the message to process, of type DEREGISTER_REQUEST
	//       clt   : the client who sent the message
	// POST: reply : the relpy to send back to the client
	//       - Returns 0 if the connection has to be closed (i.e. for UDP protocol) 
	//                 or if an error occured which implies a deconnection of 
	//                 the client
	//       - Returns 1 otherwise


	int treatAssignRequestRSAIP(RSIPMessage* &msg, RSIPMessage* &reply, RSIPClient* clt);
	// PRE : msg   : the message to process, of type ASSIGN_REQUEST_RSAIP
	//       clt   : the client who sent the message
	// POST: reply : the relpy to send back to the client
	//       - Returns 0 if an error occured which implies a deconnection of 
	//                 the client
	//       - Returns 1 otherwise


	int treatAssignRequestRSAPIP(RSIPMessage* &msg, RSIPMessage* &reply, RSIPClient* clt);
	// Same specifications... but for ASSIGN_REQUEST_RSAPIP


	int treatExtendRequest(RSIPMessage* &msg, RSIPMessage* &reply, RSIPClient* clt);
	// Same specifications... but for EXTEND_REQUEST


	int treatFreeRequest(RSIPMessage* &msg, RSIPMessage* &reply, RSIPClient* clt);
	// Same specifications... but for FREE_REQUEST


	int treatListenRequest(RSIPMessage* &msg, RSIPMessage* &reply, RSIPClient* clt);
	// Same specifications... but for LISTEN_REQUEST


	int treatQueryRequest(RSIPMessage* &msg, RSIPMessage* &reply, RSIPClient* clt);
	// Same specifications... but for QUERY_REQUEST
	// Not implemented, return an UNSUPPORTED_MESSAGE Error


	int treatUnsupportedRequest(RSIPMessage* &msg, RSIPMessage* &reply, RSIPClient* clt);
	// Same specifications... but for an UNSUPPORTED REQUEST
	// Return an UNSUPPORTED_MESSAGE Error


	uint16_t check(struct RSIPMsgParams& msgp, RSIPClient* context, uint8_t msgType);
	// Check the state, the client ID and the bind ID of the client
	// PRE : /
	// POST: Returns an error number > 0 if one of these checked parameters is not valid
	//       else return 0.
	// NOTE: the error is the RFC3103 error number


	uint16_t assign(RSIPClient* cl,
			struct RSIPMsgParams& msgp,
			uint16_t msgType,
			Bind* &b);
	// Assign a bind to a client and create the tunnel if needed.
	// PRE : - b is not NULL
	//       - msgType is ASSIGN_REQUEST_RSAIP, ASSIGN_REQUEST_RSAPIP 
	//         or LISTEN_REQUEST
	// POST: - if success, then the resources specified have been allocated to 
	//         the client, and RETURNS 1 and
	//         localPorts and addrLocal have been modified to reflect the
	//         real assignement
	//       - otherwise, nothing is done and RETURNS RFC3103 ERROR_CODE

};

#endif
