//**************************************************************************
//* netEndPoint.h 
//*
//* description    : Virtual class netEndPoint. 
//*                  Modelize an end point of  a TCP or UDP connection
//* copyright      : (C) 2001 Cedric de Launois
//*                  Universite Catholique de Louvain, Belgium
//* begin          : Sept 11 2001
//* email          : cedric@delaunois.org
//* compilation flags : -lnsl -lsocket
//**************************************************************************

/****************************************************************************
 * Copyright (c) 1998 Free Software Foundation, Inc.                        *
 *                                                                          *
 * Permission is hereby granted, free of charge, to any person obtaining a  *
 * copy of this software and associated documentation files (the            *
 * "Software"), to deal in the Software without restriction, including      *
 * without limitation the rights to use, copy, modify, merge, publish,      *
 * distribute, distribute with modifications, sublicense, and/or sell       *
 * copies of the Software, and to permit persons to whom the Software is    *
 * furnished to do so, subject to the following conditions:                 *
 *                                                                          *
 * The above copyright notice and this permission notice shall be included  *
 * in all copies or substantial portions of the Software.                   *
 *                                                                          *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS  *
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF               *
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.   *
 * IN NO EVENT SHALL THE ABOVE COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,   *
 * DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR    *
 * OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR    *
 * THE USE OR OTHER DEALINGS IN THE SOFTWARE.                               *
 *                                                                          *
 * Except as contained in this notice, the name(s) of the above copyright   *
 * holders shall not be used in advertising or otherwise to promote the     *
 * sale, use or other dealings in this Software without prior written       *
 * authorization.                                                           *
 ****************************************************************************/

class NetworkEndPoint;

#ifndef _NETWORK_ENDPOINT_H_
#define _NETWORK_ENDPOINT_H_

#include <netinet/in.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <arpa/inet.h>
#include <unistd.h>
#include "proto.h"
#include "address.h"

class NetworkEndPoint {  // Abstract class
 public:

	virtual void openPort(const char* iface, unsigned short port, const Address *)=0;
	// SPEC: Opens the communication port
	// PRE : - iface : the name of an interface or NULL
	//       - port : valid communication port
	//       - address : optional :
	//          - if address is NULL, then opens the port in listen mode
	//            (for a server for example = port of type 'server')
	//          - otherwise 'address' must be a valid IPv4 address and then
	//            opens a connection towards the server with this address.
	// POST: - Port initialized as described
	//       - if iface!=NULL, packets will be accepted from iface only
	// EXCEPTIONS : SocketOpeningError, BindError, ListenError, BadPort, ConnectError


	virtual int readn(char* ptr, int nbytes, int timeout, Address& from)=0;
	// SPEC: reads n bytes on the listening port
	// PRE : - port opened
	//       - nbytes >= 0 : the number of characters to read
	//       - timeout : the time out in MICROseconds, < 1 000 000
	//                 -1 for an infinite time out
	// POST: - ptr : the chain read
	//       - from : IP adress of the sender
	//       Returns the number of characters read OR
	//               -1 if time out.
	// EXCEPTIONS : SelectError, ClosedPortError, SocketReadError


	virtual int writen(char* ptr, int nbytes, const Address& to)=0;
	// SPEC: Sends n bytes to the client
	// PRE : - port opened
	//       - nbytes >= 0 : the number of characters to send
	//       - ptr : the characters to send
	//       - to : the client to which the bytes have to be sent
	// POST: return < 0 if error,
	//       otherwise returns the number of characters effectively sent
	// EXCEPTION  : SocketWriteError, ClosedPortError


	virtual int disconnect()=0;
	// PRE : /
	// POST: communication port closed
	//       returns 0 if error
	//               1 otherwise


	virtual int getfd() const=0;
	// PRE : /
	// POST: - returns -1 if the port is not opened
	//       - otherwise returns the file descriptor of the socket


	virtual unsigned short getPort() const=0;
	// SPEC: returns the port used
	// PRE : /
	// POST: the port used OR
	//       0 if the port is not opened


	virtual proto getProto() const=0;
	// PRE : /
	// POST: returns the protocol used by this end point
	//       returns UDP or TCP

};

#endif
