//***************************************************************************
//* protocol.h 
//*
//* description    : constants used for RSIP protocol
//*                  See RFC 3103, APPENDIX A
//*
//* copyright      : (C) 2001 Cedric de Launois
//*                  Universite Catholique de Louvain, Belgium
//* begin          : Sept 24 2001
//* email          : cedric@delaunois.org
//***************************************************************************

/****************************************************************************
 * Copyright (c) 1998 Free Software Foundation, Inc.                        *
 *                                                                          *
 * Permission is hereby granted, free of charge, to any person obtaining a  *
 * copy of this software and associated documentation files (the            *
 * "Software"), to deal in the Software without restriction, including      *
 * without limitation the rights to use, copy, modify, merge, publish,      *
 * distribute, distribute with modifications, sublicense, and/or sell       *
 * copies of the Software, and to permit persons to whom the Software is    *
 * furnished to do so, subject to the following conditions:                 *
 *                                                                          *
 * The above copyright notice and this permission notice shall be included  *
 * in all copies or substantial portions of the Software.                   *
 *                                                                          *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS  *
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF               *
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.   *
 * IN NO EVENT SHALL THE ABOVE COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,   *
 * DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR    *
 * OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR    *
 * THE USE OR OTHER DEALINGS IN THE SOFTWARE.                               *
 *                                                                          *
 * Except as contained in this notice, the name(s) of the above copyright   *
 * holders shall not be used in advertising or otherwise to promote the     *
 * sale, use or other dealings in this Software without prior written       *
 * authorization.                                                           *
 ****************************************************************************/

#ifndef _PROTOCOL_H_
#define _PROTOCOL_H_
#include <deque>
#include "address.h"

#define RSIP_VERSION                    1

// ************************ Parameter types *******************
#define PARAM_ADDRESS                   1
#define PARAM_PORT                      2
#define PARAM_LEASE_TIME                3
#define PARAM_CLIENT_ID                 4
#define PARAM_BIND_ID                   5
#define PARAM_TUNNEL_TYPE               6
#define PARAM_RSIP_METHOD               7
#define PARAM_ERROR                     8
#define PARAM_FLOW_POLICY               9
#define PARAM_INDICATOR                10
#define PARAM_MESSAGE_COUNTER          11
#define PARAM_VENDOR_SPECIFIC          12
#define PARAM_MAX_NUMBER               12 // For now

static const char* paramNames[] = {
	"UNKNOWN",
	"PARAM_ADDRESS",
	"PARAM_PORT",
	"PARAM_LEASE_TIME",
	"PARAM_CLIENT_ID",
	"PARAM_BIND_ID",
	"PARAM_TUNNEL_TYPE",
	"PARAM_RSIP_METHOD",
	"PARAM_ERROR",
	"PARAM_FLOW_POLICY",
	"PARAM_INDICATOR",
	"PARAM_MESSAGE_COUNTER",
	"PARAM_VENDOR_SPECIFIC"
};

// ************************ Message types *********************
#define ERROR_RESPONSE                  1
#define REGISTER_REQUEST                2
#define REGISTER_RESPONSE               3
#define DEREGISTER_REQUEST              4
#define DEREGISTER_RESPONSE             5
#define ASSIGN_REQUEST_RSAIP            6
#define ASSIGN_RESPONSE_RSAIP           7
#define ASSIGN_REQUEST_RSAPIP           8
#define ASSIGN_RESPONSE_RSAPIP          9
#define EXTEND_REQUEST                 10
#define EXTEND_RESPONSE                11
#define FREE_REQUEST                   12
#define FREE_RESPONSE                  13
#define QUERY_REQUEST                  14
#define QUERY_RESPONSE                 15
#define LISTEN_REQUEST                 16
#define LISTEN_RESPONSE                17
#define MESSAGE_MAX_NUMBER             17 // For now

static const char* msgNames[] = {
	"UNKNOWN",
	"ERROR_RESPONSE",
	"REGISTER_REQUEST",
	"REGISTER_RESPONSE",
	"DEREGISTER_REQUEST",
	"DEREGISTER_RESPONSE",
	"ASSIGN_REQUEST_RSAIP",
	"ASSIGN_RESPONSE_RSAIP",
	"ASSIGN_REQUEST_RSAPIP",
	"ASSIGN_RESPONSE_RSAPIP",
	"EXTEND_REQUEST",
	"EXTEND_RESPONSE",
	"FREE_REQUEST",
	"FREE_RESPONSE",
	"QUERY_REQUEST",
	"QUERY_RESPONSE",
	"LISTEN_REQUEST",
	"LISTEN_RESPONSE"
};


// ************************ Error types ***********************
// 100 : General errors
#define UNKNOWN_ERROR                 101
#define USE_TCP                       102
#define FLOW_POLICY_VIOLATION         103
#define INTERNAL_SERVER_ERROR         104
#define MESSAGE_COUNTER_REQUIRED      105
#define UNSUPPORTED_RSIP_VERSION      106
#define GEN_ERROR_MAX_NUMBER          106

static const char* genErrorNames[] = {
	"UNKNOWN",
	"UNKNOWN_ERROR",
	"USE_TCP",
	"FLOW_POLICY_VIOLATION",
	"INTERNAL_SERVER_ERROR",
	"MESSAGE_COUNTER_REQUIRED",
	"UNSUPPORTED_RSIP_VERSION"
};

// 200 : Parameter and message errors
#define MISSING_PARAM                 201
#define DUPLICATE_PARAM               202
#define EXTRA_PARAM                   203
#define ILLEGAL_PARAM                 204
#define BAD_PARAM                     205
#define ILLEGAL_MESSAGE               206
#define BAD_MESSAGE                   207
#define UNSUPPORTED_MESSAGE           208
#define PM_ERROR_MAX_NUMBER           208

static const char* pmErrorNames[] = {
	"UNKNOWN",
	"MISSING_PARAM",
	"DUPLICATE_PARAM",
	"EXTRA_PARAM",
	"ILLEGAL_PARAM",
	"BAD_PARAM",
	"ILLEGAL_MESSAGE",
	"BAD_MESSAGE",
	"UNSUPPORTED_MESSAGE"
};

// 300 : Permission, resource and policy errors
#define REGISTER_FIRST                301
#define ALREADY_REGISTERED            302
#define ALREADY_UNREGISTERED          303
#define REGISTRATION_DENIED           304
#define BAD_CLIENT_ID                 305
#define BAD_BIND_ID                   306
#define BAD_TUNNEL_TYPE               307
#define LOCAL_ADDR_UNAVAILABLE        308
#define LOCAL_ADDRPORT_UNAVAILABLE    309
#define LOCAL_ADDR_INUSE              310
#define LOCAL_ADDRPORT_INUSE          311
#define LOCAL_ADDR_UNALLOWED          312
#define LOCAL_ADDRPORT_UNALLOWED      313
#define REMOTE_ADDR_UNALLOWED         314
#define REMOTE_ADDRPORT_UNALLOWED     315
#define PRP_ERROR_MAX_NUMBER          315

static const char* prpErrorNames[] = {
	"UNKNOWN",
	"REGISTER_FIRST",
	"ALREADY_REGISTERED",
	"ALREADY_UNREGISTERED",
	"REGISTRATION_DENIED",
	"BAD_CLIENT_ID",
	"BAD_BIND_ID",
	"BAD_TUNNEL_TYPE",
	"LOCAL_ADDR_UNAVAILABLE",
	"LOCAL_ADDRPORT_UNAVAILABLE",
	"LOCAL_ADDR_INUSE",
	"LOCAL_ADDRPORT_INUSE",
	"LOCAL_ADDR_UNALLOWED",
	"LOCAL_ADDRPORT_UNALLOWED",
	"REMOTE_ADDR_UNALLOWED",
	"REMOTE_ADDRPORT_UNALLOWED"
};

// ***************************** Other definitions ************
#define TUNNEL_IP_IP                    1
#define TUNNEL_GRE                      2
#define TUNNEL_L2TP                     3
#define METHOD_RSA_IP                   1
#define METHOD_RSAP_IP                  2
#define LOCAL_POLICY_MACRO              1
#define LOCAL_POLICY_MICRO              2
#define REMOTE_POLICY_MACRO             1
#define REMOTE_POLICY_MICRO             2
#define REMOTE_NO_POLICY                3
#define ADDR_IPV4                       1
#define ADDR_IPV4_NETMASK               2
#define ADDR_IPV6                       3
#define ADDR_FQDN                       4

// *********************** Message description ****************
// RSIP paquet structure
#define MSG_HEADER_LEN 4               // Size of the header
struct RSIPMsgHeader {
	unsigned int ver:8;            // RSIP version
	unsigned int type:8;           // Message type
	unsigned int length:16;        // Overall length (in bytes)
};

// RSIP parameter header structure
#define PARAM_HEADER_LEN 3             // Size of the header
struct RSIPParamHeader {
	unsigned int type:8;           // Parameter type
	unsigned int length:16;        // Parameter length (in bytes)
};

// RSIP parameter structure
struct RSIPParam {
	struct RSIPParamHeader header; // Header
	uint8_t* value;                // Value of the du parameter
};

// RSIP port parameter structure
struct RSIPPortParam {
	uint8_t number;                // Number of ports
	deque<uint16_t> ports;         // List of ports
};

// RSIP address parameter structure
struct RSIPAddressParam {
	uint8_t addrType;              // Address Type
	Address address;               // The address
};

// VENDOR_SPECIFIC parameter header structure
#define PARAM_VS_HEADER_LEN 6          // Size of the header
struct RSIPVsidHeader {
	unsigned int vid:16;           // Vendor ID
	unsigned int type:16;          // Vendor specific parameter
	unsigned int length:16;        // Data length
};

// VENDOR_SPECIFIC parameter structure
struct RSIPVsid {
	struct RSIPVsidHeader header;  // Header
	uint8_t* data;                 // Data
};

// Common parameters that may be present in a client RSIP message
struct RSIPMsgParams {
	uint32_t counter;
	uint32_t leaseTime;
	uint32_t clientID;
	uint32_t bindID;
	struct RSIPAddressParam address1;
	struct RSIPPortParam    ports1;
	struct RSIPAddressParam address2;
	struct RSIPPortParam    ports2;
	uint8_t tunnelType;
	struct RSIPAddressParam tunnelEP;
	uint16_t errorType;
	unsigned int flags;
};

#define FLAG_UNDEFINED          0
#define FLAG_NONE               0
#define FLAG_COUNTER            1
#define FLAG_LEASE_TIME         2
#define FLAG_CLIENT_ID          4
#define FLAG_BIND_ID            8
#define FLAG_ADDRESS_1          16
#define FLAG_PORTS_1            32
#define FLAG_ADDRESS_2          64
#define FLAG_PORTS_2            128
#define FLAG_TUNNEL_TYPE        256
#define FLAG_TUNNEL_EP          512
#define FLAG_ERROR_TYPE         1024
#define FLAG_OTHER              8192

// Tunnel flags
#define TFLAG_NONE              0
#define TFLAG_IP_IP             1
#define TFLAG_GRE               2
#define TFLAG_L2TP              4

#define PTYPE(x) ((uint8_t)((x) & 0xFF))
#define Required 0x100
#define Multiple 0x200
#define REQUIRED 0x100
#define MULTIPLE 0x200

#define PARAM_ADDRESS_EP        20

// RSIP Messages description
static const int RSIPMessages[][11] = {

	// UNSPECIFIED MESSAGE
	{ 0 },

	// ERROR_RESPONSE
	{ REQUIRED|PARAM_ERROR, PARAM_MESSAGE_COUNTER, PARAM_CLIENT_ID, 
	  PARAM_BIND_ID, 0 },

	// REGISTER_REQUEST
	{ PARAM_MESSAGE_COUNTER, 0 },

	// REGISTER_RESPONSE
	{ REQUIRED|PARAM_CLIENT_ID, REQUIRED|PARAM_LEASE_TIME, 
	  REQUIRED|PARAM_FLOW_POLICY, PARAM_MESSAGE_COUNTER, PARAM_RSIP_METHOD,
	  MULTIPLE|PARAM_TUNNEL_TYPE, 0 },

	// DEREGISTER_REQUEST
	{ REQUIRED|PARAM_CLIENT_ID, PARAM_MESSAGE_COUNTER, 0 },

	// DEREGISTER_RESPONSE
	{ REQUIRED|PARAM_CLIENT_ID, PARAM_MESSAGE_COUNTER, 0 },

	// ASSIGN_REQUEST_RSAIP
	{ REQUIRED|PARAM_CLIENT_ID, REQUIRED|PARAM_ADDRESS, 
	  REQUIRED|PARAM_ADDRESS, REQUIRED|PARAM_PORT,
	  PARAM_MESSAGE_COUNTER, PARAM_LEASE_TIME, PARAM_TUNNEL_TYPE, 0 },

	// ASSIGN_RESPONSE_RSAIP
	{ REQUIRED|PARAM_CLIENT_ID, REQUIRED|PARAM_BIND_ID, 
	  REQUIRED|PARAM_ADDRESS, REQUIRED|PARAM_ADDRESS, REQUIRED|PARAM_PORT,
	  REQUIRED|PARAM_LEASE_TIME, REQUIRED|PARAM_TUNNEL_TYPE, 
	  PARAM_ADDRESS, PARAM_MESSAGE_COUNTER, 0 },

	// ASSIGN_REQUEST_RSAPIP
	{ REQUIRED|PARAM_CLIENT_ID, REQUIRED|PARAM_ADDRESS, 
	  REQUIRED|PARAM_PORT, REQUIRED|PARAM_ADDRESS, REQUIRED|PARAM_PORT,
	  PARAM_MESSAGE_COUNTER, PARAM_LEASE_TIME, PARAM_TUNNEL_TYPE, 0 },

	// ASSIGN_RESPONSE_RSAPIP
	{ REQUIRED|PARAM_CLIENT_ID, REQUIRED|PARAM_BIND_ID, 
	  REQUIRED|PARAM_ADDRESS, REQUIRED|PARAM_PORT,
	  REQUIRED|PARAM_ADDRESS, REQUIRED|PARAM_PORT,
	  REQUIRED|PARAM_LEASE_TIME, REQUIRED|PARAM_TUNNEL_TYPE, 
	  PARAM_ADDRESS, PARAM_MESSAGE_COUNTER, 0 },

	// EXTEND_REQUEST
	{ REQUIRED|PARAM_CLIENT_ID, REQUIRED|PARAM_BIND_ID, 
	  PARAM_LEASE_TIME, PARAM_MESSAGE_COUNTER, 0 },

	// EXTEND_RESPONSE
	{ REQUIRED|PARAM_CLIENT_ID, REQUIRED|PARAM_BIND_ID, 
	  REQUIRED|PARAM_LEASE_TIME, PARAM_MESSAGE_COUNTER, 0 },

	// FREE_REQUEST
	{ REQUIRED|PARAM_CLIENT_ID, REQUIRED|PARAM_BIND_ID, 
	  PARAM_MESSAGE_COUNTER, 0 },

	// FREE_RESPONSE
	{ REQUIRED|PARAM_CLIENT_ID, REQUIRED|PARAM_BIND_ID, 
	  PARAM_MESSAGE_COUNTER, 0 },

	// QUERY_REQUEST
	// Ignoring such request
	{ REQUIRED|PARAM_CLIENT_ID, PARAM_MESSAGE_COUNTER, 0 },

	// QUERY_RESPONSE
	// 'Don't know' reply only
	{ REQUIRED|PARAM_CLIENT_ID, PARAM_MESSAGE_COUNTER, 0 },

	// LISTEN_REQUEST
	{ REQUIRED|PARAM_CLIENT_ID, REQUIRED|PARAM_ADDRESS, 
	  REQUIRED|PARAM_PORT, REQUIRED|PARAM_ADDRESS, REQUIRED|PARAM_PORT,
	  PARAM_MESSAGE_COUNTER, PARAM_LEASE_TIME, 0 },

	// LISTEN_RESPONSE
	{ REQUIRED|PARAM_CLIENT_ID, REQUIRED|PARAM_BIND_ID, 
	  REQUIRED|PARAM_ADDRESS, REQUIRED|PARAM_PORT,
	  REQUIRED|PARAM_ADDRESS, REQUIRED|PARAM_PORT,
	  REQUIRED|PARAM_TUNNEL_TYPE, REQUIRED|PARAM_LEASE_TIME, 
	  PARAM_ADDRESS, PARAM_MESSAGE_COUNTER, 0 }
};


// Macros, by Jose Vasconcellos
#define GET16(ptr) ntohs(*((uint16_t*)(ptr)))
#define GET32(ptr) ntohl(*((uint32_t*)(ptr)))
#define SET16(ptr, val) do {*((uint16_t*)(ptr)) = htons(val);} while (0)
#define SET32(ptr, val) do {*((uint32_t*)(ptr)) = htonl(val);} while (0)

#endif
