//***************************************************************************
//* protocolHdlr.cpp
//*
//* description    : class ProtocolHandler Implementation
//*                  Process the messages of the RSIP protocol.
//*
//* copyright      : (C) 2001 Cedric de Launois
//*                  Universite Catholique de Louvain, Belgium
//* begin          : Sept 24 2001
//* email          : cedric@delaunois.org
//***************************************************************************

/****************************************************************************
 * Copyright (c) 1998 Free Software Foundation, Inc.                        *
 *                                                                          *
 * Permission is hereby granted, free of charge, to any person obtaining a  *
 * copy of this software and associated documentation files (the            *
 * "Software"), to deal in the Software without restriction, including      *
 * without limitation the rights to use, copy, modify, merge, publish,      *
 * distribute, distribute with modifications, sublicense, and/or sell       *
 * copies of the Software, and to permit persons to whom the Software is    *
 * furnished to do so, subject to the following conditions:                 *
 *                                                                          *
 * The above copyright notice and this permission notice shall be included  *
 * in all copies or substantial portions of the Software.                   *
 *                                                                          *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS  *
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF               *
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.   *
 * IN NO EVENT SHALL THE ABOVE COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,   *
 * DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR    *
 * OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR    *
 * THE USE OR OTHER DEALINGS IN THE SOFTWARE.                               *
 *                                                                          *
 * Except as contained in this notice, the name(s) of the above copyright   *
 * holders shall not be used in advertising or otherwise to promote the     *
 * sale, use or other dealings in this Software without prior written       *
 * authorization.                                                           *
 ****************************************************************************/

#include "protocolHdlr.h"



ProtocolHandler::ProtocolHandler(Options* o) 
{
	assert(o!=NULL);
	opts = o;
	proxy = new Proxy(sh, opts);
	proxy->start();

	if (pthread_cond_init(&condition, NULL)) throw NotEnoughMemory();
	if (pthread_mutex_init(&mutex, NULL)) throw NotEnoughMemory();

	if (pthread_create(&thprocess, NULL, (void* (*) (void*)) threadProcess, (void*)this))
		throw NotEnoughMemory();

}



ProtocolHandler::~ProtocolHandler() 
{
	status = STOPPING;
	
	// Flush the message list
	while (msgs.size()) {
		pthread_mutex_lock(&mutex);
		pthread_cond_signal(&condition);
		pthread_mutex_unlock(&mutex);
	}

	// Waiting for the thread to terminate
	// This is usually extremely rapid...
	while (status!=STOPPED) {
		pthread_mutex_lock(&mutex);
		pthread_cond_signal(&condition);
		pthread_mutex_unlock(&mutex);
		usleep(100);
	}

	pthread_join(thprocess, NULL);

	// INV: Message are no longer processed now
	//      Clients states do no longer change
	
	if (pthread_cond_destroy(&condition)!=0) throw Exception();
	if (pthread_mutex_destroy(&mutex)!=0) throw Exception();

	// This will do a delete on all the clients' states
	RSIPClient::disconnectAll(true);

	delete proxy;
}



int ProtocolHandler::process(NetworkEndPoint* nep) 
{
	MsgElem m;

	if (status==STOPPING || status==STOPPED) return -1;

	// Reads the message
	m.msg = new RSIPMessage();
	m.nep = nep;
	if (m.msg->receive(nep, 1)!=1) {
		delete m.msg;

		// We invalidate this nep so that clients no longer
		// use this nep
		while (status==BUSY) usleep(100);
		RSIPClient::invalid(nep);
		return 0;
	}

	// Adds the message the list of messages to process
	pthread_mutex_lock(&mutex);
	status = BUSY;
	msgs.push_back(m);
	pthread_cond_signal(&condition);
	pthread_mutex_unlock(&mutex);

	return 1;
}



int ProtocolHandler::popMessage(RSIPMessage* &msg, NetworkEndPoint* &nep, int timeout) 
{
	int result, n;
	struct timespec t;
	t.tv_nsec = 0;
	time(&t.tv_sec);
	t.tv_sec += timeout;
	MsgElem m;

	// Critical section
	pthread_mutex_lock(&mutex);

	if (msgs.size()==0) {
		if (timeout>=0) n=pthread_cond_timedwait(&condition,&mutex,&t);
		else n=pthread_cond_wait(&condition,&mutex);

		switch (n) {

		case 0:
			if (msgs.size()>0) {
				m = msgs.front();
				msgs.pop_front();
				result = 1;
			} else result = -1; // Connection closed
			break;

		case ETIMEDOUT:
			if (msgs.size()>0) {
				m = msgs.front();
				msgs.pop_front();
				result = 1;
			} else result = 0; // Time out
			break;

		default:
			// Should not occur
			result = -2;
			break;
		}

	} else { // At least 1 message is in the stack

		m = msgs.front();
		msgs.pop_front();
		result = 1;

	}

	pthread_mutex_unlock(&mutex);
	// End of critical section

	if (result) {
		msg = m.msg;
		nep = m.nep;
	}
	return result;
}



int ProtocolHandler::doProcess() 
{
	RSIPMessage *msg = NULL;
	RSIPMessage *reply = NULL;
	RSIPClient* cl;
	NetworkEndPoint* nep;
	MessageProcessor mp(sh, proxy, opts);
	Address from;
	int retval;

	while (status!=STOPPING) {
			
		// Pop one message from message list
		status = RUNNING;
		retval = popMessage(msg, nep, 1);
		if (retval<=-2) break;    // Unexpected error
		if (retval<=0) continue;  // Time-out or client disconnected
		
		if (opts->dump) msg->dump();
		from = msg->getAddress();
		
		// A message has been received
		// The message is finally delete by a cl->updateLastMsg() of a new
		// message which makes the message no more the last one


		// Sanity check
		if ((retval = msg->check())!=1) {
			// Bad message received
			logmsg(LOG_WARNING, "Bad or illegal message received from host %s (UDP).\n", from.toChar());
			// Silently discarded !
			delete msg;
			continue; 
		}


		// Getting the state of the client, and lock it if possible
		RSIPClient::lockdb();
		if (RSIPClient::search(from, cl)==0) {
			
			// It is a new client
			cl = new RSIPClient(from, nep, opts);
			cl->lock();

		} else {

			if (msgs.size()==1) usleep(100);	

			if (cl->trylock() == EBUSY) {
				// The client is busy, re-pushing the message
				// on the list for later processing
				MsgElem m = { nep, msg };
				pthread_mutex_lock(&mutex);
				msgs.push_back(m);
				pthread_cond_signal(&condition);
				pthread_mutex_unlock(&mutex);
				RSIPClient::unlockdb();
				continue;
			}

		}
		RSIPClient::unlockdb();
		
		// INV: we have a lock on the client
		// Check for retransmission and updating last msg received
		if (cl->updateLastMsg(msg)) {
			// The message has already been received !
			// msg is now the last message transmitted
			
			logmsg(LOG_INFO, "UDP Retransmission for host %s\n", from.toChar());
			
			cl->send(msg, nep);
			cl->unlock();
			continue;
		}
		
		
		// Process the new message received
		retval = mp.process(msg, reply, cl);
		
		
		// Sending the reply, do not delete it since it is stored in the
		// last message transmitted of the client
		if (reply) cl->send(reply, nep);
		
		cl->unlock();

		// We can destroy the state of the client if retval==0
		if (retval==0) {
			RSIPClient::lockdb();
			delete cl;
			RSIPClient::unlockdb();
		}

	}

	status = STOPPED;

	return 0;
}



void* threadProcess(ProtocolHandler* hdlr) 
{
	hdlr->doProcess();
	pthread_exit(0);
}


