//***************************************************************************
//* protocolHdlr.h
//*
//* description    : class ProtocolHandler
//*                  Process the messages of the RSIP protocol.
//*
//* copyright      : (C) 2001 Cedric de Launois
//*                  Universite Catholique de Louvain, Belgium
//* begin          : Sept 24 2001
//* email          : cedric@delaunois.org
//***************************************************************************

/****************************************************************************
 * Copyright (c) 1998 Free Software Foundation, Inc.                        *
 *                                                                          *
 * Permission is hereby granted, free of charge, to any person obtaining a  *
 * copy of this software and associated documentation files (the            *
 * "Software"), to deal in the Software without restriction, including      *
 * without limitation the rights to use, copy, modify, merge, publish,      *
 * distribute, distribute with modifications, sublicense, and/or sell       *
 * copies of the Software, and to permit persons to whom the Software is    *
 * furnished to do so, subject to the following conditions:                 *
 *                                                                          *
 * The above copyright notice and this permission notice shall be included  *
 * in all copies or substantial portions of the Software.                   *
 *                                                                          *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS  *
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF               *
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.   *
 * IN NO EVENT SHALL THE ABOVE COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,   *
 * DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR    *
 * OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR    *
 * THE USE OR OTHER DEALINGS IN THE SOFTWARE.                               *
 *                                                                          *
 * Except as contained in this notice, the name(s) of the above copyright   *
 * holders shall not be used in advertising or otherwise to promote the     *
 * sale, use or other dealings in this Software without prior written       *
 * authorization.                                                           *
 ****************************************************************************/

class ProtocolHandler;

#ifndef _PROTOCOL_HANDLER_H_
#define _PROTOCOL_HANDLER_H_

#include <pthread.h>
#include <deque>
#include "rsipMessage.h"
#include "rsipClient.h"
#include "msgProcessor.h"
#include "common.h"
#include "netEndPoint.h"
#include "string.h"
#include "proto.h"
#include "proxy.h"

class ProtocolHandler {

 public:
	enum State { RUNNING, STOPPING, STOPPED, BUSY };


	ProtocolHandler(Options* o);
	// Constructor. Create a processing module for RSIP messages 
	// coming from an end point
	// PRE : o not NULL
	// POST: protocol handler started


	~ProtocolHandler();
	// Destructor
	// Stop and destroy the protocol handler :
	// Flush the message list and disconnect all the clients


	int process(NetworkEndPoint* nep);
	// Read data received from nep and put the message on a list
	// of message to be processed
	// PRE : a message is ready to be read on nep
	// POST: return 1 if success,
	//              0 if the socket is closed
	//             -1 if the handler is STOPPED
	// NOTE: the function immediately read the incoming message,
	//       then put it on a list of received messages to process,
	//       and then returns.
	//       So, the message is not processed entirely when the
	//       function returns.


	inline State state() { return status; }
	// Return the status of the manager
	// PRE : /
	// POST: returns RUNNING, BUSY, STOPPING or STOPPED


 private:

	typedef struct msgListElem {
		NetworkEndPoint* nep;
		RSIPMessage* msg;
	} MsgElem;

	State status;           // Protocol Handler State
	deque<MsgElem> msgs;    // List of messages to process
	Options* opts;          // Options
	Proxy* proxy;           // A proxy to handle traffic forwarding

	pthread_t thprocess;
	pthread_cond_t condition;
	pthread_mutex_t mutex;


	int popMessage(RSIPMessage* &m, NetworkEndPoint* &nep, int timeout);
	// Waits for a message to process, pops the messages from the message list
	// PRE : timeout : a time out in seconds
	// POST: - m is the new message to process
	//         OR NULL if a time out occured
	//       - nep is the End Point from which the message comes from,
	//         OR undefined if a timeout occured
	//       - returns 0 if a time out occured
	//                 1 if success
	//                -1 if client disconnected
	//                -2 if unexpected error


	int doProcess();
	// Pop the messages processed from the list and process them
	// until status is STOPPING


	friend void* threadProcess(ProtocolHandler*);
	// The main thread

};
	
void* threadProcess(ProtocolHandler*);

#endif
