//**************************************************************************
//*
//* description    : proxy.cpp
//*                  Proxy RSAP
//* copyright      : (C) 2001 Cedric de Launois
//*                  Universite Catholique de Louvain, Belgium
//* begin          : Sept 11 2001
//* email          : cedric@delaunois.org
//**************************************************************************

/****************************************************************************
 * Copyright (c) 1998 Free Software Foundation, Inc.                        *
 *                                                                          *
 * Permission is hereby granted, free of charge, to any person obtaining a  *
 * copy of this software and associated documentation files (the            *
 * "Software"), to deal in the Software without restriction, including      *
 * without limitation the rights to use, copy, modify, merge, publish,      *
 * distribute, distribute with modifications, sublicense, and/or sell       *
 * copies of the Software, and to permit persons to whom the Software is    *
 * furnished to do so, subject to the following conditions:                 *
 *                                                                          *
 * The above copyright notice and this permission notice shall be included  *
 * in all copies or substantial portions of the Software.                   *
 *                                                                          *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS  *
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF               *
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.   *
 * IN NO EVENT SHALL THE ABOVE COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,   *
 * DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR    *
 * OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR    *
 * THE USE OR OTHER DEALINGS IN THE SOFTWARE.                               *
 *                                                                          *
 * Except as contained in this notice, the name(s) of the above copyright   *
 * holders shall not be used in advertising or otherwise to promote the     *
 * sale, use or other dealings in this Software without prior written       *
 * authorization.                                                           *
 ****************************************************************************/

#include "proxy.h"

#define for_each(set,i) for (i = (set).begin(); i != (set).end(); ++i )

Proxy::Proxy(Shell* shl, Options* opts)
{
	int i=0;
	struct stat statbuf;
	char *iptablesPaths[] = { 
		"/sbin/iptables",
		"/usr/sbin/iptables",
		"/usr/local/sbin/iptables",
		0
	};

	bufsize = 65550;   // Eth HDR + IP max length
	status = STOPPED;

	if (opts==NULL) {
		logmsg(LOG_ERR, "Invalid Options\n");
		throw Exception();
	} else o = opts;

	if ((s = shl)==NULL) {
		logmsg(LOG_ERR, "Invalid Shell\n");
		throw Exception();
	}

	if ((extifindex = if_nametoindex(o->ext_if))==0) {
		logmsg(LOG_ERR, "Unknown interface %s\n", o->ext_if);
		throw Exception();
	}

	while ((iptables = iptablesPaths[i++])) {
		if (stat(iptables, &statbuf)==0)
			break;
	}

	if (!iptables) {
		logconsole(LOG_ERR, "Can't find iptables binary.\n");
		throw Exception();
	}

	// Disabling the log of martian packets
	FILE* file;
	if ((file=fopen("/proc/sys/net/ipv4/conf/all/log_martians", "r+"))) {
		logMartianBkpAll = fgetc(file);
		rewind(file);
		fputc('0',file);
		fclose(file);
	}
	if ((file=fopen("/proc/sys/net/ipv4/conf/default/log_martians", "r+"))) {
		logMartianBkpDflt = fgetc(file);
		rewind(file);
		fputc('0',file);
		fclose(file);
	}

	if (pthread_mutex_init(&lock, NULL)) throw NotEnoughMemory();
	if (pthread_mutex_init(&icmplock, NULL)) throw NotEnoughMemory();
	if (pthread_mutex_init(&fraglock, NULL)) throw NotEnoughMemory();
}



Proxy::~Proxy()
{
	status = STOPPING;
	map<Address, AddressTableEntry*>::iterator i;

	// Stopping the routing thread
	pthread_cancel(thproxy);
	pthread_join(thproxy, NULL);

	// Deleting the addressTable
	for_each(addressTable, i) delete i->second;
	addressTable.clear();

	// Restoring the log of martian packets
	FILE* file;
	if ((file=fopen("/proc/sys/net/ipv4/conf/all/log_martians", "w"))) {
		fputc(logMartianBkpAll,file);
		fclose(file);
	}
	if ((file=fopen("/proc/sys/net/ipv4/conf/default/log_martians", "w"))) {
		fputc(logMartianBkpDflt,file);
		fclose(file);
	}

	if (pthread_mutex_destroy(&lock)!=0) throw Exception();
	if (pthread_mutex_destroy(&icmplock)!=0) throw Exception();
	if (pthread_mutex_destroy(&fraglock)!=0) throw Exception();
}



void Proxy::start() 
{
	if (status!=STOPPED) return;

	// Creating the routing thread
	if (pthread_create(&thproxy, NULL, (void* (*) (void*)) router, (void *)this))
		throw NotEnoughMemory();

	status = STARTED;
}



int Proxy::addRoute(Address a, unsigned short port, const char* iface, unsigned short number) 
{
	pthread_mutex_lock(&lock);
	unsigned int index;
	char strport[13];

	if ((index=if_nametoindex(iface))==0) {
		logmsg(LOG_ERR, "Unknown interface %s\n", iface);
		pthread_mutex_unlock(&lock);
		return -2;
	}

	// Getting the AddressTable Entry
	map<Address, AddressTableEntry*>::iterator i;
	AddressTableEntry* ate;
	if ((i=addressTable.find(a))==addressTable.end()) {
		ate = new AddressTableEntry;
		ate->all = false;
		ate->addr = a;
		addressTable.insert(make_pair(a, ate));
	} else ate = i->second;

	// Case 1: Routing of a whole IP address, only if a is the server IP
	if ((a == o->ext_ip) && (port==0 || ate->all)) {

		if (ate->portTable.size()) {
			// Some ports are currently being routed
			pthread_mutex_unlock(&lock);
			return -4;
		}
		if (s->execute(iptables,"-I","INPUT 1","-d",o->ext_ip.toChar(),"-j","DROP",NULL)!=0) {
			logmsg(LOG_ERR, "iptables error.\n");
			pthread_mutex_unlock(&lock);
			return -3;
		}
		Interval i(0,65535);
		ate->portTable.insert(make_pair(i, index));
		ifaceAdd(ate, index, 1);
		ate->all = true;
		pthread_mutex_unlock(&lock);
		return 1;
	}

	Interval ivl(port, port+number-1);

	// Case 2: Standard case : Routing of a port
	if (ate->portTable.find(port)!=ate->portTable.end()) {
		pthread_mutex_unlock(&lock);
		return -1;
	}


	if (a == o->ext_ip) {

		// If a is the address of this computer, then we have to 
		// prevent packets from being catched by this computer
		if (number > 1) snprintf(strport, 13, "%i:%i", port, port+number-1);
		else {
			number = 1;
			snprintf(strport, 6, "%i", port);
		}

		// Inserting iptables rules
		if (s->execute(iptables,"-I","INPUT 1","--protocol","tcp","--dport",strport,"-j","DROP",NULL)!=0) {
			logmsg(LOG_ERR, "iptables error.\n");
			pthread_mutex_unlock(&lock);
			return -3;
		}

		if (s->execute(iptables,"-I","INPUT 1","--protocol","udp","--dport",strport,"-j","DROP",NULL)!=0) {
			// Backtracking
			s->execute(iptables,"-D","INPUT","--protocol","tcp","--dport",strport,"-j","DROP",NULL);
			logmsg(LOG_ERR, "iptables error.\n");
			pthread_mutex_unlock(&lock);
			return -3;
		}

	}

	// Updating routing table port<->Iface
        ate->portTable.insert(make_pair(ivl, index));

	// Updating Iface information
	ifaceAdd(ate, index, number);

	pthread_mutex_unlock(&lock);
	return 1;
}



int Proxy::delRoute(Address a, unsigned short port, unsigned short number) 
{
	pthread_mutex_lock(&lock);
	unsigned int index;
	char strport[13];
	AddressTableEntry* ate;
	Interval ivl(port, port+number-1);
	map<Address, AddressTableEntry*>::iterator i;
	map<Interval, unsigned int>::iterator j;

	if ((i=addressTable.find(a))==addressTable.end()) return -1;
	else ate = i->second;

	if ((j=ate->portTable.find(ivl))==ate->portTable.end()) {
		// Unknown port
		pthread_mutex_unlock(&lock);
		return -1;
	} else index = j->second;

	if (a == o->ext_ip) {

		if (port==0 || ate->all) {
			ate->all = false;
			if (ate->ifaces.find(index)==ate->ifaces.end()) {
				pthread_mutex_unlock(&lock);
				return 0;
			}
			s->execute(iptables,"-D","INPUT","-d",o->ext_ip.toChar(),"-j","DROP",NULL);
			ate->portTable.erase(j);
			ifaceDel(ate, index, 1);
			pthread_mutex_unlock(&lock);
			return 1;
		}
		
		if (number > 1) snprintf(strport, 13, "%i:%i", port, port+number-1);
		else {
			number = 1;
			snprintf(strport, 6, "%i", port);
		}

		s->execute(iptables,"-D","INPUT","--protocol","tcp","--dport",strport,"-j","DROP",NULL);
		s->execute(iptables,"-D","INPUT","--protocol","udp","--dport",strport,"-j","DROP",NULL);
	}

	ate->portTable.erase(j);
	ifaceDel(ate, index, number);

	if (!ate->portTable.size()) {
		addressTable.erase(i);
		delete ate;
	}

	pthread_mutex_unlock(&lock);
	return 1;
}



int Proxy::openRawEth(int &sock, const char *ifname, int type)
{
	struct sockaddr_ll addr;
	bzero((char*)&addr, sizeof(addr));

	if ( (sock = socket(PF_PACKET, type, htons(ETH_P_IP))) < 0) {
		logmsg(LOG_ERR, "Can't open socket\n");
		return 0;
	}

	if (!ifname) return 1;

	addr.sll_family   = AF_PACKET;
	addr.sll_protocol = htons(ETH_P_IP);
	addr.sll_ifindex  = if_nametoindex(ifname);
	if (bind(sock, (struct sockaddr*)&addr, sizeof(addr)) < 0) {
		logmsg(LOG_ERR, "Can't bind socket\n");
		return 0;
	}

	return 1;
}



int Proxy::openRawIP(int &sock)
{
	if ( (sock = socket(PF_INET, SOCK_RAW, IPPROTO_RAW)) < 0) {
		logmsg(LOG_ERR, "Can't open raw IP socket\n");
		return 0;
	}
	
	return 1;
}



void Proxy::doRoute()
{
	pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, NULL);

	int nread;
	unsigned int localIf = if_nametoindex("lo");
	char buf[bufsize];
	struct sockaddr_ll ethfrom;
	unsigned int s = 0;
	fd_set readfds;

	if (!openRawEth(rawSockIn, NULL, SOCK_DGRAM)) return;
	if (!openRawIP(ipSockOut)) return;

	pthread_cleanup_push((void (*) (void*))close, (void*)&rawSockIn);
	pthread_cleanup_push((void (*) (void*))close, (void*)&ipSockOut);


	// Loop to receive and route both ingress and outgress RSIP packets received
	// This loop is stopped asynchronously by a pthread_cancel() on this thread

	while(1) {

		FD_ZERO(&readfds);
		FD_SET(rawSockIn, &readfds);


		// Start of cancellable section
		pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL);

		select(FD_SETSIZE, &readfds, NULL, NULL, NULL);

		nread = recvfrom(rawSockIn, buf, bufsize, 0, (struct sockaddr*)&ethfrom, &s);

		// End of cancellable section
		pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, NULL);

		// No routing needed if there is no RSIP client
		if (!addressTable.size() || nread < 20) continue;

		// Routing the packet received
		if ((unsigned int)ethfrom.sll_ifindex == extifindex) {
			// Packet coming from the external interface
			// i.e.: Internet->RSIP host
			// Routing packet

			routeIngress(buf, nread, ethfrom, rawSockIn);

		} else if ((unsigned int)ethfrom.sll_ifindex != localIf) {
			// Other packet, but not a packet coming from local Iface (ifindex==1)

			routeOutgress(buf, nread, ethfrom, ipSockOut);

		}
	}


	pthread_cleanup_pop(1);
	pthread_cleanup_pop(1);
}



void Proxy::routeIngress(char* pkt, int len, struct sockaddr_ll &ethfrom, int sock)
{
	struct sockaddr_ll ethto;
	struct sockaddr_in ipto;
	unsigned short port;
	AddressTableEntry* ate = NULL;
	map<Address, AddressTableEntry*>::iterator i;
	map<Interval, unsigned int>::iterator j;

	// Getting destination IP address
	memcpy((char*)&ipto.sin_addr.s_addr, pkt+16, 4);
	Address dst(ipto.sin_addr.s_addr);

	if ((i=addressTable.find(dst))==addressTable.end()) 
		return;
	else 
		ate = i->second;

	if (ate->all) port = 0; // The whole port range must be forwarded
	else {
		if (pkt[9]==1)
			// Processing ICMP messages
			port = routeIngressICMP(pkt, len, ethfrom, sock);
		else
			// Processing Regular Packets
			port = routeIngressRegular(pkt, len, ethfrom, sock);

		if (!port)
			return;
	}

	Interval ivl(port,port);
	if ((j=ate->portTable.find(ivl))!=ate->portTable.end()) {
		ethto.sll_family   = ethfrom.sll_family;
		ethto.sll_protocol = ethfrom.sll_protocol;
		ethto.sll_ifindex = j->second;
		if (o->dump) dump(pkt, ethfrom, len);

		if (sendto(sock, pkt, len, 0, 
				(struct sockaddr*)&ethto, sizeof(ethto)) < 0) {

			if (errno == EMSGSIZE)
				logmsg(LOG_ERR, "Can't send rsip packet. MTU problem.\n");
			else logmsg(LOG_ERR, "Can't send rsip packet err %d\n", errno);
		}
	}
}



unsigned short Proxy::routeIngressRegular(char* pkt, int len, struct sockaddr_ll &ethfrom, int sock)
{
	char* payload = pkt + ((int)(pkt[0]&0xf))*4;  // Points to the payload of the IP packet
	map<FragID, uint16_t>::iterator i;
	uint16_t port =  ntohs(*((unsigned short*)(payload+2)));
	bool DF = pkt[6]&0x40;
	bool MF = pkt[6]&0x20;
	unsigned short offset = (ntohs(*((uint16_t*)(pkt+6)))&0x1FFF);

	if (DF || (offset==0 && MF==0))
		return port;

	cerr << "FRAG" << endl;

	// INV: DF==0
	FragID fid;
	fid.id = *((uint16_t*)&pkt[4]);
	fid.proto = pkt[9];
	fid.src = *((uint32_t*)(pkt+12));
	fid.dst = *((uint32_t*)(pkt+16));

	if (offset==0) {

		// INV: MF==1 && offset==0
		// Storing data for next fragments

		pthread_mutex_lock(&fraglock);
		if ((i=frag.find(fid))==frag.end()) {
			frag.insert(make_pair(fid, port));
			cerr << "Frag id inserted p=" << port << endl;

			// Scheduling the removal of the FragID id in 10 sec.
			tmr->schedule((void*(*)(void*))removeOldFragID,
				      (void*)this, 
				      10);
		}
		pthread_mutex_unlock(&fraglock);

		return port;
	}

	// INV: offset!=0 && DF==0

	// Searching the port related to this fragment
	if ((i=frag.find(fid))==frag.end())
		return 0; // Unknown fragment

	cerr << "Frag pkt routed p=" << i->second << endl;
		
	return i->second;
}



unsigned short Proxy::routeIngressICMP(char* pkt, int len, struct sockaddr_ll &ethfrom, int sock)
{
	map<unsigned int, unsigned int>::iterator j;
	struct sockaddr_ll ethto;
	char* payload = pkt + ((int)(pkt[0]&0xf))*4;  // Points to the payload of the IP packet

	if (payload[0]==0) {

		// Processing an ICMP echo reply
		// Getting an identifier for this ICMP request
		// The identifier is the concatenation of <Identifier><Sequence Number>

		unsigned int id = ntohl(*((unsigned int*)(payload+4)));
		
		pthread_mutex_lock(&icmplock);

		if ((j=icmp.find(id))!=icmp.end()) {
			ethto.sll_family   = ethfrom.sll_family;
			ethto.sll_protocol = ethfrom.sll_protocol;
			ethto.sll_ifindex = j->second;
			if (o->dump) dump(pkt, ethfrom, len);
			sendto(sock, pkt, len, 0, (struct sockaddr*)&ethto, sizeof(ethto));
		} 

		pthread_mutex_unlock(&icmplock);
		return 0;
		
	}

	// Processing ICMP error messages
	return ntohs(*((unsigned short*)(payload+28))); 
}



void Proxy::routeOutgress(char* pkt, int len, struct sockaddr_ll &ethfrom, int ipsock)
{
	struct sockaddr_in ipto, ipfrom;
	AddressTableEntry* ate = NULL;
	map<Address, AddressTableEntry*>::iterator i;

	// Getting src and dst IP addresses
	memcpy((char*)&ipfrom.sin_addr.s_addr, pkt+12, 4);
	memcpy((char*)&ipto.sin_addr.s_addr, pkt+16, 4);
	Address src(ipfrom.sin_addr.s_addr);
	Address dst(ipto.sin_addr.s_addr);

	// Checking that the packet is coming from a RSIP interface
	if ((i=addressTable.find(src))==addressTable.end()) return;
	else ate = i->second;

	if (ate->ifaces.find((unsigned int)ethfrom.sll_ifindex)==ate->ifaces.end())
		return;

	char* payload = pkt + ((int)(pkt[0]&0xf))*4;  // Points to the payload of the IP packet
		
	if (pkt[9]==1 && payload[0]==8) {

		// Processing an ICMP echo request
		// Getting an identifier for this ICMP request
		// The identifier is the concatenation of <Identifier><Sequence Number>

		unsigned int id = ntohl(*((unsigned int*)(payload+4)));

		pthread_mutex_lock(&icmplock);
		if (icmp.find(id)==icmp.end()) {
			icmp.insert(make_pair(id, ethfrom.sll_ifindex));
			// Scheduling the removal of the ICMP id in 10 sec.
			tmr->schedule((void*(*)(void*))removeOldICMP,
				      (void*)this, 
				      10);
		}
		pthread_mutex_unlock(&icmplock);
	}

	// Forwarding the packet, routing is done by the kernel

	sendto(ipsock, pkt, len, 0,  dst.getSockaddr(), sizeof(struct sockaddr_in));
	if (o->dump) dump(pkt, ethfrom, len);

}



int Proxy::ifaceAdd(AddressTableEntry* ate, unsigned int index, unsigned int number)
{
	assert(ate!=NULL);

	char name[IF_NAMESIZE];
	map<unsigned int, unsigned int>::iterator i;

	if ((i=ate->ifaces.find(index))!=ate->ifaces.end())
		ate->ifaces[index] = i->second + number;
	else {
		ate->ifaces.insert(make_pair(index, number));
		if (ate->addr != o->ext_ip) {
			s->execute(iptables,"-I","INPUT 1","-s",ate->addr.toChar(),"-j","DROP",NULL);
			s->execute(iptables,"-I","FORWARD 1","-s",ate->addr.toChar(),"-j","DROP",NULL);
			s->execute(iptables,"-I","INPUT 1","-d",ate->addr.toChar(),"-j","DROP",NULL);
			s->execute(iptables,"-I","FORWARD 1","-d",ate->addr.toChar(),"-j","DROP",NULL);
		} else {
			s->execute(iptables,"-I","INPUT 1","-s",o->ext_ip.toChar(),"-i",if_indextoname(index, name),"-j","DROP",NULL);
			s->execute(iptables,"-I","OUTPUT 1","-s",o->ext_ip.toChar(),"-j","ACCEPT",NULL);
		}
	}

	return 1;
}



int Proxy::ifaceDel(AddressTableEntry* ate, unsigned int index, unsigned int number)
{
	assert(ate!=NULL);

	unsigned int nb = 0;
	char name[IF_NAMESIZE];
	map<unsigned int, unsigned int>::iterator i;

	if ((i=ate->ifaces.find(index))==ate->ifaces.end()) return 0;
	else nb = i->second;

	nb-=number;
	if (nb<=0) {
		ate->ifaces.erase(i);
		if (ate->addr != o->ext_ip) {
			s->execute(iptables,"-D","INPUT","-s",ate->addr.toChar(),"-j","DROP",NULL);
			s->execute(iptables,"-D","FORWARD","-s",ate->addr.toChar(),"-j","DROP",NULL);
			s->execute(iptables,"-D","INPUT","-d",ate->addr.toChar(),"-j","DROP",NULL);
			s->execute(iptables,"-D","FORWARD","-d",ate->addr.toChar(),"-j","DROP",NULL);
		} else {
			s->execute(iptables,"-D","INPUT","-s",o->ext_ip.toChar(),"-i",if_indextoname(index, name),"-j","DROP",NULL);
			s->execute(iptables,"-D","OUTPUT","-s",o->ext_ip.toChar(),"-j","ACCEPT",NULL);
		}
	} else ate->ifaces[index] = nb;

	return 1;
}



void router(Proxy* p)
{
	p->doRoute();
}



void removeOldICMP(Proxy* p)
{

	// All time outs are identical. So when we have to remove an ICMP id,
	// we are sure it is the first (the older) one...
	pthread_mutex_lock(&(p->icmplock));
	p->icmp.erase(p->icmp.begin());
	pthread_mutex_unlock(&(p->icmplock));
}



void removeOldFragID(Proxy* p)
{

	// All time outs are identical. So when we have to remove a Frag ID,
	// we are sure it is the first (the older) one...
	pthread_mutex_lock(&(p->fraglock));
	p->frag.erase(p->frag.begin());
	pthread_mutex_unlock(&(p->fraglock));
	cerr << "Frag id removed" << endl;
}



void Proxy::dump(char* pkt, struct sockaddr_ll &ethfrom, int len) 
{
	// Dumping packets

	struct sockaddr_in ipfrom, ipto;
	ipto.sin_family = AF_INET;
	char* payload = pkt + ((int)(pkt[0]&0xf))*4;  // Points to the payload of the IP packet
	char name[IF_NAMESIZE];

	// Retrieving the IP addresses of the packet
	memcpy((char*)&ipto.sin_addr.s_addr, pkt+16, 4);
	memcpy((char*)&ipfrom.sin_addr.s_addr, pkt+12, 4);
	unsigned short portfrom = ntohs(*((unsigned short*)(payload)));
	unsigned short portto = ntohs(*((unsigned short*)(payload+2)));

	cerr << "From Iface:" << if_indextoname(ethfrom.sll_ifindex, name) << "(" << ethfrom.sll_ifindex << ")";
	cerr << " src=" << inet_ntoa(ipfrom.sin_addr) << ":";
	if (pkt[9]!=1) cerr << portfrom; else cerr << "ICMP";
	cerr << " dst=" << inet_ntoa(ipto.sin_addr) << ":";
	if (pkt[9]!=1) cerr << portto; else cerr << "ICMP";
	cerr <<  " len=" << len << endl;
	if (pkt[7]!=0) cerr << " F" << endl;
	dumpHex(pkt, len);
}



