//**************************************************************************
//*
//* description    : proxy.h
//*                  Proxy RSAP
//* copyright      : (C) 2001 Cedric de Launois
//*                  Universite Catholique de Louvain, Belgium
//* begin          : Sept 11 2001
//* email          : cedric@delaunois.org
//**************************************************************************

/****************************************************************************
 * Copyright (c) 1998 Free Software Foundation, Inc.                        *
 *                                                                          *
 * Permission is hereby granted, free of charge, to any person obtaining a  *
 * copy of this software and associated documentation files (the            *
 * "Software"), to deal in the Software without restriction, including      *
 * without limitation the rights to use, copy, modify, merge, publish,      *
 * distribute, distribute with modifications, sublicense, and/or sell       *
 * copies of the Software, and to permit persons to whom the Software is    *
 * furnished to do so, subject to the following conditions:                 *
 *                                                                          *
 * The above copyright notice and this permission notice shall be included  *
 * in all copies or substantial portions of the Software.                   *
 *                                                                          *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS  *
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF               *
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.   *
 * IN NO EVENT SHALL THE ABOVE COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,   *
 * DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR    *
 * OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR    *
 * THE USE OR OTHER DEALINGS IN THE SOFTWARE.                               *
 *                                                                          *
 * Except as contained in this notice, the name(s) of the above copyright   *
 * holders shall not be used in advertising or otherwise to promote the     *
 * sale, use or other dealings in this Software without prior written       *
 * authorization.                                                           *
 ****************************************************************************/

class Proxy;

#ifndef _PROXY_H_
#define _PROXY_H_

#include <stdio.h>
#include <iostream>
#include <stdlib.h>
#include <errno.h>
#include <net/if.h>
#include <net/if_arp.h>
#include <sys/types.h>
#include <sys/socket.h>       // For sockets in general
#include <features.h>         // for the glibc version number
#if __GLIBC__ >= 2 && __GLIBC_MINOR >= 1
#include <netpacket/packet.h>
#include <net/ethernet.h>     // the L2 protocols
#else
#include <asm/types.h>
#include <linux/if_packet.h>
#include <linux/if_ether.h>   // The L2 protocols
#endif
#include <sys/time.h>         // For select()...
#include <sys/stat.h>
#include <sys/types.h>
#include <arpa/inet.h>
#include <unistd.h>           // For close()...
#include <strings.h>          // For bzero()...
#include <pthread.h>
#include <map>

#include "shell.h"
#include "common.h"
#include "log.h"
#include "netutils.h"
#include "address.h"
#include "interval.h"
#include "fragid.h"
#include "assert.h"

// This code is a RSIP proxy which routes messages destined to RSIP hosts
// according to the port from which messages are coming from.
//
// This is needed when RSAP-IP is used. In this case, the RSIP host may
// have the same address as the proxy ! So we need to distinguish between
// messages addressed to the proxy and other messages addressed to the
// RSIP host. The proxy just inspects the correspondance between a host and
// the port which has been allocated to it and the port from which the
// message is coming from.
//
// NOTE : we need a PACKET socket from incoming messages since these messages
// are correctly routed by the kernel. Indeed, the kernel believes the message
// is to address locally since those messages shows an IP address of the proxy.
// We must reroute these messages towards RSIP hosts through tunnels.

class Proxy {

 public:

	Proxy(Shell*, Options* o);
	// Constructor


	~Proxy();
	// Destructor


	void start();
	// Start RSAP proxy router
	// PRE : /
	// POST: proxy router started


	int addRoute(Address a, unsigned short port, const char* iface, unsigned short number);
	// PRE : /
	// POST: - add an entry in the address and port tables of the proxy :
	//         packets coming from port 'port' with dst address 'a' will be
	//         forwarded to interface named 'iface'
	//       - if number is > 1, then ports 'port' to 'port+number-1'
	//         will have this behaviour
	//       - if port is 0, then the whole ip will be rerouted 
	// RETURNS : 1 if success,
	//           -1 if the port is already allocated
	//           -2 if the interface is unknown
	//           -3 if another problem occured


	int delRoute(Address a, unsigned short port, unsigned short number);
	// PRE : /
	// POST: - delete an entry in the address and port tables of the proxy
	//         If the entry doesn't exist, then nothing is done
	//       - if number is > 1, then ports 'port' to 'port+number-1'
	//         will be deleted
	// RETURNS : 1 if a route was deleted
	//           0 otherwise


 private:

	enum State { STOPPED, STARTED, STOPPING } status;
	unsigned int extifindex;
	unsigned char logMartianBkpAll, logMartianBkpDflt;
	int bufsize;
	int rawSockIn;
	int ipSockOut;
	char* iptables;
	pthread_mutex_t lock;
	pthread_mutex_t icmplock;
	pthread_mutex_t fraglock;
	pthread_t thproxy;
	Options* o;
	Shell* s;

	typedef struct ATE {

		bool all;
		Address addr;

		map<Interval, unsigned int> portTable; 
		// Table <interval of ports, interface index>
		// The iface index is given by a if_nametoindex()
		
		map<unsigned int, unsigned int> ifaces; 
		// Table <interface index, number of port to fwd to this interface>
		// When the number of port is 0, then the iface may be deleted

	} AddressTableEntry;

	map<Address, AddressTableEntry*> addressTable; 
	// The global routing table : one port routing table for each address

	map<unsigned int, unsigned int> icmp; 
	// Table <icmp identifier, interface index>

	map<FragID, uint16_t> frag;
	// Table <fragment id, port>

	void doRoute();
	// Function that effectively do the proxy work
	// until status == STOPPING


	void routeIngress(char* pkt, int len, struct sockaddr_ll &ethfrom, int sock);
	// Route the packet coming from the public interface and send it to the
	// right host through socket sock.
	// PRE : pkt is not NULL
	//       len is the length of the packet (without any header)
	//       ethfrom is the IP header of the packet
	// POST: If the packet is an ICMP one, the reply is re-sent to the right host
	//         according to the port (ICMP error) or to the identifier (ICMP echo
	//         reply) of the ICMP packet
	//       If the packet is an IP one, the packet is re-sent to the right host
	//         eccording to the port contained in the UDP or TCP header

	unsigned short routeIngressICMP(char* pkt, int len, struct sockaddr_ll &ethfrom, int sock);
	unsigned short routeIngressRegular(char* pkt, int len, struct sockaddr_ll &ethfrom, int sock);

	void routeOutgress(char* pkt, int len, struct sockaddr_ll &ethfrom, int sock);
	// Receive the packet coming from the private interface and send it to the
	// public host through socket sock.
	// PRE : pkt is not NULL
	//       len is the length of the packet (without any header)
	//       ethfrom is the IP header of the packet
	// POST: If the packet is an ICMP echo request, an identifier is build and 
	//         recorded, in order to correctly demultiplex the reply. This 
	//         identifier is automatically deleted after 10 seconds.
	//       Otherwise, the packet is simply re-sent through sock.


	int openRawEth(int &sock, const char *ifname, int type);
	// Opens a PACKET socket bind to interface 'ifname'
	// PRE : type is SOCK_RAW or SOCK_DGRAM


	int openRawIP(int &sock);
	// Opens a RAW IP socket


	int openRawICMP(int &sock);
	// Opens a RAW IP socket


	int ifaceAdd(AddressTableEntry*, unsigned int index, unsigned int number);
	// Add 'number' uses of the interface indexed by 'index'
	// If the iface doesn't yet exist, then the iface is created


	int ifaceDel(AddressTableEntry*, unsigned int index, unsigned int number);
	// Remove 'number' uses of the interface indexed by 'index'
	// If number becomes 0, then the iface is really deleted


	void dump(char*, struct sockaddr_ll &ethfrom, int len);


	friend void removeOldICMP(Proxy* p);
	friend void removeOldFragID(Proxy* p);

	friend void router(Proxy* p);
	// The proxy thread

};

void removeOldICMP(Proxy* p);
void removeOldFragID(Proxy* p);

void router(Proxy* p);

#endif
