//****************************************************************************
//* resources.cpp 
//*
//* description    : class Resources
//*                  Pool of RSIP resources
//* copyright      : (C) 2001 Cedric de Launois
//*                  Universite Catholique de Louvain, Belgium
//* begin          : Sept 24 2001
//* email          : cedric@delaunois.org
//***************************************************************************

/****************************************************************************
 * Copyright (c) 1998 Free Software Foundation, Inc.                        *
 *                                                                          *
 * Permission is hereby granted, free of charge, to any person obtaining a  *
 * copy of this software and associated documentation files (the            *
 * "Software"), to deal in the Software without restriction, including      *
 * without limitation the rights to use, copy, modify, merge, publish,      *
 * distribute, distribute with modifications, sublicense, and/or sell       *
 * copies of the Software, and to permit persons to whom the Software is    *
 * furnished to do so, subject to the following conditions:                 *
 *                                                                          *
 * The above copyright notice and this permission notice shall be included  *
 * in all copies or substantial portions of the Software.                   *
 *                                                                          *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS  *
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF               *
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.   *
 * IN NO EVENT SHALL THE ABOVE COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,   *
 * DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR    *
 * OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR    *
 * THE USE OR OTHER DEALINGS IN THE SOFTWARE.                               *
 *                                                                          *
 * Except as contained in this notice, the name(s) of the above copyright   *
 * holders shall not be used in advertising or otherwise to promote the     *
 * sale, use or other dealings in this Software without prior written       *
 * authorization.                                                           *
 ****************************************************************************/

#include "resources.h"

#define for_each(set,i) for (i = (set).begin(); i != (set).end(); ++i )

Resource & Resource::operator=(const Resource &r) {
	if (this==&r) return *this;
	address = r.address;
	port = r.port;
	return *this;
}



int operator==(const Resource &r1, const Resource &r2) {
	return ((r1.address==r2.address) && (r1.port==r2.port));
}



int operator!=(const Resource &r1, const Resource &r2) {
	return ((r1.address!=r2.address) || (r1.port!=r2.port));
}



ostream& operator << (ostream &out, const Resource &r) {
	out << r.address << " " << r.port;
	return out;
}



Resources::Resources(Options* opt) 
{
	if (pthread_mutex_init(&mutex, NULL)) throw NotEnoughMemory();

	opts = opt;
	parsePoolFile(opts->cfg_file);
}



Resources::~Resources() 
{
	pthread_mutex_destroy(&mutex);
}



int Resources::getResourceList(Address &address, deque<uint16_t> &portList, 
			       bool aDC, bool listen)
{
	Resource r;
	int availability = 0;

	pthread_mutex_lock(&mutex);

	// Trying to allocate the ports with the address desired
	// We first check if all the resources are allowed
	for (uint16_t i=0; i<portList.size(); i++) {
		r = Resource(address, portList[i]);
		if (!isAllowed(r, listen)) {
			pthread_mutex_unlock(&mutex);
			return -2;
		}
	}

	// We check if the resources are all free
	for (uint16_t i=0; i<portList.size(); i++) {
		r = Resource(address, portList[i]);
		if ((availability = isFree(r))!=1) break;
	}

	if (availability>0) {
		// All resources available.
		// Assigning them
	
		for (uint16_t i=0; i<portList.size(); i++) {
			r = Resource(address, portList[i]);
			if (doGetResource(r, false, false, 1, listen)<=0) {
				// Should not occur
				pthread_mutex_unlock(&mutex);
				return -1;
			}
		}

		pthread_mutex_unlock(&mutex);
		return 1;

	} else if (!aDC) {
		pthread_mutex_unlock(&mutex);
		return availability;
	}

	// Trying other addresses
	map<Address, vector<int>*>::iterator i;

	for_each(pool, i) {
		address = i->first;
		availability = 1;

		for (uint16_t i=0; i<portList.size(); i++) {
			r = Resource(address, portList[i]);
			if ((availability = isFree(r))!=1) break;
		}

		if (availability>0) {
			// All resources available.
			// Assigning them
			portList.front();
			for (uint16_t i=0; i<portList.size(); i++) {
				r = Resource(address, portList[i]);
				if (doGetResource(r, false, false, 1, listen)<=0) {
					// Should not occur
					pthread_mutex_unlock(&mutex);
					return -1;
				}
			}

			pthread_mutex_unlock(&mutex);
			return 1;
		}
	}

	// Could not assign the resources
	pthread_mutex_unlock(&mutex);
	return 0;
}



void Resources::ungetResource(Resource &r, unsigned char nbPorts) 
{
	if (!nbPorts) return;

	pthread_mutex_lock(&mutex);
	map<Address, vector<int>*>::iterator i;
	vector<int>* portsUsed;
	vector<int>* newPorts;
	vector<int>::iterator j;

	if ((i=pool.find(r.address))==pool.end()) {
		logmsg(LOG_DEBUG, "Release of a deprecated resource: %s\n",
		       r.address.toChar());
		pthread_mutex_unlock(&mutex);
		return;
	}

	portsUsed = i->second;
	if (!portsUsed) {
		logmsg(LOG_ERR, "Release of an already released resource!\n");
		pthread_mutex_unlock(&mutex);
		return;
	}

	if (r.port==-1) {

		// RSAIP
		pool.erase(i);
		delete portsUsed;
		logmsg(LOG_DEBUG, "Resource released : %s\n",
		       r.address.toChar());

	} else {

		// RSAP-IP
		newPorts = new vector<int>;
		int limit = r.port + nbPorts;
		for_each(*portsUsed, j) {
			if (*j < r.port || *j >= limit) 
				newPorts->push_back(*j);
		}

		if (newPorts->size()==0) {
			delete newPorts;
			newPorts = NULL;
		}

		pool[r.address] = newPorts;
		delete portsUsed;

		if (nbPorts==1) { 
			logmsg(LOG_DEBUG, "Resource released : %s:%d\n", 
			       r.address.toChar(), r.port);
		} else {
			logmsg(LOG_DEBUG, "Resources released : %s:%d-%d\n", 
			       r.address.toChar(), r.port, r.port+nbPorts-1);
		}
	}

	pthread_mutex_unlock(&mutex);
}



void Resources::update(const Address &old, const Address &nw)
{
	pthread_mutex_lock(&mutex);

	vector<int>* lst = NULL;
	map<Address, vector<int>*>::iterator i;

	if ((i=pool.find(old))!=pool.end()) {
		if (i->second) delete i->second;
		pool.erase(i);
	}

	if ((i=pool.find(nw))==pool.end()) 
		pool.insert(make_pair(nw, lst));

	pthread_mutex_unlock(&mutex);
}



int Resources::parsePoolFile(char* filename) 
{
	FILE* file;
	int m, n;
	int stop = 0;
	int valid = 0;       // Nb of valid addresses
	int ignored = 0;     // Nb of ignored addresses
	char line[80];       // A line of the config file
	char* status = NULL;
	char optName[80];    // Option name
	char address[16];
	char address2[16];

	if ((file=fopen(filename, "r"))==NULL) {
		// Pas de fichier de configuration detecte
		logmsg(LOG_WARNING, "%s: file does not exist\n", filename);
		return -1;
	}

	while(!stop) {
		status = fgets(line, 80, file);

		if (status==NULL) stop=1;
		if (line[0]=='#' || line[0]=='\n' || line[1]=='\n') continue;
		
		n = sscanf(line, "%s", optName);
		if (n !=EOF && n!=0 && (strcmp(optName, "POOL")==0)) {
			m = sscanf(line, 
				   "%s %[01234567890.]-%[01234567890.]", 
				   optName, address, address2);
			switch (m) {
			case EOF: case 0:
				stop = -1;
				break;
				
			case 2:
				// INV: We may have here an address
				// Test of the address validity
				if (!checkAddress(address)) { 
					logmsg(LOG_WARNING, 
					       "%s is not a valid IPv4 address. Address ignored.\n",
					       address);
					ignored++;

				} else {

					valid++;
					Address a = address;
					if (pool.find(a)==pool.end())
						pool.insert(make_pair(a, (vector<int>*)NULL));
				}
				break;
				
			case 3:
				if (!checkAddress(address2)) { 
					logmsg(LOG_WARNING, 
					       "%s is not a valid IPv4 address. Address ignored.\n", 
					       address);
					ignored++;
				} else valid+=expandRange(address, address2);
				break;
			default:
				stop = -1;
			}
		} else stop = 1;
	}

	fclose(file);

	if (!pool.size()) pool.insert(make_pair(opts->ext_ip, (vector<int>*)NULL));

	// If the external IP is in the pool, then we have to check to see if it changes sometimes
	if (pool.find(opts->ext_ip)!=pool.end()) {
		opts->check_ip = true;
	} else opts->check_ip = false;

	return stop;
}



int Resources::checkAddress(char* address) 
{
	int a[5];
	int n;
	for (n=0; n<5; n++) a[n]=-1;
	n = sscanf(address, "%d.%d.%d.%d.%d", &a[0], &a[1], &a[2], &a[3], &a[4]);
	if (n!=4) return 0; 

	for (n=0; n<4; n++)
		if ((a[n]<0) || (a[n]>255)) return 0;
 
	return 1; 
}



int Resources::expandRange(char* fA, char* tA) 
{
	int i,j, n;
	int a1[4];
	int a2[4];
	int deb, fin;
	string addrA, addrB, addrC;
	char str[4] = { 0, 0, 0, 0 };

	n=0;     // Nombre d'adresse effectivement ajoutees
	sscanf(fA, "%d.%d.%d.%d", &a1[0], &a1[1], &a1[2], &a1[3]);
	sscanf(tA, "%d.%d.%d.%d", &a2[0], &a2[1], &a2[2], &a2[3]);

	if ((a1[0]!=a2[0]) || (a1[1]!=a2[1])) { // pas plus de 255*255 adresses
		logmsg(LOG_WARNING, "Sorry, range %s-%s is to large.\n", fA, tA);
		logmsg(LOG_WARNING, "Max 255*255 addresses. Range ignored.\n");
	} else {
		if ((a2[2]<a1[2]) || ((a2[2]==a1[2]) && (a2[3]<a1[3])) ) {
			logmsg(LOG_WARNING, "%s-%s : not a valid range.\n", fA, tA);
		} else {
			// INV : a1[0]==a2[0] ^ a1[1]==a2[1] ^ a1[2]<=a2[2]
			//       ^ Vi: ((0<=a1[i]<=255) ^ (0<=a2[i]<=255))
			sprintf(str, "%d", a1[0]);
			addrA = string(str) + ".";
			sprintf(str, "%d", a1[1]);
			addrA = addrA + string(str) + ".";

			for (i=a1[2]; i<=a2[2]; i++) {
				sprintf(str, "%d", i);
				addrB = addrA + string(str) + ".";
				if ((i>0) && (i<255)) {
					if (i==a1[2]) deb=a1[3]; else deb=1;
					if (i==a2[2]) fin=a2[3]; else fin=254;
					for (j=deb; j<=fin; j++) {
						sprintf(str, "%d", j);
						addrC = addrB + string(str);
						pool.insert(make_pair(addrC.c_str(), (vector<int>*)NULL));
						//fprintf(stderr,"%s\n", addrC.toChar());
						n++;
					}
				}
			}

		}
	}

	return n;
}



int Resources::doGetResource(Resource &r, bool aDC, bool pDC, unsigned char nbPorts, bool listen) 
{
	int result;
	bool defined;
	Address a;

	// RSA-IP
	if (nbPorts == 0) {
		result = getResourceRSAIP(r, aDC);
		return result;
	}

	// RSAP-IP
	if (r.address.isDefined()) a = r.address;
	if (pool.find(a)!=pool.end()) defined = true;
	else defined = false;

	// a is 0 if the address is a don't care

	// Trying the address desired...
	unsigned short lLimit = (listen ? opts->listen_low_port : opts->rsap_low_port);
	unsigned short uLimit = (listen ? opts->listen_high_port : opts->rsap_high_port);


	if (r.port<lLimit) {
		if (!pDC) return -2;
		else r.port=lLimit;
	}
	if (r.port>uLimit) {
		if (!pDC) return -2;
		else r.port=uLimit-nbPorts+1;
	}
	if ((result = getResourceRSAPIP(r, nbPorts))>0) return 1;

	//... failed

	if (!aDC) {
		if (!defined) return -1;

		// Case "address required" && "port required"
		if (!pDC) return result;

		// Case "address required" && "port don't care"
		for (r.port=lLimit; r.port+nbPorts<=uLimit; r.port++) {
			if ((result = getResourceRSAPIP(r, nbPorts))>0) break;
		}

		return result;
	}

	// INV: the address is a don't care

	// Trying other addresses
	map<Address, vector<int>*>::iterator i;
	for_each(pool, i) {
		r.address = i->first;
		if (getResourceRSAPIP(r, nbPorts)>0) return 1;
	}

	// INV: here, we failed to assign the ports desired

	// Case "address don't care" && "port required"
	if (!pDC) return 0;

	// Case "address don't care" && "port don't care"
	// We may try to find other ports with another address
	for_each(pool, i) {
		r.address = i->first;
		for (r.port=lLimit; r.port+nbPorts<=uLimit; r.port++) {
			if ((result = getResourceRSAPIP(r, nbPorts))>0) break;
		}

		if (result==1) return 1;
	}

	return 0; // Impossible to assign the resource
}



int Resources::getResourceRSAIP(Resource &r, bool aDC) 
{
	int result = 1;
	int available = 1;
	Address a;

	vector<int>* ports = NULL;
	map<Address, vector<int>*>::iterator i;

	if (opts->rsa_method==0) return -1;

	if (r.address.isDefined()) a = r.address;
	
	// First simple case : no address available
	if (!pool.size()) return -1;
	
	if ((i=pool.find(a))==pool.end()) {
		available = 0;
		if (!aDC) return -1;
	} else {
		ports = i->second;
		if (ports!=NULL) available = 0;
	}

	if (opts->rsa_method==2 && a==opts->ext_ip) available=0;

	if (!available) {
		// We couldn't find the desired address, or the resource
		// is used. So we assign the first unused address.
		for_each(pool, i) {
			ports = i->second;
			if (ports==NULL) {
				if (opts->rsa_method==2 && a==opts->ext_ip) continue;
				else break; // Found !
			}
		}

		if (ports!=NULL) return 0; // We couldn't find any free address

		result = 2;
	}

	// INV: ports is NULL, the resource is not used
	ports = new vector<int>;
	ports->push_back(-1);      // -1 = all the ports
	pool[a] = ports;

	r.address = a;
	r.port = -1;
	logmsg(LOG_INFO, "Resource taken : %s\n", r.address.toChar());

	return result;
}


int Resources::getResourceRSAPIP(Resource &r, unsigned char nbPorts) 
{
	Address a;
	vector<int>* ports;
	map<Address, vector<int>*>::iterator i;

	if (r.address.isDefined()) a = r.address;
	
	// Case 1: no address defined in config file
	if (!pool.size() || r.port<1 || (r.port+nbPorts > 65535)) return -1;
	
	// Case 2: address not defined in config file
	if ((i=pool.find(a))==pool.end()) return -1;
	ports = i->second;

	// Case 3 : normal case, address defined
	if (!ports) {

		// No ports assigned yet for this address
		ports = new vector<int>;
		for (int j=r.port; j<nbPorts+r.port; j++)
			ports->push_back(j);

		if (nbPorts==1) {
			logmsg(LOG_INFO, "Resource taken : %s:%d\n", 
			       r.address.toChar(), r.port);
		} else {
			logmsg(LOG_INFO, "Resources taken : %s:%d-%d\n", 
			       r.address.toChar(), r.port, r.port+nbPorts-1);
		}

		pool[a] = ports;

	} else {

		// At least one port has been assigned for this address
		// Checking that the requested address is free
		if ((*ports)[0]==-1) return 0; // All ports already used for this address

		// Checking that all the ports required are free
		for (unsigned int j=0; j<ports->size(); j++) {
			if ((*ports)[j]>=r.port && (*ports)[j]<r.port+nbPorts)
				return 0;
		}

		// OK, all the ports are free. Assigning them.
		for (int i=r.port; i<nbPorts+r.port; i++)
			ports->push_back(i);

		if (nbPorts==1) {
			logmsg(LOG_INFO, "Resource taken : %s:%d\n", 
			       r.address.toChar(), r.port);
		} else {
			logmsg(LOG_INFO, "Resources taken : %s:%d-%d\n", 
			       r.address.toChar(), r.port, r.port+nbPorts-1);
		}
	}

	return 1;
}



int Resources::isFree(Resource &r) 
{
	Address a;
	vector<int>* ports;
	map<Address, vector<int>*>::iterator i;

	if (r.address.isDefined()) a = r.address;
	if ((i=pool.find(a))==pool.end()) return -1;

	// ports is the list of ports used for this address
	if ((ports = i->second)!=NULL) {
		if ((*ports)[0]==-1) return 0; // The whole address is used

		for (unsigned int j=0; j<ports->size(); j++) 
			if ((*ports)[j]==r.port) return 0;
	} 
	return 1;
}



bool Resources::areAllowed(Resource& r, unsigned char nbPorts, bool listen)
{
	if (listen) {
		for (uint16_t p=r.port; p<=r.port+nbPorts; p++) {
			if (r.port < opts->listen_low_port || r.port > opts->listen_high_port)
				return false;
		}

	} else {
		for (uint16_t p=r.port; p<=r.port+nbPorts; p++) {
			if (r.port < opts->rsap_low_port || r.port > opts->rsap_high_port)
				return false;
		}
	}

	return true;
}
