//***************************************************************************
//* resources.h
//*
//* description    : Pool of RSIP resources
//* copyright      : (C) 2001 Cedric de Launois
//*                  Universite Catholique de Louvain, Belgium
//* begin          : Sept 24 2001
//* email          : cedric@delaunois.org
//***************************************************************************

/****************************************************************************
 * Copyright (c) 1998 Free Software Foundation, Inc.                        *
 *                                                                          *
 * Permission is hereby granted, free of charge, to any person obtaining a  *
 * copy of this software and associated documentation files (the            *
 * "Software"), to deal in the Software without restriction, including      *
 * without limitation the rights to use, copy, modify, merge, publish,      *
 * distribute, distribute with modifications, sublicense, and/or sell       *
 * copies of the Software, and to permit persons to whom the Software is    *
 * furnished to do so, subject to the following conditions:                 *
 *                                                                          *
 * The above copyright notice and this permission notice shall be included  *
 * in all copies or substantial portions of the Software.                   *
 *                                                                          *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS  *
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF               *
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.   *
 * IN NO EVENT SHALL THE ABOVE COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,   *
 * DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR    *
 * OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR    *
 * THE USE OR OTHER DEALINGS IN THE SOFTWARE.                               *
 *                                                                          *
 * Except as contained in this notice, the name(s) of the above copyright   *
 * holders shall not be used in advertising or otherwise to promote the     *
 * sale, use or other dealings in this Software without prior written       *
 * authorization.                                                           *
 ****************************************************************************/

class Resources;
class Resource;

#ifndef _RESOURCES_H_
#define _RESOURCES_H_

#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <fcntl.h>
#include <pthread.h>
#include <time.h>
#include <string>
#include <deque>
#include <vector>
#include <map>
#include "address.h"

#define LOCK(lock,statement) do { pthread_mutex_lock(&lock); statement pthread_mutex_unlock(&lock); } while (0);

// A RSIP resource
class Resource {

	friend int operator==(const Resource &, const Resource &);
	friend int operator!=(const Resource &, const Resource &);
	friend ostream& operator << (ostream &, const Resource &);

 public:
	Resource() : port(0) {}
	Resource(Address& a, int p) : address(a), port(p) {}
	~Resource() {}
	Resource(const Resource& r) : address(r.address), port(r.port) {}
	Resource & operator=(const Resource &);
	
	Address address;
	int     port;
};


#include "common.h"

// The pool of RSIP resources
class Resources {
 public:


	Resources(Options* opt);
	// Constructor
	// PRE : /
	// POST: initialize the address pool by reading the address written
	//       in file opt->cfg_file

	
	~Resources();
	// Destructor
	
	
	inline int getResource(Resource &r, bool aDC, bool pDC,
			       unsigned char nbPorts, bool listen) {
		int res; 
		LOCK(mutex,res=doGetResource(r, aDC, pDC, nbPorts, listen);); 
		return res;
	}
	// PRE : - r: resource desired or required (depending of aDC or pDC)
	//       - aDC: true for "don't care" address
	//       - pDC: true for "don't care" port
	//       - nbPorts: - number of consecutive ports starting at r.port
	//                    for the address r.address (max 256 ports)
	//                  - 0 to request all the ports for this addres
	//       - listen : true to request a listen port
	// POST: returns  1 if the resource requested was succesfully assigned
	//                2 if another resource was successfully assigned
	//                  (only if aDC or pDC is set)
	//                0 if resource is in use
	//               -1 if the resource is not available for RSIP
	//               -2 if resource is not allowed


	int getResourceList(Address &address, deque<uint16_t> &portList,
			    bool aDC, bool listen);
	// Assign multiple ports of the same address
	// PRE : - portList initialized with the resources desired
	//       - address is the address to be used with the ports specified
	//       - aDC : true for a "Don't care" address
	//       - listen : true to request a listen port
	// POST: returns  1 if the resources with the address desired
	//                  were succesfully assigned
	//                2 if another address was chosen for the ports
	//                0 if the resources were not all available
	//               -1 if one of the resource is not available for RSIP
	//               -2 if resource is not allowed


	void ungetResource(Resource &r, unsigned char nbPorts);
	// PRE : /
	// POST: If the resource is a RSA-IP resource (i.e. r.port == -1),
	//         then returns the IP to the pool
	//       Else returns resources <r.address, r.port> up to 
	//         <r.address, r.port + nbPorts - 1> to the pool
	//       If nbPorts is 0, then no resource is returned to the pool


	inline int isEmpty() { int r; LOCK(mutex,r=!pool.size();); return r; }
	// PRE : /
	// POST: returns 0 if no address available
	//               1 otherwise


	void update(const Address &old, const Address &nw);
	// PRE : old exist in the pool
	// POST: new has replaced old in the pool

	
	inline unsigned int size() { int r; LOCK(mutex,r=pool.size();); return r; }
	// PRE : /
	// POST: renvoie le nombre d'adresses completes encore disponibles.
	

 private:
	
	int parsePoolFile(char* filename);
	// Parsing du fichier contenant les ressources.
	// PRE : filename est un nom de fichier
	// POST: pool est remplie avec les ressources valides contenues dans
	//       le fichier filename
	//       -1 en cas d'erreur, 1 sinon

	
	int checkAddress(char* address);
	// Verifie si l'adresse est valide
	// PRE : /
	// POST: renvoie 1 si address est de la forme a.b.c.d
	//                 avec 0 <= a, b, c, d <= 255
	//       renvoie 0 sinon

	
	int expandRange(char* fromAddr, char* toAddr);
	// PRE : fromAddr et toAddr sont des adresses IPv4 VALIDES
	// POST: les adresses comprises entre fromAddr et toAddr ont
	//       ete ajoute a la liste des adresses disponibles.
	//       Renvoie le nombre d'adresses ajoutees. 0 si erreur.

	
	int doGetResource(Resource &r, bool aDC, bool pDC, unsigned char nbPorts, bool listen);
	// PRE : - r: resource desired or required (depending of aDC or pDC)
	//       - aDC: true for "don't care" address
	//       - pDC: true for "don't care" port
	//       - nbPorts: - number of consecutive ports starting at r.port
	//                    for the address r.address (max 256 ports)
	//                  - 0 to request all the ports for this addres
	//       - listen : true to request a listen port
	// POST: returns -1 if the resource is not available for RSIP
	//                0 if resource not available
	//                1 if the resource requested was succesfully assigned
	//                2 if another resource was successfully assigned
	//                 (only if aDC or pDC is set)

	int getResourceRSAIP(Resource &r, bool aDC);
	// PRE : - r is the desired or requested resource (depending of aDC)
	//         aDC=true for a "Don't care" for the address
	// POST: - r is the assigned resource
	//       - returns -1 if the resource is not available for RSIP
	//                  0 if the resource is not available
	//                  1 if the requested resources has been assigned
	//                  2 if another resource has been assigned

	
	int getResourceRSAPIP(Resource &r, unsigned char nbPorts);
	// PRE : - r is the requested resource
	//       - nbPorts is the number of consecutive ports to assign,
	//         beginning at r.port
	// POST: returns -1 if the resource is not available for RSIP
	//                0 is the resource could not be assign entirely
	//                1 if success


	int isFree(Resource& r);
	// PRE : r is the resource to test if free
	// POST: returns -1 if the resource is not available for RSIP
	//                1 if r is free
	//                0 if r is not free

	
	bool areAllowed(Resource& r, unsigned char nbPorts, bool listen);
	// Test if the ports r.port to r.port+nbPorts are allowed by the options
	// PRE : /
	// POST: IF (listen==TRUE) return
	//           TRUE if (listenLowerLimit <= port <= listenUpperLimit)
	//       IF (listen==FALSE) return
	//           TRUE if (lowerLimit <= port <= upperLimit)
	//       WHERE 'port' is any port between r.port and r.port+nbPorts
	//       Otherwise return FALSE
	//       r does not change


	inline bool isAllowed(Resource& r, bool listen) 
		{ return areAllowed(r, 1, listen); }
	
	map<Address, vector<int>*> pool; 
	// Table of resources
	// <address, <list of ports used for this address>*>

	
	pthread_mutex_t mutex;
	Options* opts;
};

#endif

