//***************************************************************************
//* RSIPClient.cpp
//*
//* description    : class RSIPClient
//*                  State and resources of a RSIP client
//*
//* copyright      : (C) 2001 Cedric de Launois
//*                  Universite Catholique de Louvain, Belgium
//* begin          : Sept 24 2001
//* email          : cedric@delaunois.org
//***************************************************************************

/****************************************************************************
 * Copyright (c) 1998 Free Software Foundation, Inc.                        *
 *                                                                          *
 * Permission is hereby granted, free of charge, to any person obtaining a  *
 * copy of this software and associated documentation files (the            *
 * "Software"), to deal in the Software without restriction, including      *
 * without limitation the rights to use, copy, modify, merge, publish,      *
 * distribute, distribute with modifications, sublicense, and/or sell       *
 * copies of the Software, and to permit persons to whom the Software is    *
 * furnished to do so, subject to the following conditions:                 *
 *                                                                          *
 * The above copyright notice and this permission notice shall be included  *
 * in all copies or substantial portions of the Software.                   *
 *                                                                          *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS  *
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF               *
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.   *
 * IN NO EVENT SHALL THE ABOVE COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,   *
 * DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR    *
 * OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR    *
 * THE USE OR OTHER DEALINGS IN THE SOFTWARE.                               *
 *                                                                          *
 * Except as contained in this notice, the name(s) of the above copyright   *
 * holders shall not be used in advertising or otherwise to promote the     *
 * sale, use or other dealings in this Software without prior written       *
 * authorization.                                                           *
 ****************************************************************************/

#include "rsipClient.h"

#define for_each(map,i) for (i = (map).begin(); i != (map).end(); ++i )

uint32_t RSIPClient::nextCltID = 1;
map<Address, RSIPClient*> RSIPClient::clients; // Table (client IP, client state)
pthread_mutex_t RSIPClient::lockID = PTHREAD_MUTEX_INITIALIZER;
pthread_mutex_t RSIPClient::mutexdb = PTHREAD_MUTEX_INITIALIZER;


RSIPClient::RSIPClient(Address &addr, NetworkEndPoint* n, Options* o)
{
	assert(n!=NULL);

	state             = STATE_UNREGISTERED;
	clientID          = 0;
	registerDeadline  = 0;
	ports             = 0;
	IPs               = 0;
	id                = 0;
	nextReservedPort  = 0;
	address           = addr;
	protocol          = n->getProto();
	nep               = n;
	lastMsgTx         = NULL;
	lastMsgRx         = NULL;
	opts              = o;

	if (pthread_mutex_init(&mutex, NULL)) throw NotEnoughMemory();

	clients.insert(make_pair(address, this));
}



RSIPClient::~RSIPClient() 
{
	if (lastMsgRx) delete lastMsgRx;
	lastMsgRx = NULL;
		
	if (lastMsgTx) delete lastMsgTx;
	lastMsgTx = NULL;

	clients.erase(clients.find(address));

	if (pthread_mutex_destroy(&mutex)!=0) throw Exception();
}



//******************** Operations on the client itseltf ******************

uint32_t RSIPClient::doRegister()
{
	if (state != STATE_UNREGISTERED) return 0;

	// Assigns an ID to the client
	clientID = getID();

	RSIPClient::TimerParam *p = new RSIPClient::TimerParam;
	p->address  = address;
	p->clientID = clientID;

	// Initialization of the register lease time
	registerDeadline = time(0) + opts->initial_register_lease_time;
	tmr->schedule((void*(*)(void*))checkRegistering, 
		      (void*)p, 
		      opts->initial_register_lease_time);

	// Update the state of the client
	state = STATE_REGISTERED;

	return clientID;
}



void RSIPClient::deregister(bool warn)
{
	if (state == STATE_UNREGISTERED) return;

	freeAll(warn);

	if (warn) sendDeRegResp();

	// Scheduling the removing of the client if
	// the client did not disconnect after 5 sec

	RSIPClient::TimerParam *p = new RSIPClient::TimerParam;
	p->address  = address;
	p->clientID = clientID;

	tmr->schedule((void*(*)(void*))removeClient, 
		      (void*)p,5);

	state            = STATE_UNREGISTERED;
	registerDeadline = 0;
}



uint32_t RSIPClient::extend(uint32_t bindID, uint32_t ext) 
{
	Bind* b;
	map<uint32_t, Bind*>::iterator i;

	if ((i=binds.find(bindID))!=binds.end()) {

		b = i->second;
		deadlines.erase(deadlines.find(b->deadline));

		// Extension of the lease time
		time_t currentLease = b->deadline - time(0);
		if (currentLease+ext > opts->max_lease_time) 
			ext = opts->max_lease_time - currentLease;

		b->deadline += ext;

		// We just increase LT of 1sec if another deadline
		// exists at the same time. Should be rare.
		while (deadlines.find(b->deadline)!=deadlines.end()) {
			b->deadline++;
			ext++;
		}

		// Calculating the next deadline
		deadlines.insert(make_pair(b->deadline, bindID));

		RSIPClient::TimerParam *p = new RSIPClient::TimerParam;
		p->address  = address;
		p->clientID = clientID;
		tmr->schedule((void*(*)(void*))checkAssignments,
			      (void*)p, 
			      b->deadline-time(0));


		// Adjust the register lease time if needed
		if (registerDeadline <= b->deadline) 
			registerDeadline = b->deadline;

		p = new RSIPClient::TimerParam;
		p->address  = address;
		p->clientID = clientID;
		tmr->schedule((void*(*)(void*))checkRegistering, 
			      (void*)p, 
			      registerDeadline-time(0));

		return (currentLease + ext);

	} else return 0;

}



bool RSIPClient::free(uint32_t bindID, bool warn) 
{
	if (state != STATE_ASSIGNED) return false;

	Bind* b;
	b = del(bindID); // This update the state

	if (b) {
		if (warn) sendFreeResp(b->bindID);
		if (b->tunnel) delete b->tunnel;
		delete b;
		return true;

	} else return false;
}



void RSIPClient::freeAll(bool warn) 
{
	if (state != STATE_ASSIGNED) return;

	Bind* b;

	// Releasing all resources
	while ((b = pop())) {
		if (warn) sendFreeResp(b->bindID);

		// Destruction of the binds
		if (b->tunnel) delete b->tunnel;
		delete b;
	}
	
	state = STATE_REGISTERED;
}



int RSIPClient::send(RSIPMessage* &msg, NetworkEndPoint* ep)
{
	int result;

	msg->setAddress(address);

	result = msg->send(ep);

	if (opts->dump) msg->dump();
	if (lastMsgTx && msg!=lastMsgTx)
		delete lastMsgTx;

	lastMsgTx = msg;
	msg = NULL;

	return result;
}



bool RSIPClient::updateLastMsg(RSIPMessage* &msg)
{
	if (lastMsgRx && protocol==UDP) {
		if (*lastMsgRx == *msg) {
			delete msg;
			msg = lastMsgTx;
			return true;
		}
	}

	if (lastMsgRx) delete lastMsgRx;
	lastMsgRx = msg;
	return false;
}



//******************* Static operations on all clients ****************

void RSIPClient::invalid(NetworkEndPoint* n)
{
	map<Address, RSIPClient*>::iterator i;
	
	lockdb();
	if (!clients.size()) {
		unlockdb();
		return;
	}

	for_each(clients, i) {
		i->second->lock();
		if (i->second->nep==n) i->second->nep=NULL;
		i->second->unlock();
	}
	unlockdb();
}



string RSIPClient::getInfo() 
{
	string result;
	char temp[300];
	result = "Client           Leased Address  STATE        ID    BID   LT    RLT   TUNNEL    PROTO  PORTS\n";
	map<Address, RSIPClient*>::iterator i;
	map<uint32_t, Bind*>::iterator j;
	Address k;
	RSIPClient* cl;
	Resource r;
	Bind* bind;

	lockdb();

	for_each(clients, i) {

		k = i->first;
		cl = i->second;

		cl->lock();
		
		switch(cl->state) {
		case STATE_ASSIGNED:

			for_each(cl->binds, j) {
				bind = j->second;
				string tnlName;
				int firstport = bind->getFirst().port;
				if (bind->tunnel) tnlName = bind->tunnel->getName();
				snprintf(temp, 200,
					 "%-15s %15s  ASSIGNED     %.5i %.5i %.5ld %.5ld %-9s %-6s %i%s(%i)\n",
					 cl->address.toChar(),
					 bind->getFirst().address.toChar(), 
					 cl->clientID, 
					 bind->bindID, 
					 bind->deadline-time(0), 
					 (time_t)cl->registerDeadline-time(0),
					 tnlName.c_str(),
					 (cl->protocol==UDP)?"UDP":"TCP",
					 (firstport>0)?firstport:0,
					 (bind->nbPorts()>1)?"...":"",
					 (firstport>0)?bind->nbPorts():65535
					 );
				result += temp;
			}
			break;
		case STATE_REGISTERED:
			sprintf(temp, "%-15s %15s  REGISTERED   %.5d none  %.5d %.5ld %-9s %-6s\n", 
				cl->address.toChar(),
				"none", 
				cl->clientID, 
				0, 
				(time_t)cl->registerDeadline-time(0), 
				"",
				(cl->protocol==UDP)?"UDP":"TCP"
				);
			result += temp;
			break;
		case STATE_UNREGISTERED:
			sprintf(temp, "%-15s %15s  UNREGISTERED none %18s %9s %-6s\n", 
				cl->address.toChar(),
				"", 
				"",
				"",
				(cl->protocol==UDP)?"UDP":"TCP"
				);
			result += temp;
			break;
		}
		
		cl->unlock();
		
	} // end for

	unlockdb();

	return result;
}



void RSIPClient::disconnectAll(bool warn) 
{
	map<Address, RSIPClient*>::iterator i;

	lockdb();

	for_each(clients, i) {
		i->second->deregister(warn);
		delete i->second;
	}

	clients.clear();

	unlockdb();
}


void RSIPClient::freeAll(const Resource &r, bool warn) 
{
	lockdb();

	map<Address, RSIPClient*>::iterator i;

	for_each(clients, i) {
		RSIPClient* cl = i->second;
		deque<Bind*> deprecatedBinds;
		map<uint32_t, Bind*>::iterator j;

		cl->lock();

		// Search for all client's deprecated resources
		for_each(cl->binds, j) {
			Bind* bind = j->second;
			if (bind->getFirst().address == r.address) {
				deprecatedBinds.push_back(bind);
			}
		}

		// Free them
		while (deprecatedBinds.size()) {
			cl->free((deprecatedBinds.front())->bindID, warn);
			deprecatedBinds.pop_front();
		}

		cl->unlock();
	}

	unlockdb();
}



//************************* Private Functions *********************

uint32_t RSIPClient::getID() {
	pthread_mutex_lock(&lockID);
	uint32_t result = nextCltID;
	nextCltID++;
	if (!nextCltID) nextCltID = 1;
	pthread_mutex_unlock(&lockID);
	return result;
}



void RSIPClient::sendFreeResp(uint32_t bindID)
{
	struct RSIPMsgParams msgp;
	RSIPMessage* msg;

	logmsg(LOG_INFO, "Freeing bind %i, client %s\n", bindID, address.toChar());

	msgp.clientID = clientID;
	msgp.bindID = bindID;
	msgp.flags = 0;

	if (protocol==UDP) {
		msgp.counter = 0;
		msgp.flags |= FLAG_COUNTER;
	}

	if (!nep) {
		logmsg(LOG_WARNING, "Cannot warn client %s (connection is closed)!\n", address.toChar());
		return;
	}

	msg = new RSIPMessage(FREE_RESPONSE);
	msg->forge(msgp);
	msg->setAddress(address);

	msg->send(nep);

	if (opts->dump) msg->dump();
	delete msg;
}



void RSIPClient::sendDeRegResp()
{
	struct RSIPMsgParams msgp;
	RSIPMessage* msg;

	logmsg(LOG_INFO, "Deregistering client %s\n", address.toChar());

	msgp.clientID = clientID;
	msgp.flags = 0;

	if (protocol==UDP) {
		msgp.counter = 0;
		msgp.flags |= FLAG_COUNTER;
	}

	if (!nep) {
		logmsg(LOG_WARNING, "Cannot warn client %s (connection is closed)!\n", address.toChar());
		return;
	}

	msg = new RSIPMessage(DEREGISTER_RESPONSE);
	msg->forge(msgp);
	msg->setAddress(address);

	msg->send(nep);

	if (opt->dump) msg->dump();
	delete msg;
}



void removeClient(RSIPClient::TimerParam* p)
{
	RSIPClient* cl;

	// Retrieving the client state (if exist)
	RSIPClient::lockdb();
	if (!RSIPClient::search(p->address, cl)) {
		RSIPClient::unlockdb();
		delete p;
		return;
	}

	cl->lock();

	// Removing the client if in state UNREGISTERED
	if (cl->clientID==p->clientID && cl->state==STATE_UNREGISTERED) {
		logmsg(LOG_INFO, "Closing connection with client %s\n", 
		       p->address.toChar());
		cl->unlock();
		delete cl;
	} else cl->unlock();
	RSIPClient::unlockdb();

	delete p;
}



void checkRegistering(RSIPClient::TimerParam* p)
{
	RSIPClient* cl;

	// Retrieving the client state (if exist)
	RSIPClient::lockdb();
	if (!RSIPClient::search(p->address, cl)) {
		RSIPClient::unlockdb();
		delete p;
		return;
	}

	cl->lock();
	RSIPClient::unlockdb();

	// Check that no time out occurred for the register lease time
	if (cl->clientID==p->clientID && time(0)>=cl->registerDeadline) {

		if (cl->state != STATE_UNREGISTERED) {
			logmsg(LOG_INFO, "Time out for client %s\n", p->address.toChar());
			cl->deregister(true);
		}

	}

	delete p;
	cl->unlock();
}



void checkAssignments(RSIPClient::TimerParam* p)
{
	RSIPClient* cl;
	map<time_t, uint32_t>::iterator i;

	// Retrieving the client state (if exist)
	RSIPClient::lockdb();
	if (!RSIPClient::search(p->address, cl)) {
		RSIPClient::unlockdb();
		delete p;
		return;
	}

	cl->lock();
	RSIPClient::unlockdb();

	if (cl->clientID!=p->clientID) {
		// Check that no time out occurred for any bind
		while ((i=cl->deadlines.begin()) != cl->deadlines.end()) {
			if (time(0) >= i->first) {
				logmsg(LOG_INFO, 
				       "Time out for bind %i, client %s\n", 
				       i->second, p->address.toChar());
				cl->free(i->second, true);
			} else break;
		}
	}

	delete p;
	cl->unlock();
}



//********************* Operations on client's binds *********************

Bind* RSIPClient::search(uint32_t bindID) 
{
	map<uint32_t, Bind*>::iterator i;

	if ((i=binds.find(bindID))!=binds.end()) return i->second;
	else return NULL;
}



uint32_t RSIPClient::add(Bind* b) 
{
	id++;
	if (id==0) id++; // To avoid a NULL bind id

	// Searching for a free ID (should be rare since: id++)
	while (binds.find(id)!=binds.end()) id++;

	// Ok, we found a free ID.
	b->bindID = id;

	// Now we adjust the deadline so that no more
	// than one deadline occurs for the same client
	// at the same time

	// We just increase the deadline of 1sec if another deadline
	// exists at the same time
	while (deadlines.find(b->deadline)!=deadlines.end()) 
		b->deadline++;

	// OK. Now inserting the new bind
	binds.insert(make_pair(id, b));

	// Updating the number of ports used by the client
	ports += b->nbPorts();
	IPs += b->nbIPs();

	// Inserting the new deadline
	deadlines.insert(make_pair(b->deadline, id));

	RSIPClient::TimerParam *p = new RSIPClient::TimerParam;
	p->address  = address;
	p->clientID = clientID;
	tmr->schedule((void*(*)(void*))checkAssignments,
		      (void*)p, 
		      b->deadline-time(0));

	// Adjust the register lease time if needed
	if (registerDeadline <= b->deadline) 
		registerDeadline = b->deadline;

	p = new RSIPClient::TimerParam;
	p->address  = address;
	p->clientID = clientID;
	tmr->schedule((void*(*)(void*))checkRegistering, 
		      (void*)p, 
		      registerDeadline-time(0));

	state = STATE_ASSIGNED;

	return id;
}



Bind* RSIPClient::del(uint32_t bindID) 
{
	Bind* b;
	map<uint32_t, Bind*>::iterator i;

	if ((i=binds.find(bindID))!=binds.end()) {
		b = i->second;
		deadlines.erase(deadlines.find(b->deadline));
		binds.erase(i);

		// Updating the number of ip and ports used by the client
		ports -= b->nbPorts();
		IPs -= b->nbIPs();

		// Changing state if needed
		if (binds.size()==0) state = STATE_REGISTERED;

		return b;
	}

	return NULL;
}



Bind* RSIPClient::pop() 
{
	Bind* b;
	map<uint32_t, Bind*>::iterator i;

	if ((i=binds.begin())!=binds.end()) {
		b = i->second;
		deadlines.erase(deadlines.find(b->deadline));
		binds.erase(i);

		// Updating the number of ip and ports used by the client
		ports -= b->nbPorts();
		IPs -= b->nbIPs();

		// Changing state if needed
		if (binds.size()==0) state = STATE_REGISTERED;

		return b;
	} else return NULL;
}



