//***************************************************************************
//* RSIPClient.h
//*
//* description    : class RSIPClient
//*                  State and resources of a RSIP client
//*
//* copyright      : (C) 2001 Cedric de Launois
//*                  Universite Catholique de Louvain, Belgium
//* begin          : Sept 24 2001
//* email          : cedric@delaunois.org
//***************************************************************************

/****************************************************************************
 * Copyright (c) 1998 Free Software Foundation, Inc.                        *
 *                                                                          *
 * Permission is hereby granted, free of charge, to any person obtaining a  *
 * copy of this software and associated documentation files (the            *
 * "Software"), to deal in the Software without restriction, including      *
 * without limitation the rights to use, copy, modify, merge, publish,      *
 * distribute, distribute with modifications, sublicense, and/or sell       *
 * copies of the Software, and to permit persons to whom the Software is    *
 * furnished to do so, subject to the following conditions:                 *
 *                                                                          *
 * The above copyright notice and this permission notice shall be included  *
 * in all copies or substantial portions of the Software.                   *
 *                                                                          *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS  *
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF               *
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.   *
 * IN NO EVENT SHALL THE ABOVE COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,   *
 * DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR    *
 * OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR    *
 * THE USE OR OTHER DEALINGS IN THE SOFTWARE.                               *
 *                                                                          *
 * Except as contained in this notice, the name(s) of the above copyright   *
 * holders shall not be used in advertising or otherwise to promote the     *
 * sale, use or other dealings in this Software without prior written       *
 * authorization.                                                           *
 ****************************************************************************/

class RSIPClient;

#ifndef RSIP_CLIENT_H
#define RSIP_CLIENT_H

#include <netinet/in.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <pthread.h>
#include <time.h>
#include <string>
#include <deque>
#include <map>
#include "proto.h"
#include "resources.h"
#include "bind.h"
#include "tunnel.h"
#include "exceptions.h"
#include "rsipMessage.h"
#include "netEndPoint.h"
#include "address.h"
#include "common.h"

#define STATE_UNREGISTERED          0
#define STATE_REGISTERED            1
#define STATE_ASSIGNED              2

class RSIPClient {
 public:

	RSIPClient(Address &addr, NetworkEndPoint* n, Options* opts);
	// Constructor
	// PRE : - addr is the IP address of the client
	//       - nep is a socket to the client
	// POST: - client initialized
	//       - client inserted in the list of all clients


	~RSIPClient();
	// Destructor
	// PRE : /
	// POST:  - client removed from the list of clients, if known
	//        - nep NOT closed


	//********** Operations on the client itseltf **********

	uint32_t doRegister();
	// PRE : - client in the UNREGISTERED state
	// POST: - client in the REGISTERED state
	//       - returns the new client ID of the client


	void deregister(bool warn);
	// PRE : the nep of the client is still valid if warn is true
	// POST: - the client is in the UNREGISTERED state
	//       - all resources have been released by the call of the
	//         function freeAll with warn as parameter
	//       - A DEREGISTER_RESPONSE has been sent 
	//         if warn is true
	//       - the client will be deleted automatically 5sec later 
	//         if it was not already deleted (function removeClient())


	inline uint32_t assign(Bind* b) { return add(b); }
	// Add a bind to the list of client's binds
	// PRE : - b configured except the bindID
	//       - the lease time of b is the lease time requested
	// POST: - returns the bind ID for the new bind
	//       - the lease time of the b has been changed to the real 
	//         lease time assigned
	//       - the bindID of b has been set


	inline bool checkBind(uint32_t bindID) { return (binds.find(bindID)!=binds.end()); }
	// Check that the bind ID exist


	uint32_t extend(uint32_t BindID, uint32_t leaseExtension);
	// Extend the leaseTime of the bind
	// PRE : BindID is the bind ID of an existing bind
	//       leaseExtension is the extension requested
	// POST: returns the updated lease time,
	//       0 if the bind does not exist

	
	bool free(uint32_t bindID, bool warn);
	// Free the bind with this bindID
	// PRE : the nep of the client is still valid if warn is true
	// POST: - the bind is freed and returned to the pool if it existed.
	//       - returns true if the bind existed
	//                 false otherwise
	//       - A FREE_RESPONSE has been sent if warn is true


	void freeAll(bool warn);
	// Remove all the binds of this client
	// PRE : the nep of the client is still valid if warn is true
	// POST: - All bind are freed and returned to the pool
	//         The client's state is STATE_REGISTERED
	//       - A FREE_RESPONSE has been sent 
	//         for each bind if warn is true
	

	int send(RSIPMessage* &msg, NetworkEndPoint* nep);
	// PRE : msg not NULL
	// POST: - sends the message to the client through the nep
	//       - last message transmitted updated
	//       - msg is NULL ! This is because the msg is now
	//         internally stored as the last message received
	//         and this prevent the message to be deleted.
	//       - returns 1 if the msg has been properly sent
	//                 0 otherwise


	bool updateLastMsg(RSIPMessage* &msg);
	// Check for a duplicate message
	// PRE : msg is the new message received which must be checked
	// POST: - if the message has already been received
	//         and if the protocol used is UDP, THEN
	//         - msg is the last message transmitted
	//         - returns true
	//       - oterwise returns false and msg is unchanged
	//         and the last message received has been updated


	inline void setNextReservedPort(unsigned short port) { nextReservedPort=port; };
	inline unsigned short getNextReservedPort() { return nextReservedPort; };

	inline int lock() { return pthread_mutex_lock(&mutex); }
	inline int trylock() { return pthread_mutex_trylock(&mutex); }
	inline int unlock() { return pthread_mutex_unlock(&mutex); }
	

	//************************ Accessors ***********************


	inline int getState() { return state; }
	// PRE : /
	// POST: the state of the client


	inline uint32_t getClientID() { return clientID; }
	// PRE : /
	// POST: the client ID of the client


	inline proto getProto() { return protocol; }
	// PRE : /
	// POST: the protocol used by the client


	inline Address getAddress() { return address; }
	// PRE : /
	// POST: the address of the client


	inline unsigned int nbBinds() { return binds.size(); }
	// Return the number of binds for this client
	// PRE : /
	// POST: a positive value : the number of binds
	

	inline unsigned int nbPorts() { return ports; }
	// Return the number of binds for this client
	// PRE : /
	// POST: a positive value : the number of binds
	

	inline unsigned int nbIPs() { return IPs; }


	//******************* Static operations on all clients ****************

	static inline int search(Address &id, RSIPClient* &inf) { 
		map<Address, RSIPClient*>::iterator i;
		if ((i = clients.find(id)) != clients.end()) {
			inf = i->second;
			return 1;
		} else return 0;
	}
	// Search atomically for a client
	// PRE : id is the address of the client
	// POST: - If the client is known, returns 1 and inf is the client 
	//         associated with this address
	//       - Otherwise returns 0


	static inline void lockdb() { pthread_mutex_lock(&mutexdb); }
	static inline void unlockdb() { pthread_mutex_unlock(&mutexdb); }


	static void invalid(NetworkEndPoint* nep);
	// Invalid the nep for all clients using it
	// NOTE: call lockdb()

	static string getInfo();
	// Returns informations about all the clients of the database
	// PRE  : /
	// POST : the informations
	// NOTE: call lockdb()


	static void disconnectAll(bool warn);
	// Deregister and delete all the clients
	// NOTE: call lockdb()


	static void freeAll(const Resource &r, bool warn);
	// Remove all binds of all clients corresponding to the address
	// of this resource. Useful when an IP of the pool has changed.
	// PRE : /
	// POST: - All bind corresponding to this resource are freed 
	//         and returned to the pool
	//       - A FREE_RESPONSE has been sent 
	//         for each bind if warn is true
	// NOTE: call lockdb()


	static inline unsigned int size() { return clients.size(); }
	// Returns the total number of clients


 private:

	map<uint32_t, Bind*> binds;               // Table <BindID, Bind>
	map<time_t, uint32_t> deadlines;          // Table <deadline, BindID>
	int state;                                // State of the client
	Address address;                          // IP address of the client
	uint32_t clientID;                        // Client ID
	time_t registerDeadline;                  // Absolute Deadline of the client register
	proto protocol;                           // Protocol used by the client : UDP || TCP
	unsigned int ports;                       // Number of ports assigned to this client
	unsigned int IPs;                         // Number of ip assigned to this client
	unsigned short nextReservedPort;          // max_host_ports are reserved for each client
	RSIPMessage* lastMsgRx;                   // Last message received
	RSIPMessage* lastMsgTx;                   // Last message transmitted
	NetworkEndPoint* nep;                     // End point towards the client
	Options* opts;                            // Options for this client
	uint32_t id;                              // To generate the bind ID
	pthread_mutex_t mutex;                    // Lock for this client

	// static variables
	static uint32_t nextCltID;                // Next ID to assign to a client
	static map<Address, RSIPClient*> clients; // Table (client IP, client)
	static pthread_mutex_t lockID;            // Lock client ID
	static pthread_mutex_t mutexdb;

	struct TimerParam {
		Address  address;
		uint32_t clientID;
	};

	uint32_t getID();
	// PRE : /
	// POST: Returns a new unique client ID,
	//       starting from 1


	void sendFreeResp(uint32_t bindID);
	// Send an asynchronous FREE_RESPONSE message to the client 
	// for this bind if nep is not null.


	void sendDeRegResp();
	// Send an asynchronous DE-REGISTER_RESPONSE message to the 
	// client if nep is not null..


	friend void removeClient(TimerParam*);
	// Do a delete on the client
	// if and only if the client is in STATE_UNREGISTERED
	// NOTE: call lockdb()


	friend void checkRegistering(TimerParam*);
	// Check that the client is still registered. 
	// If not, then function deregister() is called
	// for this client.
	// NOTE: call lockdb()


	friend void checkAssignments(TimerParam*);
	// Check that no bind time out occured for the client.
	// If yes, then the bind is released by calling
	// function free(bind ID, true)
	// NOTE: call lockdb()


	//************ Operations on client's binds ************

	Bind* search(uint32_t BindID);
	// Search for a bind
	// PRE : /
	// POST: returns the bind with this bind ID if it exists
	//       else return NULL
	// NOTE: Do NOT delete the returned value. Use del() instead.


	uint32_t add(Bind* b);
	// Add a bind to the list of client's binds
	// PRE : - b configured except the bindID
	//       - the lease time of b is the lease time requested
	// POST: - returns the bind ID for the new bind
	//       - the lease time of the b has been changed to the real 
	//         lease time assigned
	//       - the bindID of b has been set

	
	Bind* del(uint32_t bindID);
	// Remove the bind with this bindID from the list of binds
	// PRE : /
	// POST: the bind is removed from the list of client's binds, 
	//       if it existed. This does not release the resources but
	//       returns the bind if the bind existed
	//               NULL otherwise


	Bind* pop();
	// Remove & returns the first bind with from the list of binds
	// PRE : /
	// POST: returns the bind if a bind exists
	//               NULL if the list of binds is empty
	
};

void checkRegistering(RSIPClient::TimerParam* p);
void removeClient(RSIPClient::TimerParam* p);
void checkAssignments(RSIPClient::TimerParam* p);

#endif










