//***************************************************************************
//* RSIPMessage.cpp
//*
//* description    : class RSIPMessage
//*                  Modelize a RSIP message
//*
//* copyright      : (C) 2001 Cedric de Launois
//*                  Universite Catholique de Louvain, Belgium
//* begin          : Sept 24 2001
//* email          : cedric@delaunois.org
//***************************************************************************

/****************************************************************************
 * Copyright (c) 1998 Free Software Foundation, Inc.                        *
 *                                                                          *
 * Permission is hereby granted, free of charge, to any person obtaining a  *
 * copy of this software and associated documentation files (the            *
 * "Software"), to deal in the Software without restriction, including      *
 * without limitation the rights to use, copy, modify, merge, publish,      *
 * distribute, distribute with modifications, sublicense, and/or sell       *
 * copies of the Software, and to permit persons to whom the Software is    *
 * furnished to do so, subject to the following conditions:                 *
 *                                                                          *
 * The above copyright notice and this permission notice shall be included  *
 * in all copies or substantial portions of the Software.                   *
 *                                                                          *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS  *
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF               *
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.   *
 * IN NO EVENT SHALL THE ABOVE COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,   *
 * DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR    *
 * OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR    *
 * THE USE OR OTHER DEALINGS IN THE SOFTWARE.                               *
 *                                                                          *
 * Except as contained in this notice, the name(s) of the above copyright   *
 * holders shall not be used in advertising or otherwise to promote the     *
 * sale, use or other dealings in this Software without prior written       *
 * authorization.                                                           *
 ****************************************************************************/

#include "rsipMessage.h"


RSIPMessage::RSIPMessage(uint8_t type) 
{
	header.ver = RSIP_VERSION;
	header.type = type;
	header.length = MSG_HEADER_LEN;
	bufsize = (opt->pktsize < MIN_SIZE) ? MIN_SIZE : opt->pktsize;
	buffer = new uint8_t[bufsize];
	wptr = buffer;
	rptr = buffer;

	*wptr++ = header.ver;
	*wptr++ = header.type;
	SET16(wptr, MSG_HEADER_LEN);
	wptr+=2;

	bzero((void*)&address, sizeof(address));
	length = MSG_HEADER_LEN;
}



RSIPMessage::~RSIPMessage() 
{
	delete [] buffer;
}



int RSIPMessage::send(NetworkEndPoint* nep) 
{
	int sent;

	try {
		// Send the message
		sent = nep->writen((char*)buffer, header.length, address);
		if (sent < 0) {
			logmsg(LOG_WARNING, "SendPacket error\n");
			return 0;
		}
		if ((unsigned int)sent != header.length) {
			logmsg(LOG_WARNING, "SendPacket error\n");
			return 0;
		}
	} catch (NetworkException) {
		logmsg(LOG_WARNING, "SendPacket exception\n");
		return 0;
	}
	return 1;
}



int RSIPMessage::receive(NetworkEndPoint* nep, int timeout) 
{
	assert(nep!=NULL);
	Address from;
	int n;

	try {
		n = nep->readn((char*)buffer, bufsize, timeout, from);
		if (n==-1) return 0; // Time Out
		if (n<=0) return -1;
		length = (uint16_t)(n & 0xFFFF);
		setAddress(from);
	} catch (NetworkException) {
		return -1;
	}
		
	rptr = wptr = buffer;
	header.ver = *wptr++;
	header.type = *wptr++;
	header.length = GET16(wptr);
	wptr += 2;
		
	return 1;
}



void RSIPMessage::addParam(uint8_t type, uint8_t* val, uint16_t lenValue) 
{
	if (wptr + PARAM_HEADER_LEN + lenValue > buffer + bufsize) return;

	*wptr++ = type;
	SET16(wptr, lenValue);
	wptr += 2;
	memcpy(wptr, val, lenValue);
	wptr += lenValue;
	header.length = header.length + PARAM_HEADER_LEN + lenValue;
	length = header.length;
	SET16(buffer+2, header.length);
}



void RSIPMessage::addParamAddress(const struct RSIPAddressParam &a) 
{
	uint8_t lenValue = 1;
	uint8_t* val;

	if (!a.address.isDefined()) val = NULL;
	else {
		unsigned long addr;

		switch (a.addrType) {
		case ADDR_IPV4: 
		case ADDR_IPV4_NETMASK: 
			lenValue += 4;
			addr = a.address.getAddr();
			val = (uint8_t*)&addr;
			break;
		case ADDR_IPV6:
			lenValue += 16;
			addr = a.address.getAddr();
			val = (uint8_t*)&addr;
			break;
		case ADDR_FQDN:
			val = (uint8_t*)a.address.toChar();
			lenValue += strlen((char*)val);
			break;
		default:
			return;
		}
	}

	if (wptr + PARAM_HEADER_LEN + lenValue > buffer + bufsize) return;

	*wptr++ = PARAM_ADDRESS;
	SET16(wptr, lenValue);
	wptr += 2;
	*wptr++ = a.addrType;
	memcpy(wptr, val, lenValue-1);
	wptr = wptr + lenValue - 1;
	header.length = header.length + PARAM_HEADER_LEN + lenValue;
	length = header.length;
	SET16(buffer+2, header.length);
}



void RSIPMessage::addParamPort(const struct RSIPPortParam &p)
{
	uint8_t lenValue = 1 + p.ports.size()*2;

	if (wptr + PARAM_HEADER_LEN + lenValue > buffer + bufsize) return;

	*wptr++ = PARAM_PORT;
	SET16(wptr, lenValue);
	wptr += 2;
	*wptr++ = p.number;

	for (unsigned int i=0; i<p.ports.size(); i++) {
		uint16_t pt = p.ports[i];
		SET16(wptr, pt);
		wptr += 2;
	}

	header.length = header.length + PARAM_HEADER_LEN + lenValue;
	length = header.length;
	SET16(buffer+2, header.length);
}



void RSIPMessage::dump() 
{
	const char* paramType;
	const char* errorType;
	struct RSIPParam p;
	p.value = NULL;

	FILE* debugFile = stderr;

	if (length < MSG_HEADER_LEN) {
		fprintf(debugFile, "***************** BAD MESSAGE ***************\n");
		fprintf(debugFile, "* Address : %s (%lu)\n", address.toChar(), address.getAddr());
		dumpHex((char*)buffer, length);
		fprintf(debugFile, "*********************************************\n");
	}

	fprintf(debugFile, "******************** HEADER *****************\n");
	fprintf(debugFile, "* Address         : %s (%lu)\n", address.toChar(), address.getAddr());
	fprintf(debugFile, "* Version         : 0x%02x\n", header.ver);
	fprintf(debugFile, "* Type            : %s (0x%02x)\n", name(), header.type);
	fprintf(debugFile, "* Overall length  : %i\n", header.length);
	fprintf(debugFile, "* Real length     : %i\n", length);

	fprintf(debugFile, "*\n**************** PARAMETERS ****************\n");
	fprintf(debugFile, "* Parameters :");
	fprintf(debugFile, " (Total length : %i)\n*\n",	header.length - MSG_HEADER_LEN);

	resetParsing();
	while (getNextParam(p)) {

		if (p.header.type>PARAM_MAX_NUMBER) paramType="UNKNOWN";
		else paramType=paramNames[p.header.type];

		uint16_t len = p.header.length;

		fprintf(debugFile, "* Type     : %s (0x%02x)\n", paramType, p.header.type);
		fprintf(debugFile, "* Length   : %i (0x%04x)\n", len, p.header.length);

		if (p.header.type == PARAM_ERROR) {
			uint16_t val = GET16(p.value);
			uint16_t modval = val%100;

			errorType="Unknown";
			if (val>100 && val<=GEN_ERROR_MAX_NUMBER) errorType = genErrorNames[modval];
			if (val>200 && val<=PM_ERROR_MAX_NUMBER)  errorType = pmErrorNames[modval];
			if (val>300 && val<=PRP_ERROR_MAX_NUMBER) errorType = prpErrorNames[modval];

			fprintf(debugFile, "* Value    : %s (%i)\n", errorType, val);
			if (len>=length-MSG_HEADER_LEN) len=length-MSG_HEADER_LEN;
			dumpHex((char*)p.value, len);

		} else { 

			fprintf(debugFile, "* Value    : \n");
			if (len>=length-MSG_HEADER_LEN) len=length-MSG_HEADER_LEN;
			dumpHex((char*)p.value, len);
		}

		fprintf(debugFile, "*\n");

	} 
	fprintf(debugFile, "*********************************************\n");

}



void RSIPMessage::resetParsing() 
{
	rptr = buffer + MSG_HEADER_LEN;
}



int RSIPMessage::getFirstParam(struct RSIPParam& result) 
{
	rptr = buffer + MSG_HEADER_LEN;
	return getNextParam(result);
}



int RSIPMessage::getNextParam(struct RSIPParam& result) 
{
	// Are there other valid parameters ?
	if (rptr + PARAM_HEADER_LEN > buffer + header.length)
		return 0;

	// Copy the header
	memcpy(&result, rptr, PARAM_HEADER_LEN);
	rptr += PARAM_HEADER_LEN;

	result.header.length = ntohs(result.header.length);

	if (result.header.length==0) result.value=NULL;
	else result.value = rptr;

	rptr = rptr + result.header.length;

	return 1;
}



int RSIPMessage::parse(struct RSIPMsgParams &msgp)
{
	struct RSIPParam p;

	msgp.flags = 0;
	msgp.tunnelType = 0;
	resetParsing();

	int err, i = 0;
	while (RSIPMessages[header.type][i]) {

		if (!getNextParam(p)) {
			if (RSIPMessages[header.type][i] & REQUIRED)
				return MISSING_PARAM;
			else return 1;
		}

		if (!p.value) return BAD_PARAM;

		if ((p.header.type != PTYPE(RSIPMessages[header.type][i])) 
		    && (RSIPMessages[header.type][i] & REQUIRED))
			return MISSING_PARAM;

		// Single or multiple parameter ?
		bool SGL = !(RSIPMessages[header.type][i] & MULTIPLE);

		switch(p.header.type) {

		case PARAM_MESSAGE_COUNTER:
			if (p.header.length!=4) return BAD_PARAM;
			if ((msgp.flags & FLAG_COUNTER) && SGL) 
				return DUPLICATE_PARAM;
			msgp.flags |= FLAG_COUNTER;
			msgp.counter = GET32(p.value);
			break;

		case PARAM_LEASE_TIME:
			if (p.header.length!=4) return BAD_PARAM;
			if ((msgp.flags & FLAG_LEASE_TIME) && SGL) 
				return DUPLICATE_PARAM;
			msgp.flags |= FLAG_LEASE_TIME;
			msgp.leaseTime = GET32(p.value);
			break;

		case PARAM_CLIENT_ID:
			if (p.header.length!=4) return BAD_PARAM;
			if ((msgp.flags & FLAG_CLIENT_ID) && SGL) 
				return DUPLICATE_PARAM;
			msgp.flags |= FLAG_CLIENT_ID;
			msgp.clientID = GET32(p.value);
			break;

		case PARAM_BIND_ID:
			if (p.header.length!=4) return BAD_PARAM;
			if ((msgp.flags & FLAG_BIND_ID) && SGL) 
				return DUPLICATE_PARAM;
			msgp.bindID = GET32(p.value);
			msgp.flags |= FLAG_BIND_ID;
			break;

		case PARAM_ADDRESS:
			if (msgp.flags & FLAG_ADDRESS_1) {
				if ((msgp.flags & FLAG_ADDRESS_2) && SGL) {
					if ((msgp.flags & FLAG_TUNNEL_EP) && SGL)
						return EXTRA_PARAM;
					
					// Tunnel EP address
					if ((err=parseAddress(p, msgp.tunnelEP.addrType, msgp.tunnelEP.address))!=1)
						return err;
					msgp.flags |= FLAG_TUNNEL_EP;
				}
				// Second address parameter
				if ((err=parseAddress(p, msgp.address2.addrType, msgp.address2.address))!=1)
					return err;
				msgp.flags |= FLAG_ADDRESS_2;
			} else {
				// First address parameter
				if ((err=parseAddress(p, msgp.address1.addrType, msgp.address1.address))!=1)
					return err;
				msgp.flags |= FLAG_ADDRESS_1;
			}
			break;

		case PARAM_PORT:
			if (msgp.flags & FLAG_PORTS_1) {
				if ((msgp.flags & FLAG_PORTS_2) && SGL) return EXTRA_PARAM;
				// Second port parameter
				if ((err=parsePort(p, msgp.ports2.ports, msgp.ports2.number))!=1)
					return err;
				msgp.flags |= FLAG_PORTS_2;
			} else {
				// First port parameter
				if ((err=parsePort(p, msgp.ports1.ports, msgp.ports1.number))!=1)
					return err;
				msgp.flags |= FLAG_PORTS_1;
			}
			break;

		case PARAM_TUNNEL_TYPE:
			if (p.header.length!=1) return BAD_PARAM;
			if ((msgp.flags & FLAG_TUNNEL_TYPE) && SGL) 
				return DUPLICATE_PARAM;
			msgp.flags |= FLAG_TUNNEL_TYPE;

			switch (p.value[0]) {
			case TUNNEL_IP_IP:
				msgp.tunnelType |= TFLAG_IP_IP;
				break;
			case TUNNEL_GRE:
				msgp.tunnelType |= TFLAG_GRE;
				break;
			case TUNNEL_L2TP:
				msgp.tunnelType |= TFLAG_L2TP;
				break;
			}

			break;

		case PARAM_ERROR:
			if (p.header.length!=2) return BAD_PARAM;
			if ((msgp.flags & FLAG_ERROR_TYPE) && SGL) 
				return DUPLICATE_PARAM;
			msgp.flags |= FLAG_ERROR_TYPE;
			msgp.errorType = GET16(p.value);
			break;

		default:
			if (p.header.type>PARAM_MAX_NUMBER) 
				return ILLEGAL_PARAM;
			else msgp.flags |= FLAG_OTHER;
		}

		i++;
	}

	return 1;
}



int RSIPMessage::forge(const struct RSIPMsgParams &msgp)
{
	int i = 0;
	int ap = 1; // Current address parameter to add
	int pp = 1; // Current port parameter to add
	int parameter;

	while ((parameter = RSIPMessages[header.type][i])) {
		switch(PTYPE(parameter)) {

		case PARAM_ADDRESS:
			switch(ap) {
			case 1:
				if ((parameter&REQUIRED) || (msgp.flags & FLAG_ADDRESS_1))
					addParamAddress(msgp.address1);
				ap++;
				break;
			case 2:
				if ((parameter&REQUIRED) || (msgp.flags & FLAG_ADDRESS_2))
					addParamAddress(msgp.address2);
				ap++;
				break;
			case 3:
				if ((parameter&REQUIRED) || (msgp.flags & FLAG_TUNNEL_EP))
					addParamAddress(msgp.tunnelEP);
				ap++;
				break;
			default:
				break;
			}
			break;

		case PARAM_PORT:
			if (pp==1) {
				if ((parameter&REQUIRED) || (msgp.flags & FLAG_PORTS_1))
					addParamPort(msgp.ports1);
				pp++;
			} else if (pp==2) {
				if ((parameter&REQUIRED) || (msgp.flags & FLAG_PORTS_2))
					addParamPort(msgp.ports2);
				pp++;
			}
			break;

		case PARAM_LEASE_TIME:
			if ((parameter&REQUIRED) || (msgp.flags & FLAG_LEASE_TIME)) {
				uint32_t leaseTime = htonl(msgp.leaseTime);
				addParam(PARAM_LEASE_TIME, (uint8_t*)&leaseTime, 4);
			}
			break;

		case PARAM_CLIENT_ID:
			if ((parameter&REQUIRED) || (msgp.flags & FLAG_CLIENT_ID)) {
				uint32_t clientID = htonl(msgp.clientID);
				addParam(PARAM_CLIENT_ID, (uint8_t*)&clientID, 4);
			}
			break;

		case PARAM_BIND_ID:
			if ((parameter&REQUIRED) || (msgp.flags & FLAG_BIND_ID)) {
				uint32_t bindID = htonl(msgp.bindID);
				addParam(PARAM_BIND_ID, (uint8_t*)&bindID, 4);
			}
			break;

		case PARAM_TUNNEL_TYPE:
			if ((parameter&REQUIRED) || (msgp.flags & FLAG_TUNNEL_TYPE)) {
				if (msgp.tunnelType==0) {

					if (SUPPORT_TUNNEL_IP_IP) {
						uint8_t tunnel = (uint8_t)TUNNEL_IP_IP;
						addParam(PARAM_TUNNEL_TYPE, (uint8_t*)&tunnel, 1);
					}
					if (SUPPORT_TUNNEL_GRE) {
						uint8_t tunnel = (uint8_t)TUNNEL_GRE;
						addParam(PARAM_TUNNEL_TYPE, (uint8_t*)&tunnel, 1);
					}
					if (SUPPORT_TUNNEL_L2TP) {
						uint8_t tunnel = (uint8_t)TUNNEL_L2TP;
						addParam(PARAM_TUNNEL_TYPE, (uint8_t*)&tunnel, 1);
					}

				} else addParam(PARAM_TUNNEL_TYPE, (uint8_t*)&msgp.tunnelType, 1);
			}
			break;

		case PARAM_RSIP_METHOD:
			if (SUPPORT_RSA_IP) {
				uint8_t method = (uint8_t)METHOD_RSA_IP;
				addParam(PARAM_RSIP_METHOD, (uint8_t*)&method, 1);
			}
			if (SUPPORT_RSAP_IP) {
				uint8_t method = (uint8_t)METHOD_RSAP_IP;
				addParam(PARAM_RSIP_METHOD, (uint8_t*)&method, 1);
			}
			break;

		case PARAM_ERROR:
			if ((parameter&REQUIRED) || (msgp.flags & FLAG_ERROR_TYPE)) {
				uint32_t errorType = htons(msgp.errorType);
				addParam(PARAM_ERROR, (uint8_t*)&errorType, 2);
			}
			break;

		case PARAM_FLOW_POLICY:
			if (true) {
				uint8_t policy[2];
				policy[0] = (uint8_t) LOCAL_POLICY;
				policy[1] = (uint8_t) REMOTE_POLICY;
				addParam(PARAM_FLOW_POLICY, (uint8_t*)policy, 2);
			}
			break;

		case PARAM_INDICATOR:
			break;

		case PARAM_MESSAGE_COUNTER:
			if ((parameter&REQUIRED) || (msgp.flags & FLAG_COUNTER)) {
				uint32_t counter = htonl(msgp.counter);
				addParam(PARAM_MESSAGE_COUNTER, (uint8_t*)&counter, 4);
			}
			break;

		default:
			return 0;
			break;
		}

		i++;
	}

	return 1;
}



int RSIPMessage::forgeError(uint16_t errorType, uint32_t counter, uint32_t clientID,
		       uint32_t bindID, uint32_t flags)
{
	struct RSIPMsgParams msgp;
	msgp.errorType = errorType;
	msgp.counter = counter;
	msgp.clientID = clientID;
	msgp.bindID = bindID;
	msgp.flags = flags;
	return forge(msgp);
}



void RSIPMessage::destroy() 
{
	header.ver = RSIP_VERSION;
	header.type = 0;
	header.length = MSG_HEADER_LEN;
	wptr = rptr = buffer + MSG_HEADER_LEN;
	length = MSG_HEADER_LEN;
}



int RSIPMessage::operator==(const RSIPMessage &msg) const 
{
	if (header.length != msg.header.length) return 0;
	if (header.ver != msg.header.ver) return 0;
	if (header.type != msg.header.type) return 0;
	if (memcmp(buffer,msg.buffer,msg.header.length) != 0) return 0;
	return 1;
}



int RSIPMessage::operator!=(const RSIPMessage &msg) const 
{
	if (header.length != msg.header.length) return 1;
	if (header.ver != msg.header.ver) return 1;
	if (header.type != msg.header.type) return 1;
	if (memcmp(buffer,msg.buffer,msg.header.length) != 0) return 1;
	return 0;
}



int RSIPMessage::check()
{
	// Case 1 : Message too short
	if (length<MSG_HEADER_LEN) return BAD_MESSAGE;

	// Case 2 : incorrect length
	if (length!=header.length) return BAD_MESSAGE;

	// Unknown message type
	if (header.type > MESSAGE_MAX_NUMBER) return ILLEGAL_MESSAGE;

	return true;
}



int RSIPMessage::parsePort(const struct RSIPParam& p, deque<uint16_t> &ports, uint8_t &nbPorts)
{
	nbPorts=0;

	if (!p.value) return BAD_PARAM;
	if (((p.header.length-1)%2)!=0) return BAD_PARAM;

	nbPorts = p.value[0];

	// Case 1 : only a number of ports
	if (p.header.length==1) {
		nbPorts = p.value[0];
		return 1;
	}

	// Case 2 : a least one port specified
	uint16_t port;
	for (unsigned int offset=1; offset<p.header.length; offset+=2) {
		port = GET16(p.value+offset);
		ports.push_back(port);
	}

	return 1;
}



int RSIPMessage::parseAddress(const struct RSIPParam& p, uint8_t &type, Address& address)
{
	if (!p.value) return BAD_PARAM;

	type = p.value[0];

	switch (type) {

	case ADDR_IPV4:
	case ADDR_IPV4_NETMASK:
		if (p.header.length!=5) { // Bad IP address
			if (p.header.length==1) return 1;
			else return 0;
		} else {
			address.setAddr(*((uint32_t*)(p.value+1)));
		}
		break;

	case ADDR_FQDN:
		logmsg(LOG_WARNING, "Unsupported FQDN address type\n");
		return ILLEGAL_PARAM;

	case ADDR_IPV6:
		logmsg(LOG_WARNING, "Unsupported IPv6 address type\n");
		return ILLEGAL_PARAM;

	default:
		return BAD_PARAM;
	}

	return 1;
}



void RSIPMessage::dumpHex(char const *buf, int len) 
{
	int i;
	FILE* debugFile = stderr;

	for (i=0; i<len; i++) {
		if (i == len-1 || !((i+1)%16)) {
			fprintf(debugFile, "0x%.2x\n", (unsigned char) buf[i]);
		} else if (!((i+1)%8)) {
			fprintf(debugFile, "0x%.2x - ", (unsigned char) buf[i]);
		} else if (!((i+1)%2)) {
			fprintf(debugFile, "0x%.2x ", (unsigned char) buf[i]);
		} else {
			fprintf(debugFile, "0x%.2x ", (unsigned char) buf[i]);
		}
	}
}

