//***************************************************************************
//* RSIPMessage.h 
//*
//* description    : class RSIPMessage.h 
//*                  Modelize a RSIP message
//*
//* copyright      : (C) 2001 Cedric de Launois
//*                  Universite Catholique de Louvain, Belgium
//* begin          : Sept 24 2001
//* email          : cedric@delaunois.org
//***************************************************************************

/****************************************************************************
 * Copyright (c) 1998 Free Software Foundation, Inc.                        *
 *                                                                          *
 * Permission is hereby granted, free of charge, to any person obtaining a  *
 * copy of this software and associated documentation files (the            *
 * "Software"), to deal in the Software without restriction, including      *
 * without limitation the rights to use, copy, modify, merge, publish,      *
 * distribute, distribute with modifications, sublicense, and/or sell       *
 * copies of the Software, and to permit persons to whom the Software is    *
 * furnished to do so, subject to the following conditions:                 *
 *                                                                          *
 * The above copyright notice and this permission notice shall be included  *
 * in all copies or substantial portions of the Software.                   *
 *                                                                          *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS  *
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF               *
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.   *
 * IN NO EVENT SHALL THE ABOVE COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,   *
 * DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR    *
 * OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR    *
 * THE USE OR OTHER DEALINGS IN THE SOFTWARE.                               *
 *                                                                          *
 * Except as contained in this notice, the name(s) of the above copyright   *
 * holders shall not be used in advertising or otherwise to promote the     *
 * sale, use or other dealings in this Software without prior written       *
 * authorization.                                                           *
 ****************************************************************************/

class RSIPMessage;

#ifndef _RSIP_MESSAGE_H
#define _RSIP_MESSAGE_H

#include <sys/types.h>
#include <netinet/in.h>
#include <stdlib.h>
#include <deque>
#include "exceptions.h"
#include "common.h"
#include "netEndPoint.h"
#include "assert.h"
#include "protocol.h"
#include "address.h"

#define MIN_SIZE 256  // Lowest maximum packet size allowed

class RSIPMessage {
 public:
	

	RSIPMessage(uint8_t type = 0);
	// Constructor
	// PRE : type is the message type
	// POST: - an empty RSIP message 
	//           with undefined type
	//           with RSIP version : RSIP_VERSION,
	//           with length properly set up
	//         is created.
	//       - Message well formed

	
	~RSIPMessage();
	// Destructor
	// PRE : /
	// POST: msg destroyed, memory freed

	
	int send(NetworkEndPoint* nep);
	// SPEC: Send a RSIP packet
	// PRE : - Message well formed
	//       - nep : valid End Point to the destinator, 
	//       - for UDP : address must be set to the destinator's address
	// POST: 1 if the msg has been properly sent
	//       0 otherwise

	
	int receive(NetworkEndPoint* nep, int timeout);
	// SPEC: - receive a RSIP packet
	//       - timeout : in microseconds
	//                 -1 for an infinite timeout
	// PRE : - nep : valid Network End Point
	// POST: msg is the packet received
	//       returns 1  if packet received before timeout
	//               0  if timeout
	//               -1 if connexion closed before timeout
	// NOTE: the message received may be malformed
	

	int parse(struct RSIPMsgParams &p);
	// Fill the structure with the most common parameters presents in 
	// a message coming from a client
	// PRE : Message well formed
	// POST: - p is filled with the parameters
	//       - the flags are set to reflect the parameters which are 
	//         present in the message (member p.flags)
	//       - returns the error code (see RFC3103) if an error occured
	//         else returns 1
	// NOTE: - the parameters supported are :
	//           counter, leaseTime, clientID, bindID, 
	//           address (first), port (first),
	//           address (second), port (second) and tunnelType.
	//       - FLAG_OTHER is set if there is another parameter
	//       - errors detected: 
	//           BAD_PARAM, ILLEGAL_PARAM, DUPLICATE_PARAM, 
	//           EXTRA_PARAM, MISSING_PARAM


	int forge(const struct RSIPMsgParams &msgp);
	// Build a message based on the type and on the parameters in p
	// PRE : the type of the message is set
	// POST: 1 if success,
	//       0 otherwise


	int forgeError(uint16_t errorType, uint32_t counter, uint32_t clientID,
		       uint32_t bindID, uint32_t flags);
	// Build a ERROR_REPONSE message based on the parameters
	// POST: 1 if success,
	//       0 otherwise


	void destroy();
	// PRE : /
	// POST: the message is empty
	

	int operator==(const RSIPMessage &) const;
	// PRE : Messages well formed
	// POST: true if both messages are equals
	//       false otherwise


	int operator!=(const RSIPMessage &) const;
	// PRE : Messages well formed
	// POST: true if messages are differents
	//       false otherwise


	inline int getType() const { return header.type; }
	// PRE : Message well formed
	// POST: returns the message type
	

	inline int getVersion() const { return header.ver; }
	// PRE : Message well formed
	// POST: gets the RSIP version of the message
    

	inline void setAddress(Address& a) { address=a; }
	// PRE : Message well formed
	// POST: sets the address for the message
	

	inline Address getAddress() { return address; }
	// PRE : Message well formed
	// POST: - returns the address of the message
	// NOTE: This address is the origin address if the message was received
	//       otherwise it is the destination address


	int check();
	// Checks the header of the message
	// PRE : /
	// POST: - returns 1 if the message is well formed
	//                 the error code (see RFC3130) otherwise
	// NOTE: errors detected: BAD_MESSAGE, ILLEGAL_MESSAGE


	void dump();
	// Dump the message in a human readable form
	// PRE : /
	// POST: RSIP message dumped on the standard output


	inline const char* name() { return ((header.type<=MESSAGE_MAX_NUMBER)?msgNames[header.type]:msgNames[0]); }
	// Returns the name of the message


	inline const char* pname(uint8_t type) { return paramNames[type]; }
	// Returns the parameter name


 private:

	uint16_t bufsize;
	uint8_t* buffer;             // Message buffer
	uint8_t* wptr;               // Writing pointer to the message
	uint8_t* rptr;               // Reading pointer to the message
	uint16_t length;             // Real length of the message

	struct RSIPMsgHeader header; // RSIP Message Header
	Address address;             // Origin or destination address

	void addParam(uint8_t type, uint8_t* val, uint16_t len);
	// SPEC: - add a parameter to the message
	//       - type  : the parameter type
	//       - val   : the value of the parameter
	//       - len   : the length of the parameter
	// PRE : Message well formed
	// POST: - a parameter of type 'type' and with value 'value' has been
	//         added to the message msg
	//       - the length field of the param has been adjusted
	

	void addParamAddress(const struct RSIPAddressParam &a);
	// PRE : Message well formed
	//       a.addrType is a valid address type
	// POST: - a parameter of type Address has been added to the message 
	//         msg
	//       - the length field of the param has been adjusted


	void addParamPort(const struct RSIPPortParam &p);
	// PRE : Message well formed
	// POST: - a parameter of type Port has been added to the message 
	//         msg
	//       - the length field of the param has been adjusted


	void resetParsing();
	// PRE : Message well formed
	// POST: reinitialize the parsing of parameters
	

	int getFirstParam(struct RSIPParam& p);
	// PRE : Message well formed
	// POST: - p is the first parameter of the message
	//           p.value points to the value of the parameter
	//           p.value MUST NOT be deleted nor freed
	//       - returns 1 if p is a valid parameter
	//       - renvoie 0 if there's no parameter

	
	int getNextParam(struct RSIPParam& p);
	// PRE : Message well formed
	// POST: - p is the next parameter
	//           p.value points to the value of the parameter
	//           p.value MUST NOT be deleted nor freed
	//       - returns 1 if p is a valid parameter
	//       - renvoie 0 if there's no more parameters


	int parsePort(const struct RSIPParam& p, deque<uint16_t> &ports, uint8_t &nbPorts);
	int parseAddress(const struct RSIPParam& p, uint8_t &type, Address& address);

	static void dumpHex(char const *buf, int len);

	inline void foo() { bool foo=paramNames==genErrorNames; foo=pmErrorNames==prpErrorNames; }
	// To avoid warning message when those tables are not used
	// If you knows about another way to do that, don't hesitate...
};


#endif










