//**************************************************************************
//* rsipd.cpp 
//*
//* description    : rsipd.cpp
//*                  Main part
//* copyright      : (C) 2001 Cedric de Launois
//*                  Universite Catholique de Louvain, Belgium
//* begin          : Sept 11 2001
//* email          : delaunois@info.ucl.ac.be
//**************************************************************************

/****************************************************************************
 * Copyright (c) 1998 Free Software Foundation, Inc.                        *
 *                                                                          *
 * Permission is hereby granted, free of charge, to any person obtaining a  *
 * copy of this software and associated documentation files (the            *
 * "Software"), to deal in the Software without restriction, including      *
 * without limitation the rights to use, copy, modify, merge, publish,      *
 * distribute, distribute with modifications, sublicense, and/or sell       *
 * copies of the Software, and to permit persons to whom the Software is    *
 * furnished to do so, subject to the following conditions:                 *
 *                                                                          *
 * The above copyright notice and this permission notice shall be included  *
 * in all copies or substantial portions of the Software.                   *
 *                                                                          *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS  *
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF               *
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.   *
 * IN NO EVENT SHALL THE ABOVE COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,   *
 * DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR    *
 * OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR    *
 * THE USE OR OTHER DEALINGS IN THE SOFTWARE.                               *
 *                                                                          *
 * Except as contained in this notice, the name(s) of the above copyright   *
 * holders shall not be used in advertising or otherwise to promote the     *
 * sale, use or other dealings in this Software without prior written       *
 * authorization.                                                           *
 ****************************************************************************/

#include "rsipd.h"

//****************************************************************************
//*
//* MAIN Function
//*
//****************************************************************************

int main(int argc, char *argv[]) 
{
	int err;
	struct stat statbuf;
	char *modprobePaths[] = { 
		"/sbin/modprobe",
		"/usr/sbin/modprobe",
		"/usr/local/sbin/modprobe",
		0
	};

	opt = NULL;
	tmr = NULL;
	sh = NULL;
	Resources* pool = NULL;
	Server* svr = NULL;
	atexit(clean);
	umask(0x077);

	// Initialize the signal processing:
	// Sets the signal handler functions and mask them during 
	// initialization
	signal(SIGINT, sigTerm);
	signal(SIGTERM, sigTerm);
	signal(SIGPIPE, SIG_IGN);
	signal(SIGUSR1, dumpState);

	sigset_t newmask, oldmask;
	sigemptyset(&newmask);
        sigaddset(&newmask, SIGUSR1);
        sigaddset(&newmask, SIGTERM);
        sigaddset(&newmask, SIGINT);
	pthread_sigmask(SIG_BLOCK,&newmask,&oldmask);

	setlogoutput(LOG_PERROR);

	// Initialize the options
	if ((err=init(opt, argc, argv))!=0) exit(err);
	setloglevel(opt->log_level);


	// Check permission
	if (!isRoot()) {
		fprintf(stderr, "\n");
		logconsole(LOG_ERR, "You must be root to run this server.\n");
		logconsole(LOG_ERR, "RSIP Server Version %s\n", VERSION);
		exit(ERR_BAD_RIGHTS);
	}


	// Daemonize the process if needed
	fflush(0);
	if (opt->daemon) {
		daemonStart();
		umask(0x077);
		opt->output = 0;
		setlogoutput(LOG_CONS);
	}

	int i=0;
	char* modprobe;
	while ((modprobe = modprobePaths[i++])) {
		if (stat(modprobe, &statbuf)==0)
			break;
	}

	if (!modprobe) {
		logconsole(LOG_ERR, "Can't find modprobe binary.\n");
		exit(ERR_MODPROBE);
	}

	// Launch the 'shell' that executes external commands
	// The shell forks when intialized, and forks for each command
	// it issues. Must be initialized before the pool, otherwise: mem leak.
	if (opt->daemon) sh = new Shell(false);
	else sh = new Shell(true);
	sh->start();

	// Check modules and load them if needed
	if (checkModule("tunl0")<0) {
		fprintf(stderr, "\n");
		logconsole(LOG_WARNING, "IPIP tunnels not available. Trying to load ipip module...\n");
		sh->execute(modprobe, "ipip", NULL);
		if (checkModule("tunl0")<0) {
			logconsole(LOG_ERR, "IPIP tunnels not available. Change kernel option !\n");
			logconsole(LOG_ERR, "Starting failed.\n");
			delete sh;
			exit(ERR_IPIP);
		} else logconsole(LOG_WARNING, "Module successfully loaded.\n");
	}

#if SUPPORT_TUNNEL_GRE
	if (checkModule("gre0")<0) {
		fprintf(stderr, "\n");
		logconsole(LOG_WARNING, "GRE tunnels not available. Trying to load ip_gre module...\n");
		sh->execute(modprobe, "ip_gre", NULL);
		if (checkModule("gre0")<0) {
			logconsole(LOG_ERR, "GRE tunnels not available. Will use ipip tunnels only.\n");
		} else logconsole(LOG_WARNING, "Module successfully loaded.\n");
	}
#endif


	// Intialize the pool of resources
	init(pool, opt);
	Bind::setPool(pool);


	// Start the timer module
	tmr = new Timer();


 	// Start the server
	unsigned int t = 0;
	svr = new Server(opt);
	try {
		// Actual start of the server
		if (svr->start(UDP|TCP)) {
			logmsg(LOG_NOTICE, "Server started [PID %i]. Protocols : UDP TCP\n", getpid());
			writepid(opt->pid_file);
#ifdef _HAVE_SLP
			// Registers with SLP
			Address a = opt->int_ip;
			a.setPort(opt->port);
			registerWithSlp(a);
#endif
		} else goto end;

		// Monitoring and waiting for a termination signal
		pthread_sigmask(SIG_SETMASK,&oldmask,NULL);
		while (!opt->stop) {
			sleep(1);
			if (!opt->check_ip) continue;

			// Check for a change of the server IP
			if (t++>=opt->refresh) { t=0; monitor(opt, pool); }
		}

		// Stopping the server
		svr->stop();
		monitor(opt, pool);

	} catch (NetworkException) {
		logmsg(LOG_ERR,"Network Exception: %s\n", strerror(errno));
	} catch (NotEnoughMemory) {
		logmsg(LOG_ERR,"Not enough memory !\n");
	} catch (...) {
		logmsg(LOG_ERR,"Exception: \n", strerror(errno));
	}

	unlink(opt->pid_file);


 end:
	// Stopping program
	logmsg(LOG_NOTICE, "Server stopped.\n");
	if (!opt->stop) opt->stop = true;
	if (pool) delete pool;
	if (svr) delete svr;

	exit(RET_SUCCESS);
}



//****************************************************************************
//* Function sigTerm
//* SPEC: process signals received
//* PRE : src is the signal received
//****************************************************************************
void sigTerm(int src) 
{
	logmsg(LOG_NOTICE, "Signal %i received. Stopping server.\n", src);


	// Stops the server
	if (src==SIGALRM) {
		logmsg(LOG_ERR, "Server not responding. Killing it...\n");
		kill(getpid(), SIGKILL);

	} else {
		opt->stop = true;
		signal(SIGALRM, sigTerm); // If the server crashed...
		alarm(3);
	}
}



//****************************************************************************
//* Function clean
//* SPEC: gracefully terminate the program
//****************************************************************************
void clean(void) 
{
	if (tmr) delete tmr;
	if (sh) delete sh;
	if (opt) {
		if (opt->cfg_file) free(opt->cfg_file);
		if (opt->ext_if) free(opt->ext_if);
		if (opt->ext_if) free(opt->int_if);
		delete opt;
	}
}



//****************************************************************************
//* Function isRoot
//* SPEC: check that the program is running as root
//* PRE : / 
//* POST: returns 1 if root
//*               0 otherwise
//****************************************************************************
int isRoot() 
{
	return getuid() ? 0 : 1;
}



//****************************************************************************
//* Function usage. 
//* PRE : name of the program
//* POST: shows usage of the program
//****************************************************************************
void usage(char const *name) 
{
	fprintf(stdout, "Name: %s   -   RSIP server\n", name);
	fprintf(stdout, "Usage: %s [ -pdVhcDa ]\n", name);
	fprintf(stdout, "Options\n");
	fprintf(stdout, "   -p port  -- Specify RSIP port (default: 4555)\n");
	fprintf(stdout, "   -d       -- Dump packets received\n");
	fprintf(stdout, "   -V       -- Print version and exit\n");
	fprintf(stdout, "   -h       -- Print usage information and exit\n");
	fprintf(stdout, "   -c file  -- Specify configuration file (default: /etc/rsipd.conf)\n");
	fprintf(stdout, "   -D       -- Debug mode (no daemon process)\n");
	fprintf(stdout, "   -a       -- Add ARP entry for each leased RSIP address\n");
	fprintf(stdout, "\n");
	fprintf(stdout, "RSIP Server Version %s\n", VERSION);
	fprintf(stdout, "C. de Launois\n");
	fprintf(stdout, "Universite Catholique de Louvain, Belgium\n");
}



//****************************************************************************
//* Function monitor
//* SPEC: Monitor the IP of the server
//* PRE : server started
//* POST: pool updated if IP has changed
//*       returns 1 if succes, 0 if an error occured
//****************************************************************************
int monitor(Options* o, Resources* pool) 
{
	Address actualAddress = o->ext_ip;

	// Getting the actual address of the external interface
	if (getIP(o->ext_if, actualAddress)!=1)
		actualAddress.undefine();

	// Managing the change of the IP of the external interface
	if (o->ext_ip!=actualAddress) {
		// The external IP address has changed !

		if (actualAddress.isDefined())
			logmsg(LOG_INFO, "External IP address changed !");
		else
			logmsg(LOG_INFO, "External interface seems to be down !");

		pool->update(o->ext_ip, actualAddress);

		// Free resource
		Resource r(o->ext_ip, -1);
		RSIPClient::freeAll(r, true);

		pool->update(o->ext_ip,actualAddress);
	
		o->ext_ip = actualAddress;
	}

	return 1;
}




//****************************************************************************
//* Function dumpState
//* SPEC: dump the current state of the server
//****************************************************************************
void dumpState(int src) {
	char s[160];

	time_t now = time(0);
	snprintf(s, 160, "RSIP server status file. %s", ctime(&now));
	string text(RSIPClient::getInfo());

	FILE* f = fopen(opt->status_file, "w");
	if (!f) {
		logmsg(LOG_ERR,"Can't create state file %s\n", opt->status_file);
		return;
	}

	fprintf(f,"%s",s);
	fprintf(f,"%s",text.c_str());
	fclose(f);

	logmsg(LOG_NOTICE,"Dumped state available in file %s\n", opt->status_file);
	if (opt->output==1) {
		fprintf(stdout,"%s",s);
		fprintf(stdout,"%s",text.c_str());
	}
}



//****************************************************************************
//*
//* Fonction daemonStart
//*
//* Detach a daemon process from login session context
//*
//****************************************************************************
void daemonStart() 
{
	int childpid;

	// If started by init (process 1) from the /etc/inittab file, there's no
	// need to detach
	if (getppid() == 1) goto out;

	// Ignore the terminal stop signals
#ifdef SIGTTOU
	signal(SIGTTOU, SIG_IGN);
#endif
#ifdef SIGTTIN
	signal(SIGTTIN, SIG_IGN);
#endif
#ifdef SIGTSTP
	signal(SIGTSTP, SIG_IGN);
#endif
 
	// If we were not started in the background, fork and let the parent exit.
	if ( (childpid = fork()) < 0) logconsole(LOG_ERR, "Can't fork first child\n");
	else if (childpid > 0) exit(0); // Parent

	// First child process
	// Disassociate from controlling terminal and process group

	if (setpgrp() == -1) logconsole(LOG_ERR, "Can't change process group\n");

	signal(SIGHUP, SIG_IGN);

	if ( (childpid = fork()) < 0) logconsole(LOG_ERR, "Can't fork second child\n");
	else if (childpid > 0) exit(0); // First child

	// Second child

 out:


	errno = 0;

	// Clear any inherited file mode creation mask
	umask(0);
}



//****************************************************************************
//* Function init
//* SPEC: Initialize the options
//* PRE : /
//* POST: - options initialized
//*       - Returns >=0 if success
//*                 the error code <0 otherwise
//****************************************************************************
int init(Options* &o, int argc, char** argv) 
{
	if (o==NULL) o = new Options;

	o->dump                        = DFLT_DUMP;
	o->port                        = DFLT_RSIP_PORT;
	o->daemon                      = DFLT_DAEMON;
	o->output                      = 1;
	o->initial_lease_time          = DFLT_INITIAL_LEASE_TIME; 
	o->initial_register_lease_time = DFLT_INITIAL_REGISTER_LEASE_TIME;  
	o->max_lease_time              = DFLT_MAXIMUM_LEASE_TIME;               
	o->min_lease_time              = DFLT_MINIMUM_LEASE_TIME;               
	o->cfg_file                    = strdup(DFLT_CFG_FILE);
	o->status_file                 = DFLT_STATUS_FILE;
	o->pid_file                    = DFLT_PID_FILE;
	o->stop                        = false;
	o->refresh                     = DFLT_REFRESH;
	o->pktsize                     = DFLT_PKTSIZE;
	o->rsa_method                  = DFLT_RSA_METHOD;
	o->max_tcp_clients             = DFLT_MAX_TCP_CLIENTS;
	o->max_clients                 = DFLT_MAX_CLIENTS;
	o->max_host_ports              = DFLT_MAX_HOST_PORTS;
	o->ext_if                      = strdup(DFLT_EXT_IF);
	o->int_if                      = strdup(DFLT_INT_IF);
	o->log_level                   = DFLT_LOG_LEVEL;
	o->arp                         = DFLT_ARP;
	o->rsap_low_port               = DFLT_RSAP_LOW_PORT;
	o->rsap_high_port              = DFLT_RSAP_HIGH_PORT;
	o->listen_low_port             = DFLT_LISTEN_LOW_PORT;
	o->listen_high_port            = DFLT_LISTEN_HIGH_PORT;
	o->force_iface                 = DFLT_FORCE_IFACE;
	o->net_delay                   = DFLT_NET_DELAY;
	o->check_ip                    = DFLT_CHECK_IP;
	o->mtu                         = DFLT_MTU;

	// Parsing parameters
	int arg;
	bool cfgFile = 0; // 0 if no cfg file given
	int port = -1;
	int arp = -1;

	while ((arg = getopt(argc, argv, "p:hdDVc:a")) != -1) {
		switch (arg) {
		case 'p':
			port = (int)strtol(optarg, NULL, 10);
			break;
		case 'h':
			usage(argv[0]);
			return RET_USAGE;
		case 'd': 
			o->dump = true;
			break;
		case 'D': 
			o->daemon = false;
			break;
		case 'V':
			printf("RSIP Server Version %s\n", VERSION);
			return RET_VERSION;
		case 'c':
			free(o->cfg_file);
			o->cfg_file = strdup(optarg);
			if (readOptions(o, o->cfg_file)<0) {
				logconsole(LOG_ERR,"Error in configuration file.\n");
				return ERR_CONFIG;
			}
			cfgFile = true;
			break;
		case 'a': 
			arp = 1;
			break;
		default: 
			usage(argv[0]);
			return ERR_BAD_OPTION;
		}
	}
	
	if (!cfgFile) {
		if (readOptions(o, o->cfg_file)<0) {
			fprintf(stderr,"Error in configuration file %s\n", o->cfg_file);
			return ERR_CONFIG;
		}
	}

	if (getIP(o->ext_if, o->ext_ip)==-1) {
		logconsole(LOG_ERR, "Can't get IP address of external interface %s.\n", o->ext_if);
		return ERR_EXT_IP;
	}

	if (getIP(o->int_if, o->int_ip)==-1) {
		logconsole(LOG_ERR, "Can't get IP address of internal interface %s.\n", o->int_if);
		return ERR_INT_IP;
	}

	if (!o->ext_ip.isDefined()) {
		logconsole(LOG_ERR, "External interface %s is down. Please set it up first!\n", o->ext_if);
		return ERR_EXT_IF;
	}

	if (!o->int_ip.isDefined()) {
		logconsole(LOG_ERR, "Internal interface %s is down. Please set it up first!\n", o->int_if);
		return ERR_INT_IF;
	}

	if (o->rsap_low_port > o->rsap_high_port) {
		logconsole(LOG_ERR, "Bad PORT_RANGE parameter %i-%i\n", o->rsap_low_port, o->rsap_high_port);
		return ERR_RANGE;
	}

	if (o->initial_lease_time < 60)            o->initial_lease_time = 60;
	if (o->initial_register_lease_time < 60)   o->initial_register_lease_time = 60;
	if (o->min_lease_time < 30)                o->min_lease_time = 30;
	if (o->max_lease_time < o->min_lease_time) o->max_lease_time = o->min_lease_time;

	FILE* file;
	if ((file=fopen("/proc/sys/net/ipv4/ip_local_port_range", "r"))==NULL) {
		logconsole(LOG_WARNING,"Cannot open /proc/sys/net/ipv4/ip_local_port_range\n");
		return ERR_CONFIG;
	}

	int low, high;
	if (fscanf(file, "%d %d", &low, &high)==2) {
		if (!(o->rsap_high_port<low || o->rsap_low_port>high)) {
			int range = o->rsap_high_port - o->rsap_low_port;
			int found = 0;
			
			logconsole(LOG_WARNING,
				   "Local port range %i-%i used by RSIP conflicts\n",
				   o->rsap_low_port, o->rsap_high_port);

			logconsole(LOG_WARNING, 
				   "with range %i-%i used by Linux.\n", low, high);

			for (int i=8192; i+range<65535; i+=range) {
				if (i+range<low || i>high) {
					found = 1;
					o->rsap_low_port = i;
					o->rsap_high_port = i + range;
					break;
				}
			}
			
			if (found)
				logconsole(LOG_WARNING,"Using range %i-%i instead.\n", 
					   o->rsap_low_port, o->rsap_high_port);
			else {
				logconsole(LOG_ERR,"Can't find another available range !\n");
				logconsole(LOG_ERR,"Change PORT_RANGE option in %s.\n", 
					   o->cfg_file);
			}
		}
	}
	fclose(file);

	// Overwriting options with options from the command line
	if (port!=-1) o->port = (unsigned short)port;
	if (arp!=-1) o->arp = true;

	return RET_SUCCESS;
}



//****************************************************************************
//* SPEC: Initialize the pool of Resources
//* PRE : - o is initialized
//*       - p is NULL or not initialized
//* POST: - The pool is filled with resources from config file
//*       - opt->max_clients is set to the number of IP *5 if it is equal to 0
//*       - The resources in file descriptors are increased accordingly
//*       - Returns false if an error occured
//*               true otherwise
//****************************************************************************
bool init(Resources* &p, Options* o)
{
	// Intialization of the resources
	p = new Resources(o);
	int dbSize = p->size();

	if (dbSize>1) {
		logmsg(LOG_INFO,"%i RSIP addresses available.\n", dbSize);

	} else if (dbSize==1) {
		logmsg(LOG_INFO,"One RSIP address available.\n");

	} else {

		logconsole(LOG_ERR, "No RSIP address available.\n");
		logconsole(LOG_ERR, "Please configure address pool in file %s.\n", o->cfg_file);
		return false;
	}

	if (o->max_clients == 0) opt->max_clients = p->size()*5;
	if (o->max_tcp_clients == 0) o->max_tcp_clients = opt->max_clients+1;

	// Increasing resources in file descriptors
	struct rlimit rlp;
	getrlimit(RLIMIT_NOFILE, &rlp);
	rlp.rlim_cur = rlp.rlim_max = 65536;
  
	if (setrlimit(RLIMIT_NOFILE, &rlp)!=0) { 
		logconsole(LOG_ERR, "setrlimit failed.\n"); 
	}
	getrlimit(RLIMIT_NOFILE, &rlp);
	if (rlp.rlim_cur <  p->size()+4) {
		logconsole(LOG_ERR, "Warning : %i addresses available but only %i ", 
			dbSize, rlp.rlim_cur);
		logconsole(LOG_ERR, "clients may be connected at the same time !\n");
		logconsole(LOG_ERR, "Change the maximum number of files that a process can create to\n");
		logconsole(LOG_ERR, "correct this problem.\n");
	}

	return true;
}



//****************************************************************************
//*
//* Function readOptions
//*
//* Initialize parameters defined in configuration file of name "name"
//*
//* Returns  1 if no error
//*         -1 otherwise
//****************************************************************************
int readOptions(Options* o, const char* filename) 
{
	FILE* file;
	int m, n, p, q;
	int stop = 0;
	char line[80];
	char* status = NULL;
	char optName[80];
	char iface[80];

	if ((file=fopen(filename, "r"))==NULL) {
		// No config file
		logconsole(LOG_WARNING, "%s: file doesn't exist. Using default values.\n", filename);
		return 1;
	}


	while(!stop) {
		// Reads a line
		status = fgets(line, 80, file);

		if (status==NULL) {
			fclose(file);
			return 1;
		}

		// Case 1 : a # = Comment, nothing to do
		if (line[0]=='#' || line[0]=='\n' || line[1]=='\n') continue;

		// Case 2 : normal case. Parsing parameters.
		n = sscanf(line, "%s", optName);
		if (n==EOF || n==0) {
			fclose(file);
			return -1;
		}

		if (strcmp(optName, "RSIP_PORT")==0) {
			if ((m = sscanf(line, "%s %d", optName, &p))!=EOF) {
				if (m==2) {
					if (p>=0 && p<=65535) o->port = p;
					else logmsg(LOG_WARNING, "Invalid RSIP_PORT %i\n", p);
				} else stop = -1;
			} else stop = -1;
				
		} else if (strcmp(optName, "PORT_RANGE")==0) {
			if ((m = sscanf(line, "%s %d-%d", optName, &p, &q))!=EOF) {
				if (m==3) {
					o->rsap_low_port = p;
					o->rsap_high_port = q;
				} else if (m==2) {
					o->rsap_low_port = p;
					o->rsap_high_port = p;
				} 
				else stop = -1;
			} else stop = -1;
				
		} else if (strcmp(optName, "LISTEN_RANGE")==0) {
			if ((m = sscanf(line, "%s %d-%d", optName, &p, &q))!=EOF) {
				if (m==3) {
					o->listen_low_port = p;
					o->listen_high_port = q;
				} else if (m==2) {
					o->listen_low_port = p;
					o->listen_high_port = p;
				} 
				else stop = -1;
			} else stop = -1;
				
		} else if (strcmp(optName, "INITIAL_LEASE_TIME")==0) {
			if ((m = sscanf(line, "%s %d", optName, &p))!=EOF) {
				if (m==2) {
					if (p>=0) o->initial_lease_time = p;
					else logmsg(LOG_WARNING, "Invalid INITIAL_LEASE_TIME %i\n", p);
				} else stop = -1;
			} else stop = -1;
				
		} else if (strcmp(optName, "INITIAL_REGISTER_LEASE_TIME")==0) {
			if ((m = sscanf(line, "%s %d", optName, &p))!=EOF) {
				if (m==2) {
					if (p>=0) o->initial_register_lease_time = p;
					else logmsg(LOG_WARNING, "Invalid INITIAL_REGISTER_LEASE_TIME %i\n", p);
				} else stop = -1;
			} else stop = -1;
				
		} else if (strcmp(optName, "MAXIMUM_LEASE_TIME")==0) {
			if ((m = sscanf(line, "%s %d", optName, &p))!=EOF) {
				if (m==2) {
					if (p>=0) o->max_lease_time = p;
					else logmsg(LOG_WARNING, "Invalid MAXIMUM_LEASE_TIME %i\n", p);
				} else stop = -1;
			} else stop = -1;
				
		} else if (strcmp(optName, "MAX_CLIENTS")==0) {
			if ((m = sscanf(line, "%s %d", optName, &p))!=EOF) {
				if (m==2) {
					if (p>=0 && p<=65535) o->max_clients = p;
					else logmsg(LOG_WARNING, "Invalid MAX_CLIENTS %i\n", p);
				} else stop = -1;
			} else stop = -1;
				
		} else if (strcmp(optName, "MAX_PORTS")==0) {
			if ((m = sscanf(line, "%s %d", optName, &p))!=EOF) {
				if (m==2) {
					if (p>=0 && p<=65535) o->max_host_ports = p;
					else logmsg(LOG_WARNING, "Invalid MAX_PORTS %i\n", p);
				} else stop = -1;
			} else stop = -1;
				
		} else if (strcmp(optName, "RSAIP_METHOD")==0) {
			if ((m = sscanf(line, "%s %d", optName, &p))!=EOF) {
				if (m==2) {
					if (p>=0 && p<=2) o->rsa_method=p;
					else logmsg(LOG_WARNING, "Invalid RSAIP_METHOD %i\n", p);
				} else stop = -1;
			} else stop = -1;

		} else if (strcmp(optName, "EXT_IF")==0) {
			if ((m = sscanf(line, "%s %s", optName, iface))!=EOF) {
				if (m==2) {
					if (o->ext_if) free(o->ext_if);
					o->ext_if = strdup(iface);
				} else stop = -1;
			} else stop = -1;
				
		} else if (strcmp(optName, "INT_IF")==0) {
			if ((m = sscanf(line, "%s %s", optName, iface))!=EOF) {
				if (m==2) {
					if (o->int_if) free(o->int_if);
					o->int_if = strdup(iface);
				} else stop = -1;
			} else stop = -1;
				
		} else if (strcmp(optName, "SPAWN")==0) {
			// Old option

		} else if (strcmp(optName, "ARP")==0) {
			if ((m = sscanf(line, "%s %d", optName, &p))!=EOF) {
				if (m==2) {
					if (p==1) o->arp=true;
					else o->arp=false;
				} else stop = -1;
			} else stop = -1;

		} else if (strcmp(optName, "FORCE_IFACE")==0) {
			if ((m = sscanf(line, "%s %d", optName, &p))!=EOF) {
				if (m==2) {
					if (p==1) o->force_iface=true;
					else o->force_iface=false;
				} else stop = -1;
			} else stop = -1;

		} else if (strcmp(optName, "NET_DELAY")==0) {
			if ((m = sscanf(line, "%s %d", optName, &p))!=EOF) {
				if (m==2) {
					if (p>=0) o->net_delay=p;
					else logmsg(LOG_WARNING, "Invalid NET_DELAY %i\n", p);
				} else stop = -1;
			} else stop = -1;


		} else if (strcmp(optName, "MTU")==0) {
			if ((m = sscanf(line, "%s %d", optName, &p))!=EOF) {
				if (m==2) {
					if (p>=0) o->mtu=p;
					else logmsg(LOG_WARNING, "Invalid MTU %i\n", p);
				} else stop = -1;
			} else stop = -1;

		} else if (strcmp(optName, "LOG_LEVEL")==0) {
			if ((m = sscanf(line, "%s %d", optName, &p))!=EOF) {
				if (m==2) {
					if (p>=0 && p<=7) o->log_level=p;
					else logmsg(LOG_WARNING, "Invalid LOG_LEVEL %i\n", p);
				} else stop = -1;
			} else stop = -1;
				
		} else if (strcmp(optName, "POOL")!=0) {
			logmsg(LOG_WARNING, "Unrecognized option \"%s\"\n", optName);
		}

	} // End while

	fclose(file);

	return stop;
}



//****************************************************************************
//* Function writepid
//* SPEC: write the pid of the current process in the file
//****************************************************************************
void writepid(const char* file) 
{
	FILE* f = fopen(file, "w");
	fprintf(f,"%i\n",getpid());
	fclose(f);
}
