//***************************************************************************
//* shell.cpp
//*
//* description    : class which spawn a shell command
//* copyright      : (C) 2001 Cedric de Launois
//*                  Universite Catholique de Louvain, Belgium
//* begin          : Dec 21 2001
//* email          : cedric@delaunois.org
//***************************************************************************

/****************************************************************************
 * Copyright (c) 1998 Free Software Foundation, Inc.                        *
 *                                                                          *
 * Permission is hereby granted, free of charge, to any person obtaining a  *
 * copy of this software and associated documentation files (the            *
 * "Software"), to deal in the Software without restriction, including      *
 * without limitation the rights to use, copy, modify, merge, publish,      *
 * distribute, distribute with modifications, sublicense, and/or sell       *
 * copies of the Software, and to permit persons to whom the Software is    *
 * furnished to do so, subject to the following conditions:                 *
 *                                                                          *
 * The above copyright notice and this permission notice shall be included  *
 * in all copies or substantial portions of the Software.                   *
 *                                                                          *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS  *
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF               *
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.   *
 * IN NO EVENT SHALL THE ABOVE COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,   *
 * DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR    *
 * OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR    *
 * THE USE OR OTHER DEALINGS IN THE SOFTWARE.                               *
 *                                                                          *
 * Except as contained in this notice, the name(s) of the above copyright   *
 * holders shall not be used in advertising or otherwise to promote the     *
 * sale, use or other dealings in this Software without prior written       *
 * authorization.                                                           *
 ****************************************************************************/

#include "shell.h"

Shell::Shell(bool d) 
{
	started = false;
	dump=d;
	if (pthread_mutex_init(&mutex, NULL)) throw NotEnoughMemory();
}



Shell::~Shell()
{

	if (started) {
		fclose(channel[1]);
		fclose(result[0]);
		wait((int*)0);
	}
	
	pthread_mutex_destroy(&mutex);

}



int Shell::start()
{
	int fp[2];
	int res[2];

	if (pipe(fp) < 0) {
		logmsg(LOG_ERR, "Can't create pipe\n");
		return 0;
	}

	channel[0] = fdopen(fp[0], "r");
	channel[1] = fdopen(fp[1], "w");

	if (pipe(res) < 0) {
		logmsg(LOG_ERR, "Can't create pipe\n");
		return 0;
	}

	result[0] = fdopen(res[0], "r");
	result[1] = fdopen(res[1], "w");


	if ((server=fork())==0) {
		signal(SIGINT, SIG_IGN);
		signal(SIGTERM, SIG_IGN);
		signal(SIGPIPE, SIG_IGN);
		// Child
		fclose(channel[1]); // Close writing end (command pipe)
		fclose(result[0]);  // Close reading end (result pipe)
		// Child
		serve();
		fclose(channel[0]); // Close writing end (command pipe)
		fclose(result[1]);  // Close reading end (result pipe)
		exit(0);
	}

	if (server!=-1) {
		fclose(channel[0]); // Close reading end (command pipe)
		fclose(result[1]);  // Close writing end (result pipe)
		started = true;
		return 1;
	} else {
		logmsg(LOG_ERR, "Can't fork");
		return 0;
	}
}



int Shell::execute(const char *path, const char* arg, ...)
{

	int res = 1;;

	if (!started) start();

	pthread_mutex_lock(&mutex);

	fprintf(channel[1], path);
	fputc(' ', channel[1]);
	fflush(channel[1]);

	va_list ap;
	va_start(ap, arg);
	while(arg) {
		fprintf(channel[1], arg, strlen(arg));
		fputc(' ', channel[1]);
		fflush(channel[1]);
		arg = va_arg(ap, char*);
	}
	va_end(ap);

	fputc(1, channel[1]);
	fprintf(channel[1], "\n");
	fflush(channel[1]);

	fscanf(result[0], "%i", &res);;

	pthread_mutex_unlock(&mutex);

	return res;
}


int Shell::execute(const char *path, deque<const char*> arg)
{

	int res = 1;;

	if (!started) start();

	pthread_mutex_lock(&mutex);

	fprintf(channel[1], path);
	fputc(' ', channel[1]);
	fflush(channel[1]);

	int len = arg.size();
	for (int i=0; i<len; i++) {
		const char* a = arg[i];
		fprintf(channel[1], a, strlen(a));
		fputc(' ', channel[1]);
		fflush(channel[1]);
	}

	fputc(1, channel[1]);
	fprintf(channel[1], "\n");
	fflush(channel[1]);

	fscanf(result[0], "%i", &res);;

	pthread_mutex_unlock(&mutex);

	return res;
}


void Shell::serve()
{
	char buf[256];
	char* arg[260];
	int n = 0;

	while (fscanf(channel[0], "%255s", buf) >= 0) {
		buf[255]=0;
		if (buf[0]!=1) {
			if (strlen(buf)>=255) {
				// Assume overflow !
				logmsg(LOG_WARNING, "Shell: command line for script too long\n");
				fprintf(result[1], "%i", -1);
				fflush(result[1]);
				while (n>0) { n--; if (arg[n]) free(arg[n]); }
				continue;
			}

			if (n==260) {
				logmsg(LOG_WARNING, "Shell: too much arguments\n");
				fprintf(result[1], "%i", -1);
				fflush(result[1]);
				while (n>0) { n--; if (arg[n]) free(arg[n]); }
				continue;
			}

			arg[n] = strdup(buf);
		        if (dump) cerr << arg[n] << " ";
			n++;

		} else {
			// Spawning script
			if (dump) cerr << endl;
			int child;
			int status;
			arg[n] = NULL;
			if ((child=fork())==0) {
				// Child
				execv(arg[0], arg);
				logmsg(LOG_ERR, "Shell: can't execute program %s !\n", arg[0]);
			}

			if (child==-1) {
				logmsg(LOG_ERR, "Can't fork");
				fprintf(result[1], "%i ", -1);
			} else {
				waitpid(child, &status, 0);
				fprintf(result[1], "%i ", status);
			}
			fflush(result[1]);

			while (n>0) { n--; if (arg[n]) free(arg[n]); }
		}
	}

	exit(0);
}

