//***************************************************************************
//* timer.cpp
//*
//* description    : class Timer, Implementation
//*                  Schedule various tasks
//*
//* copyright      : (C) 2001 Cedric de Launois
//*                  Universite Catholique de Louvain, Belgium
//* begin          : Sept 24 2001
//* email          : cedric@delaunois.org
//***************************************************************************

/****************************************************************************
 * Copyright (c) 1998 Free Software Foundation, Inc.                        *
 *                                                                          *
 * Permission is hereby granted, free of charge, to any person obtaining a  *
 * copy of this software and associated documentation files (the            *
 * "Software"), to deal in the Software without restriction, including      *
 * without limitation the rights to use, copy, modify, merge, publish,      *
 * distribute, distribute with modifications, sublicense, and/or sell       *
 * copies of the Software, and to permit persons to whom the Software is    *
 * furnished to do so, subject to the following conditions:                 *
 *                                                                          *
 * The above copyright notice and this permission notice shall be included  *
 * in all copies or substantial portions of the Software.                   *
 *                                                                          *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS  *
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF               *
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.   *
 * IN NO EVENT SHALL THE ABOVE COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,   *
 * DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR    *
 * OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR    *
 * THE USE OR OTHER DEALINGS IN THE SOFTWARE.                               *
 *                                                                          *
 * Except as contained in this notice, the name(s) of the above copyright   *
 * holders shall not be used in advertising or otherwise to promote the     *
 * sale, use or other dealings in this Software without prior written       *
 * authorization.                                                           *
 ****************************************************************************/

#include "timer.h"

#define for_each(map,i) for (i = map.begin(); i != map.end(); ++i )


unsigned long Timer::period = DFLT_PERIOD;


Timer::Timer() 
{
	number = 0;
	status = RUNNING;
	period = DFLT_PERIOD;
	if (pthread_create(&timer, NULL, (void* (*) (void*)) runTimer, (void*)this))
		throw NotEnoughMemory();
	if (pthread_mutex_init(&mutex, NULL)!=0)
		throw NotEnoughMemory();
}



Timer::~Timer() 
{
	status = STOPPING;
	pthread_join(timer, NULL);
	map<time_t, deque<Task>*>::iterator i;

	// Removing all the remaining tasks;
	for_each(tasks, i) delete i->second;

	pthread_mutex_destroy(&mutex);
}



int Timer::schedule(void* (*function)(void*), void* arg, unsigned int sec)
{
	map<time_t, deque<Task>*>::iterator i;
	deque<Task>* l;
	Task t;

	if (status!=RUNNING) return 0;

	t.function = function;
	t.arg = arg;
	time_t deadline = time(0)+sec;

	pthread_mutex_lock(&mutex);

	if ((i = tasks.find(deadline)) == tasks.end()) {
		l = new deque<Task>();
		tasks.insert(make_pair(deadline, l));
	} else l = i->second;
	l->push_back(t);
	number++;

	pthread_mutex_unlock(&mutex);

	return 1;
}



void Timer::doSchedule()
{
	map<time_t, deque<Task>*>::iterator i;
	time_t deadline;
	deque<Task>* l;
	Task t;

	while (status!=STOPPING) {
		usleep(period);

		pthread_mutex_lock(&mutex);

		if (!tasks.size()) {
			pthread_mutex_unlock(&mutex);
			continue;
		}

		i = tasks.begin();
		deadline = i->first;
		l = i->second;

		if (deadline<=time(0)) {
			tasks.erase(i);
			pthread_mutex_unlock(&mutex);

			while (l->size()) {
				t = l->front();
				l->pop_front();
				t.function(t.arg);
				number--;
			}
			delete l;
		} else pthread_mutex_unlock(&mutex);
	}

	status = STOPPED;
}



void runTimer(Timer* t)
{
	t->doSchedule();
}
