//***************************************************************************
//* tunnel.cpp
//*
//* description    : creation and configuration of tunnels
//* copyright      : (C) 2001 Cedric de Launois
//*                  Universite Catholique de Louvain, Belgium
//* begin          : Sept 13 2001
//* email          : cedric@delaunois.org
//***************************************************************************

/****************************************************************************
 * Copyright (c) 1998 Free Software Foundation, Inc.                        *
 *                                                                          *
 * Permission is hereby granted, free of charge, to any person obtaining a  *
 * copy of this software and associated documentation files (the            *
 * "Software"), to deal in the Software without restriction, including      *
 * without limitation the rights to use, copy, modify, merge, publish,      *
 * distribute, distribute with modifications, sublicense, and/or sell       *
 * copies of the Software, and to permit persons to whom the Software is    *
 * furnished to do so, subject to the following conditions:                 *
 *                                                                          *
 * The above copyright notice and this permission notice shall be included  *
 * in all copies or substantial portions of the Software.                   *
 *                                                                          *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS  *
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF               *
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.   *
 * IN NO EVENT SHALL THE ABOVE COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,   *
 * DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR    *
 * OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR    *
 * THE USE OR OTHER DEALINGS IN THE SOFTWARE.                               *
 *                                                                          *
 * Except as contained in this notice, the name(s) of the above copyright   *
 * holders shall not be used in advertising or otherwise to promote the     *
 * sale, use or other dealings in this Software without prior written       *
 * authorization.                                                           *
 ****************************************************************************/

#include "tunnel.h"

// Static variables
Shell* Tunnel::shell = NULL;
Proxy* Tunnel::proxy = NULL;
Options* Tunnel::opts = NULL;
bool Tunnel::initRSAP = false;
map<string, unsigned int> Tunnel::tunnelsID;
map<string, string> Tunnel::tunnels;
pthread_mutex_t Tunnel::lock = PTHREAD_MUTEX_INITIALIZER;

Tunnel::Tunnel() 
{
	assert(shell!=NULL && proxy!=NULL && opts!=NULL);
	created = 0;
	name = "";
	nbPorts = -1;
	type = 1; // IPIP by default
}



Tunnel::Tunnel(Address &rf, Address &rt, Address &f, Address &t, int typ) 
{
	realFrom = rf;
	realTo = rt;
	from = f;
	to = t;
	created = 0;
	name = "";
	nbPorts = -1;
	type = typ;
}



Tunnel::~Tunnel() 
{
	if (created) {
		if (nbPorts>0) intDeleteRSAP();
		else intDeleteRSA();
	}
}



int Tunnel::create(string &n) 
{
	if (created) return 1;

	if (nbPorts>0) return intCreateRSAP(n);
	else return intCreateRSA(n);
}



void Tunnel::setup(Shell* s, Proxy* p, Options* o)
{
	shell = s;
	proxy = p;
	opts = o;
}



int Tunnel::intCreateRSA(string &n)
{
	int err;

	// Creating first the tunnel
	if ((err=createTunnel(n))<0) return err;

	if (to == opts->ext_ip) {
		if ((err=proxy->addRoute(to, 0, n.c_str(), 0))!=1) {
			intDeleteRSA();
			created = false;
			return err;
		}
	}

	created = true;
	return 1;
}



void Tunnel::intDeleteRSA()
{

	if (to == opts->ext_ip) {
		proxy->delRoute(to, 0, 0);
	}

	deleteTunnel();

	created = false;
}



int Tunnel::intCreateRSAP(string &n)
{
	int err, len;

	// Creating first the tunnel
	if ((err=createTunnel(n))<0) return err;

	// Configuring the proxy
	len = ports.size();

	if (nbPorts!=1 && len==1) {

		if ((err = proxy->addRoute(to, ports[0], n.c_str(), nbPorts))!=1) {
			intDeleteRSA();
			created = false;
			return err;
		}

	} else {

		for (int i=0; i<len; i++) {
			if ((err=proxy->addRoute(to, ports[i], n.c_str(), 1))!=1) {
				intDeleteRSA();
				created = false;
				return err;
			}
		}
	}

	name = n;
	created = true;

	return 1;
}



void Tunnel::intDeleteRSAP()
{
	int len = ports.size();
	if (nbPorts!=1 && len==1) proxy->delRoute(to, ports.front(), nbPorts);
	else {
		for (int i=0; i<len; i++) proxy->delRoute(to, ports[i], 1);
	}

	deleteTunnel();

	created = false;
}



int Tunnel::createTunnel(string &n)
{
	int err;
	string id = string(realFrom.toChar()) + ":" + string(realTo.toChar()); // The unique id for a tunnel
	unsigned int number;
	map<string, unsigned int>::iterator i;

	pthread_mutex_lock(&lock);
	if ((i=tunnelsID.find(id)) != tunnelsID.end()) {
		// A tunnel with same source and destination address exists
		i->second++;
		number = i->second;
		n = (tunnels.find(id))->second;
	} else {
		number = 1;
		tunnelsID.insert(make_pair(id, number));
		tunnels.insert(make_pair(id, n));
	}
	pthread_mutex_unlock(&lock);

	name = n;
	if (number>1) {
		// The tunnel has already been created previously
		created = true;
		return 1;
	}

	// Creating tunnel
	if (type==TUNNEL_IP_IP) {
		if (!(err = do_iptunnel("add", 
					n.c_str(), 
					IPPROTO_IPIP, 
					realFrom.toChar(), 
					realTo.toChar(), 
					opts->force_iface?opts->int_if:NULL))) {
			logmsg(LOG_ERR, "Can't create IP/IP tunnel. Error %d.\n", err);
			return -1;
		}
	} else {
		if (!(err = do_iptunnel("add", 
					n.c_str(), 
					IPPROTO_GRE, 
					realFrom.toChar(), 
					realTo.toChar(), 
					opts->force_iface?opts->int_if:NULL))) {
			logmsg(LOG_ERR, "Can't create IP/GRE tunnel. Error %d.\n", err);
			return -2;
		}
	}
	

	// Setting up tunnel
	if ((err = do_ifconfig(n.c_str(), from.toChar(), to.toChar(), opts->mtu))<=0) {
		logmsg(LOG_ERR, "Can't configure tunnel. Error %d.\n", err);
		return -3;
	}

	// Updating ARP table if needed
	if (opts->arp) {
		if (!(err = arp_set(to.toChar(), opts->ext_if))) {
			logmsg(LOG_ERR, "Can't update ARP table. Error %d.\n", err);
			return -4;
		}
	}

	created = true;
	return 1;
}



void Tunnel::deleteTunnel()
{
	int typ;
	int err;
	string id = string(realFrom.toChar()) + ":" + string(realTo.toChar()); // The unique id for a tunnel
	unsigned int number;
	map<string, unsigned int>::iterator i;

	// Here we check if the tunnel must be physically deleted
	pthread_mutex_lock(&lock);
	if ((i=tunnelsID.find(id)) != tunnelsID.end()) {
		i->second--;
		number = i->second;
		if (number==0) {
			tunnelsID.erase(i);
			tunnels.erase(tunnels.find(id));
		}
	} else {
		logmsg(LOG_ERR, "Can't remove tunnel %s\n", id.c_str());
		pthread_mutex_unlock(&lock);
		return;
	}
	pthread_mutex_unlock(&lock);

	created = false;
	if (number>0) return; // The tunnel must be keep

	typ = (type == TUNNEL_IP_IP) ? IPPROTO_IPIP : IPPROTO_GRE;
	if (!(err = do_iptunnel("del", name.c_str(), typ, realFrom.toChar(), realTo.toChar(), NULL))) {
		logmsg(LOG_ERR, "Error deleting tunnel. Error: %d.\n", err);
	}
	
	if (opts->arp) {
		if (!(err = arp_del(to.toChar(), opts->ext_if))) {
			logmsg(LOG_ERR, "Error removing tunnel. Error: %d\n", err);
		}
	}
}
