//***************************************************************************
//* tunnel.h
//*
//* description    : Declarations of class tunnel
//*                  Operations on tunnels
//* copyright      : (C) 2001 Cedric de Launois
//*                  Universite Catholique de Louvain, Belgium
//* begin          : Dec 21 2001
//* email          : cedric@delaunois.org
//***************************************************************************

/****************************************************************************
 * Copyright (c) 1998 Free Software Foundation, Inc.                        *
 *                                                                          *
 * Permission is hereby granted, free of charge, to any person obtaining a  *
 * copy of this software and associated documentation files (the            *
 * "Software"), to deal in the Software without restriction, including      *
 * without limitation the rights to use, copy, modify, merge, publish,      *
 * distribute, distribute with modifications, sublicense, and/or sell       *
 * copies of the Software, and to permit persons to whom the Software is    *
 * furnished to do so, subject to the following conditions:                 *
 *                                                                          *
 * The above copyright notice and this permission notice shall be included  *
 * in all copies or substantial portions of the Software.                   *
 *                                                                          *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS  *
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF               *
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.   *
 * IN NO EVENT SHALL THE ABOVE COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,   *
 * DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR    *
 * OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR    *
 * THE USE OR OTHER DEALINGS IN THE SOFTWARE.                               *
 *                                                                          *
 * Except as contained in this notice, the name(s) of the above copyright   *
 * holders shall not be used in advertising or otherwise to promote the     *
 * sale, use or other dealings in this Software without prior written       *
 * authorization.                                                           *
 ****************************************************************************/

class Tunnel;

#ifndef _TUNNEL_H_
#define _TUNNEL_H_

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>
#include <pthread.h>
#include <string>
#include <deque>
#include <map>

#include "common.h"
#include "iptunnel.h"
#include "ifconfig.h"
#include "arp.h"
#include "proxy.h"
#include "address.h"
#include "assert.h"

#ifndef TUNNEL_IP_IP
#define TUNNEL_IP_IP 1
#endif
#ifndef TUNNEL_GRE
#define TUNNEL_GRE   2
#endif

class Tunnel {
 public:

	Tunnel();
	// Constructor
	// PRE : Tunnel global parameters must be set by a call to setup()

	~Tunnel();
	// Destructor


	Tunnel(Address &realFrom, Address &realTo, Address &from, Address &to, int type);
	// Constructor
	// PRE : - realFrom is the real source address
	//       - realTo is the real destination address
	//       - from is the address of the beginning of the tunnel
	//       - to is the address of the end of the tunnel
	//       - type is the type of the tunnel : TUNNEL_IP_IP or TUNNEL_GRE
	// POST: tunnel defined, NO created


	int create(string &name);
	// Create the tunnel
	// PRE : name is a unique name for the tunnel
	// POST: returns 1 in case of success, 0 or a negative number otherwise
	//       The name of the tunnel may change !


	inline string getName() { return name; }
	// Returns the name of the tunnel


	inline void setPorts(unsigned char n, deque<uint16_t> p) { nbPorts=n; ports=p; }
	// PRE : n is the number of ports which restrict the use of the tunnel
	//       p : the list of ports to be used for this tunnel
	// POST: only used for RSAP-IP : restrict the use of the tunnel to
	//       the ports specified


	static void setup(Shell* s, Proxy* p, Options* o);
	// Setup global parameters for all tunnels
	// Must be call once, before using the first tunnel constructor


 private:

	bool created;
	string  name;     // The ID of the tunnel
	Address realFrom; // Real source address
	Address realTo;   // Real destination address
	Address from;     // Address of the beginning of the tunnel
	Address to;       // Address of the end of the tunnel
	int type;         // Type of the tunnel

	int nbPorts;
	deque<uint16_t> ports;

	// RSA
	int intCreateRSA(string &name);
	void intDeleteRSA();

	// RSAP
	int intCreateRSAP(string &name);
	void intDeleteRSAP();

	int createTunnel(string &name);
	void deleteTunnel();

	static bool initRSAP;
	static Shell* shell;
	static Proxy* proxy;
	static Options* opts;

	// Table <id name of tunnel, number of time this tunnel is used>
	// The id name of a tunnel is "SRC_IP:DST_IP" (unique)
	static map<string, unsigned int> tunnelsID;

	// Table <id name of tunnel, interface name>
	static map<string, string> tunnels;
	static pthread_mutex_t lock;
};

#endif
