//**************************************************************************
//* UDPEndPoint.cpp
//*
//* description    : Methods of class TCPEndPoint. 
//*                  End point of a TCP connection
//* copyright      : (C) 2001 Cedric de Launois
//*                  Universite Catholique de Louvain, Belgium
//* begin          : Sept 11 2001
//* email          : cedric@delaunois.org
//**************************************************************************

/****************************************************************************
 * Copyright (c) 1998 Free Software Foundation, Inc.                        *
 *                                                                          *
 * Permission is hereby granted, free of charge, to any person obtaining a  *
 * copy of this software and associated documentation files (the            *
 * "Software"), to deal in the Software without restriction, including      *
 * without limitation the rights to use, copy, modify, merge, publish,      *
 * distribute, distribute with modifications, sublicense, and/or sell       *
 * copies of the Software, and to permit persons to whom the Software is    *
 * furnished to do so, subject to the following conditions:                 *
 *                                                                          *
 * The above copyright notice and this permission notice shall be included  *
 * in all copies or substantial portions of the Software.                   *
 *                                                                          *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS  *
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF               *
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.   *
 * IN NO EVENT SHALL THE ABOVE COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,   *
 * DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR    *
 * OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR    *
 * THE USE OR OTHER DEALINGS IN THE SOFTWARE.                               *
 *                                                                          *
 * Except as contained in this notice, the name(s) of the above copyright   *
 * holders shall not be used in advertising or otherwise to promote the     *
 * sale, use or other dealings in this Software without prior written       *
 * authorization.                                                           *
 ****************************************************************************/


#include "udpEndPoint.h"


UDPEndPoint::UDPEndPoint() 
{
 	port = 0;
	sock = 0;
	connected = 0;
}



UDPEndPoint::~UDPEndPoint() 
{
	if (connected) close(sock);
}



void UDPEndPoint::openPort(const char* ifname, unsigned short p, const Address *addr)
{
	if (connected) close(sock);

	if (p<1024 || p>65535) throw BadPort();

	if (ifname) {
		// Get the netmask of the authorized network
		struct ifreq ifr;
		int fd;
	
		bzero(&ifr, sizeof(ifr));
		strncpy(ifr.ifr_name, ifname, IFNAMSIZ);
		if ((fd=socket(AF_INET, SOCK_DGRAM, 0))<0) goto out;

		if (ioctl(fd, SIOCGIFNETMASK, &ifr)<0) {
			close(fd);
			goto out;
		}
		authNetmask.setAddr(((struct sockaddr_in*)&ifr.ifr_netmask)->sin_addr.s_addr);

		if (ioctl(fd, SIOCGIFADDR, &ifr)<0) {
			close(fd);
			goto out;
		}
		close(fd);
		authNet.setAddr(((struct sockaddr_in*)&ifr.ifr_addr)->sin_addr.s_addr & authNetmask.getAddr());
	}
 out:

	if (addr) {  // Port of type 'client'
		address = *addr;
	} else {        // Port of type 'Server'
		address.setAddr(htonl(INADDR_ANY));
	}

	address.setPort(htons(p));

	if ( (sock = socket(PF_INET, SOCK_DGRAM, 0)) < 0) throw SocketOpeningError();

	if (!addr) {
		if (bind(sock, address.getSockaddr(), sizeof(struct sockaddr_in)) < 0) {
			close(sock);
			throw BindError();
		}
	}

	port = p;
	connected = 1;
}



int UDPEndPoint::readn(char* ptr, int nbytes, int timeout, Address& from)
{
	if (!connected) throw ClosedPortError();

	int nread, sel;
	struct timeval t;
	struct sockaddr_in cli_addr;
	fd_set readfds, errorfds;

	t.tv_sec = 0;
	t.tv_usec = timeout;
	FD_ZERO(&readfds);
	FD_ZERO(&errorfds);
	FD_SET(sock, &readfds);
	FD_SET(sock, &errorfds);


	if (timeout >= 0) sel = select(FD_SETSIZE, &readfds, NULL, &errorfds, &t);
	else sel = select(FD_SETSIZE, &readfds, NULL, &errorfds, NULL);

	if (sel==0) return -1; // Time out
	else if (sel==-1 || FD_ISSET(sock, &errorfds)) throw SelectError(); // Error

	socklen_t s = sizeof(from);

	while(1) {
		nread = recvfrom(sock, ptr, nbytes, 0, (struct sockaddr*)&cli_addr, &s);
		if (nread < 0) {
			throw SocketReadError();
		}

		from = cli_addr;

		// Check the origin if needed
		if (!authNet.isDefined()) break;
		if ((cli_addr.sin_addr.s_addr & authNetmask.getAddr()) == authNet.getAddr()) break;
		if (inet_netof(cli_addr.sin_addr) == IN_LOOPBACKNET) break;

		logmsg(LOG_WARNING, "Unauthorized packet coming from %s.\n", from.toChar());
		return 0;
	}

	return nread;  // returns >= 0, the number of bytes read
}



int UDPEndPoint::writen(char* ptr, int nbytes, const Address& dest)
{
	if (!connected) throw ClosedPortError();

	int    nwritten, result;
	struct timeval t;
	fd_set fdset;

	t.tv_sec = 0;
	t.tv_usec = 0;
	FD_ZERO(&fdset);
	FD_SET(sock, &fdset);
	result = select(FD_SETSIZE, NULL, &fdset, &fdset, &t);
	if (result<1) throw SocketWriteError();

	nwritten = sendto(sock, ptr, nbytes, 0, dest.getSockaddr(), sizeof(struct sockaddr_in));
	if (nwritten != nbytes) {
		throw SocketWriteError();
	}

	return(nwritten);  // returns >= 0
}



int UDPEndPoint::disconnect() 
{
	if (connected) close(sock);
	connected = 0;
	return 1;
}



int UDPEndPoint::getfd() const 
{
	if (connected) return sock; else return -1;
}



int operator==(const UDPEndPoint &uep1, const UDPEndPoint &uep2)
{
	if (uep1.connected && uep2.connected && uep1.sock==uep2.sock)
		return 1;
	else return 0;
}

